package no.uio.ifi.rcos.algebraic.createmodel

import org.scalatest.FlatSpec
import org.scalatest.matchers.ShouldMatchers
import no.uio.ifi.rcos.algebraic.spec.CreateModel
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import edu.unu.iist.rcos.spec.antlr.parser.SemanticError
import edu.unu.iist.rcos.mm.util.TypingException
import edu.unu.iist.rcos.spec.antlr.MyDesign._
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRenameException
import no.uio.ifi.rcos.algebraic.utility.exceptions.NoSuchOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.NoSuchOperationException
import edu.unu.iist.rcos.profile.Stereotypes
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedInterfaceException

class AddingOperationsTest extends FlatSpec with ShouldMatchers {

  "Adding an operation decl in an interface" should " not cause null pointer exception" in {
    
	val interfaceString ="interface Foo { int val; public food(int inn; int out); public baz(); " +
			"public methodWithImpl(int inn; int ut) { ut := inn }}" 
    var createModel = new CreateModel
    createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
    val foo = createModel.model.getInterface("Foo", null, null)
    foo.getOperation("food", null, null) should not be (null)
    foo.getOperation("baz", null, null) should not be (null)
    createModel.model.save()
  }

  "Adding an operation in a class" should " not cause null pointer exception" in {
	val interfaceString ="class Foo { public food(int inn; int out) {} public baz(){} }" 
    var createModel = new CreateModel
    createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
    val foo = createModel.model.getClass("Foo", null, null)
    foo.getOperation("food", null, null) should not be (null)
    foo.getOperation("baz", null, null) should not be (null)
  }

  "Fields in interfaces that is provided by a component" should "get added to the component" in {
    val prog = "interface A {int i; double d; string s; Cls cls;} component K {provided A {} class Cls{}}"
    var createModel = new CreateModel
	createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
	val k = createModel.model.getComponent("K", null, null)
	val i = createModel.model.getInterface("A",null,null)
	k.getOwnedAttribute("i", createModel.model.getTypeFromTYPE(createInt(), k)) should not be (null)
	k.getOwnedAttribute("d", createModel.model.getTypeFromTYPE(createDouble(), k)) should not be (null)
    

    //associations
	//k.getAss("s", createModel.model.getTypeFromTYPE(createDouble(), k)) should not be (null)
	//k.getOwnedAttribute("cls", createModel.model.getTypeFromTYPE(createString(), k)) should not be (null)
    
    createModel.model.save
  }
  
  "Renaming same old method more than once" should " throw DuplicateRenameException " in {

    
    val prog = "interface A {public i();} component K {provided A {public i() {}}} KRenamed = K[i <- new_i, i <- new_i_again];"
    var createModel = new CreateModel
    intercept [DuplicateRenameException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
	}
  }

  "Renaming methods to same name" should " throw DuplicateRenameException " in {
    
    val prog = "interface A {public i(); public j();} component K {provided A {public i(){} public j(){}}} KRenamed = K[i <- a, j <- a];"
    var createModel = new CreateModel
    intercept [DuplicateRenameException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
	}
  }

  "Using restriction on a method that doesnt exist " should "throw NoSuchOperationException" in {
    val prog = "interface A {int i;} component K {provided A {}} KRestricted = K \\ {dontExist};"
    var createModel = new CreateModel
    intercept [NoSuchOperationException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
	}
  } 
  
  "Provided methods" should "be added to the component" in {
    val prog = "interface A {public foo(); public bar(); public foobar();} component K { provided A {public foo(){} public bar(){} public foobar(){}} }"
    var createModel = new CreateModel
	createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
	var component = createModel.model.getComponent("K", null, null)
	component.getOwnedOperation("foo", null, null) should not be (null)
	component.getOwnedOperation("bar", null, null) should not be (null)
	component.getOwnedOperation("foobar", null, null) should not be (null)
  }

  "Using restriction on a component" should "work both on rCOS and UML level" in {
    val prog = "interface A {public foo(); public bar(); public foobar();} " +
    		"component K { provided A {public foo(){} public bar(){} public foobar(){}} }" +
    		" K_Restricted = K \\ {	foo, bar };"
    var createModel = new CreateModel
	createModel.createModelFromFileSpec("hiding.uml", "hiding", new MyRCOSModel.RCOSFromString(prog))
	var component = createModel.model.getComponent("K_Restricted", null, null)
	
	//UML level
	component.getOwnedOperation("foo", null, null) should be (null) //Hidden
	component.getOwnedOperation("bar", null, null) should be (null) //Hidden
	component.getOwnedOperation("foobar", null, null) should not be (null) //not hidden

    //rCOS level
    var rcosComp = Stereotypes.isHidingComponent(component)
    rcosComp.getMethod("foo", null, null) should not be (null)
    rcosComp.getMethod("bar", null, null) should not be (null)
    rcosComp.getMethod("foobar", null, null) should be (null)
    createModel.model.save()
  }
}
