package no.uio.ifi.rcos.algebraic.createmodel

import org.scalatest.FlatSpec
import org.scalatest.matchers.ShouldMatchers
import no.uio.ifi.rcos.algebraic.spec.CreateModel
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import scala.collection.JavaConversions._
import org.eclipse.uml2.uml.Operation
import no.uio.ifi.rcos.algebraic.utility.OperationSet
import scala.collection.mutable.Seq
import scala.collection.mutable.Buffer

class ParallelCompositionTests extends FlatSpec with ShouldMatchers {

 "A = B || C" should "calculate correct provided and required methods" in {
   
   val prog = "interface F {int i; public foo(); public bar(); } interface E { double d; public foobar(); }" +
   		"component B {provided F {public foo(){} public bar(){}}}" +
   		"component C {provided E {public foobar(){} } required F; } " +
   		"A = B || C;"
   		
   var createModel = new CreateModel
   createModel.createModelFromFileSpec("parallel.uml", "parallel", new MyRCOSModel.RCOSFromString(prog))
   val a = createModel.model.getComponent("A", null, null)
   a.getAppliedStereotype("rCOS::Parallel") should not be null
   //required F from C should not be required by A because its provided by B.
   a.getRequireds().size() should be (0)

   val ops = a.getOwnedOperations()
   ops.size() should be (3)
  
   //union of the provided
   a.getOwnedOperation("foo", null, null) should not be (null)
   a.getOwnedOperation("bar", null, null) should not be (null)
   a.getOwnedOperation("foobar", null, null) should not be (null)
   
   a.getOwnedAttributes().size should be (2)
 }
 
  "A = B || C with overlapping required methods" should "calculate correct provided and required methods" in {
   
   val prog = "interface F {int i; public foo(); public bar(); } " +
   		"interface E { double d; public foo(); }" +
   		"component B { required E;}" +
   		"component C {required F;} " +
   		"A = B || C;"
   		
   var createModel = new CreateModel
   createModel.createModelFromFileSpec("parallel.uml", "parallel", new MyRCOSModel.RCOSFromString(prog))
   val a = createModel.model.getComponent("A", null, null)
   a.getAppliedStereotype("rCOS::Parallel") should not be null
   a.getRequired("B_parallel_C_required")
 }
 
 
 "Should remove multiple equal require methods" should "work" in {
   val prog = "interface A {public foo(); public bar();}" +
   		"interface B { public foo(); public foobar(); public bard();}" +
   		"component C { required A; required B;}"
   var createModel = new CreateModel
   createModel.createModelFromFileSpec("parallel.uml", "parallel", new MyRCOSModel.RCOSFromString(prog))
   val comp = createModel.model.getComponent("C", null, null)
   val requireds = comp.getRequireds

   val lst = new OperationSet(Buffer.empty[Operation]) 

   requireds.foreach(inf => lst ++= inf.getOwnedOperations())
   lst.size should be (4)
 }
  
  "A \\ B where A and B are sets of operations" should "compute the correct result " in {
    		val prog = "interface A { public foo(); public foobar(); public bard();}" +
    				"interface B {public foo(); public bar(); public barbar();}"
    var createModel = new CreateModel
    createModel.createModelFromFileSpec("setdiff.uml", "setdiff", new MyRCOSModel.RCOSFromString(prog))
    val a = createModel.model.getInterface("A", null, null)
    val b = createModel.model.getInterface("B", null, null)

    val first = new OperationSet(Buffer.empty[Operation]) 
    val second = new OperationSet(Buffer.empty[Operation]) 

    first ++= a.getOwnedOperations()
    second ++= b.getOwnedOperations()
    
    first diff second

    first.size should be (2)

    for (op <- first) {
      println("OP:" + op.getName())
    }
  }
}