package no.uio.ifi.rcos.algebraic.createmodel

import org.scalatest.FlatSpec
import org.scalatest.matchers.ShouldMatchers
import no.uio.ifi.rcos.algebraic.spec.CreateModel
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import no.uio.ifi.rcos.algebraic.utility.exceptions.SimpleAssignmentException
import scala.collection.JavaConversions._
import no.uio.ifi.rcos.algebraic.utility.exceptions.RequiredInterfaceEnumException

class SimpleComponentTest extends FlatSpec with ShouldMatchers {

 "Creating a model of simple.rcosp" should "validate" in {
    val fileName = "testcases/modelcreator/simple.rcosp"
    val modelName = "SimpleComposition.uml"
    
    val createModel = new CreateModel
    createModel.createModelFromFileSpec(modelName, "SimpleComposition", new MyRCOSModel.RCOSFromFile(fileName))
    //createModel.model.validate() should be (true) //Causes NPE deep inside topcased stuff
    createModel.model.save()
  }

  "Expressions like A = B; " should "not parse and throw SimpleAssignmentException" in {
    val prog = "A = B;"
    var createModel = new CreateModel
    intercept [SimpleAssignmentException] {
      createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
    }
  }
 
 "Enumeration of a required interface" should "create multiple interfaces" in {
   val prog = "interface A {int k; Cls cls; public foo(int k, Cls cl; double d); public bar(); }" +
   		"component K {required A[3]; class Cls{}}"


   val createModel = new CreateModel
   createModel.createModelFromFileSpec("requiredenum.uml", "requiredenum", new MyRCOSModel.RCOSFromString(prog))

   //check that we didn't ruin the original A
   val a = createModel.model.getInterface("A", null, null) 
   a should not be (null)
   a.getOwnedOperations().size should be (2)
   a.getOwnedOperation("foo", null, null) should not be (null)
   a.getOwnedOperation("bar", null, null) should not be (null)
   
   //Assert that everything is in place with the new interfaces
   val a2 = createModel.model.getInterface("A2", null, null) 
   a2 should not be (null)
   a2.getOwnedOperations().size should be (2)
   a2.getOwnedOperation("foo2", null, null) should not be (null)
   a2.getOwnedOperation("bar2", null, null) should not be (null)

   val a3 = createModel.model.getInterface("A3", null, null) 
   a3 should not be (null)
   a3.getOwnedOperations().size should be (2)
   a3.getOwnedOperation("foo3", null, null) should not be (null)
   a3.getOwnedOperation("bar3", null, null) should not be (null)
   
   createModel.model.save
 }

 "Testing operation copying" should "test how operations get copied gets it right" in {
  val prog = "interface A {public foo(); public bar();} " +
  		"component K { provided A { public foo(){} public bar() {} }} component KK { }";
  
  val createModel = new CreateModel
  createModel.createModelFromFileSpec("opcopy.uml", "opcopy", new MyRCOSModel.RCOSFromString(prog))
  var k = createModel.model.getComponent("K", null, null)
  var kk = createModel.model.getComponent("KK", null, null)
  var foo = k.getOwnedOperation("foo", null, null)
  var bar = k.getOwnedOperation("bar", null, null)
  //Now component k loses foo and bar - which we dont want to happen
  kk.getOwnedOperations().add(foo)
  kk.getOwnedOperations().add(bar)
  
   createModel.model.save
 }

 "Renaming a method" should "work" in {
   val prog = "interface A {public foo(int i; double d); public bar(); public foobar(); }" +
   		"component K { provided A {public foo(int i; double d){} public bar() {} public foobar() {}}} Z = K [foo <- fooz, bar <- barz];"

   val createModel = new CreateModel
   createModel.createModelFromFileSpec("rename.uml", "rename", new MyRCOSModel.RCOSFromString(prog))
   createModel.model.save
 }
 
 "Renaming an overloaded method " should "work" in {
   val prog = "interface A {public foo(int i;); public foo(double d;); public bar(); public foobar(); }" +
   		"component K { provided A {public foo(int i;){} public foo(){} public bar() {} public foobar() {}}} Z = K [foo <- fooz];"

   val createModel = new CreateModel
   createModel.createModelFromFileSpec("rename2.uml", "rename2", new MyRCOSModel.RCOSFromString(prog))
   createModel.model.save
 }

 "Test creating a model" should "work" in {
   val prog = "component A{} component B{} component C{} D = A || B || C;"

   val createModel = new CreateModel
   createModel.createModelFromFileSpec("nested.uml", "nested", new MyRCOSModel.RCOSFromString(prog))
   createModel.model.save
 }

 "Test creating a model" should "work aswell" in {
   val prog = "interface M {int i; } "

   val createModel = new CreateModel
   createModel.createModelFromFileSpec("nested2.uml", "nested2", new MyRCOSModel.RCOSFromString(prog))
   createModel.model.save
 }
}