package no.uio.ifi.rcos.algebraic.spec

import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import org.eclipse.emf.common.util.URI
import org.eclipse.uml2.uml.Interface
import org.eclipse.uml2.uml.Component
import org.antlr.runtime.tree.CommonTree
import org.eclipse.uml2.uml.Usage
import edu.unu.iist.rcos.mm.util.WellformedException
import edu.unu.iist.rcos.spec.antlr.RCOSModelCheckException
import org.eclipse.uml2.uml.InterfaceRealization
import org.eclipse.uml2.uml.UMLFactory
import edu.unu.iist.rcos.profile.rCOSComponent
import org.eclipse.uml2.uml.UMLPackage
import edu.unu.iist.rcos.profile.Stereotypes
import org.eclipse.emf.common.util.EList
import edu.unu.iist.rcos.profile.RenamePair
import org.eclipse.uml2.uml.Operation
import scala.collection.JavaConversions._
import edu.unu.iist.rcos.mm.util.Utility
import edu.unu.iist.rcos.mm.util.Utility.MyPort
import org.eclipse.uml2.uml.Port
import no.uio.ifi.rcos.algebraic.utility.exceptions.NoSuchOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRenameException
import edu.unu.iist.rcos.profile.impl.{RenamePairImpl, ProfilePackageImpl}
import edu.unu.iist.rcos.profile.{Renaming, ProfilePackage, ProfileFactory}
import org.eclipse.uml2.uml.Classifier
import no.uio.ifi.rcos.algebraic.utility.exceptions.RequiredInterfaceEnumException
import org.eclipse.uml2.uml.Dependency
import no.uio.ifi.rcos.algebraic.extractors._
import scala.collection.mutable.Buffer
import scala.collection.mutable.Set
import scala.collection.mutable.Seq
import no.uio.ifi.rcos.algebraic.utility._
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRequiredInterfaceException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedInterfaceException
import no.uio.ifi.rcos.algebraic.typechecking.CompositionExpressionTypeChecker

/**
 * An extension of MyRCOSModel which add new methods and overriding existing ones for creating the UML model.
 * @author martinhb
 */
class MyExtendedRCOSModel(uri: URI, modelName : String) extends MyRCOSModel(uri, modelName) 
	with CopyAttributes with CopyOperation with CopyClasses with CompositionExpressionTypeChecker {

  /**
   * Add a required interface to a component, possibly with an enumeration (multiple required interfaces)
   */
  def addRequiredInterface(requiredInterface : Interface, component : Component, t : CommonTree, enum : Int) {

    if (component.getRequired(requiredInterface.getName) != null) 
    	throw new DuplicateRequiredInterfaceException("" +
    			"Duplicate required interface " + requiredInterface.getName + " in " + component.getName +
    			". Use enumeration instead")

    if (enum < 0) 
      throw new RequiredInterfaceEnumException("enum was below zero (was: " + enum)
    if (enum == 0 || enum == 1) { // only one interface required 
      addInterfaceUsage(requiredInterface, component)
    }
    else { //2 or more required, e.g required A[3];
      addInterfaceUsage(requiredInterface, component)
      for (interfaceNo <- 2 to enum) {
    	  val name = requiredInterface.getName + interfaceNo
    	  val existingInterface = rcos.getInterface(name, null)
    	  if (existingInterface == null) {
    	    val newInterface = addInterface(name)
    	    extractAndCopyAttributes(requiredInterface, newInterface, this)
    	    addInterfaceUsage(newInterface, component)

    	    val operations = requiredInterface.getOwnedOperations()
    	    for (operation <- operations) {
    	      extractAndCopyOperation(newInterface, operation, operation.getName + interfaceNo, this)
    	    }
    	  }
    	  else {
    		addInterfaceUsage(existingInterface, component) //interface exists already - might be source of user error - but don't handle it now
    	  }
      }
    }
  }

  /**
   * Create a usage (required interface) between an interface and a component 
   */
  def addInterfaceUsage(interface: Interface, component: Component) {
    val ir: Usage = component createUsage interface
    component.getPackagedElements add ir; 
    ir setName (interface.getLabel)
  }

  /**
   * Add a provided interface to a component
   */
  def addProvidedInterface(i : Interface, co : Component) : InterfaceRealization = {

    //already provided?
    if (co.getProvided(i.getName) != null)
      throw new DuplicateProvidedInterfaceException("Duplicate provided interface " + i.getName + "in component " + co.getName)
    //Get fields in interface and add to this component.
    extractAndCopyAttributes(i, co, this)
    val interfaceRealization = co.createInterfaceRealization(i.getName(), i)
    interfaceRealization
  }

  /**
   *   Create a parallel composition component 
   */
  def addParallelCompositionComponent(
      left: rCOSComponent,
      right : rCOSComponent) : rCOSComponent = {
    
    val leftBase = left.getBase_Component
    val rightBase = right.getBase_Component
    
    checkParallel(leftBase, rightBase) //tc

    val name = leftBase.getName() + "_parallel_" + rightBase.getName()
    val newComponent = createUMLComponent(name)
    createUMLDependency(newComponent, leftBase)
    createUMLDependency(newComponent, rightBase)

    val parallel = Stereotypes.applyParallelComponent(newComponent)
    parallel.setLeft(left)
    parallel.setRight(right)
   
    val providedOps = calculateNewProvidedInterface(leftBase, rightBase)
    createProvided(providedOps, newComponent, name)
    //required
    val requiredOps = calculateNewRequiredInterface(leftBase, rightBase)
    requiredOps diff providedOps
    createRequired(requiredOps, newComponent, name)
    extractAndCopyAttributes(leftBase, newComponent, this)
    extractAndCopyAttributes(rightBase, newComponent, this)

    //Atomic tc
    
    parallel
  }

  /**
   *   Create a disjoint composition component 
   */
  def addDisjointCompositionComponent(
      left: rCOSComponent,
      right: rCOSComponent) : rCOSComponent = {

    val leftBase = left.getBase_Component
    val rightBase = right.getBase_Component

    checkDisjoint(leftBase, rightBase) //tc


    val name = leftBase.getName() + "_disjoint_" + rightBase.getName()
    val newComponent = createUMLComponent(name)
    createUMLDependency(newComponent, leftBase)
    createUMLDependency(newComponent, rightBase)

    //apply parallel stereotype on component
    val disjoint = Stereotypes.applyDisjointComponent(newComponent)
    disjoint.setLeft(left)
    disjoint.setRight(right)
    
    //provided
    val providedOps = calculateNewProvidedInterface(leftBase, rightBase)
    createProvided(providedOps, newComponent, name)
    //required
    val requiredOps = calculateNewRequiredInterface(leftBase, rightBase)
    createRequired(requiredOps, newComponent, name)
    extractAndCopyAttributes(leftBase, newComponent, this)
    extractAndCopyAttributes(rightBase, newComponent, this)

    disjoint
  }

  /**
   *   Create a plugging composition component
   */
  def addPluggingCompositionComponent(
      left: rCOSComponent,
      right: rCOSComponent) : rCOSComponent = {
    
    val leftBase = left.getBase_Component
    val rightBase = right.getBase_Component

    checkPlugging(leftBase, rightBase) //tc

    val name = leftBase.getName() + "_plugging_" + rightBase.getName()
    val newComponent = createUMLComponent(name)
    createUMLDependency(newComponent, leftBase)
    createUMLDependency(newComponent, rightBase)

    //apply parallel stereotype on component
    val plugging = Stereotypes.applyPluggingComponent(newComponent)
    plugging.setLeft(left)
    plugging.setRight(right)

    val providedOps = calculateNewProvidedInterface(leftBase, rightBase)

    val lst = new OperationSet(Buffer.empty[Operation])
    val requiredInterfaces = rightBase.getRequireds()
    requiredInterfaces.foreach(inf => lst ++= inf.getOwnedOperations())

    //required
    val requiredOps = calculateNewRequiredInterface(leftBase, rightBase)
    requiredOps diff providedOps
    createRequired(requiredOps, newComponent, name)

    providedOps diff lst
    createProvided(providedOps, newComponent, name)
    extractAndCopyAttributes(leftBase, newComponent, this)
    extractAndCopyAttributes(rightBase, newComponent, this)

    plugging
  }
  
  /**
   *  Create a renamed composition component 
   */
  def addRenamingComponent(
      left: rCOSComponent,
      operationStringTuples : java.util.List[StringTuple]) : rCOSComponent = {

    val leftBaseComponent = left.getBase_Component()
    val name = leftBaseComponent.getName + "_renamed"
    checkRenaming(leftBaseComponent, operationStringTuples) //tc
    val newComponent = createUMLComponent(name)
    val renamed = Stereotypes.applyRenamingComponent(newComponent)
    createUMLDependency(newComponent, left.getBase_Component())
    renamed.setLeft(left)
    
    //Create RenamingPairs - extract to method?
    val leftBaseComponentOps = leftBaseComponent.getOperations()
    val lst = new OperationSet(Buffer.empty[Operation])

    var oldOps = Set.empty[String]
    for	(operationStringTuple <- operationStringTuples) {
      oldOps = oldOps + operationStringTuple.oldName
    }

    for(op <- leftBaseComponentOps) {
      //is gonna be renamed
      if (oldOps.contains(op.getName)) {
        val st = getStringTuple(op)
    	val renamePair = JavaToScalaWrapperUtil.createRenamingPair()
    	renamePair.setOld(op)

    	val newop = extractAndCopyOperation(newComponent, op, st.newName, this)
    	lst += newop
    	renamePair.setNew(newop)
    	r.getContents().add(renamePair) //gotta save the resource
        renamed.getOps().add(renamePair); //add ref to 
      }

      //Not gonna be renamed - create op in new component
      else {
    	extractAndCopyOperation(newComponent, op, op.getName, this)
    	lst += op
      }
    }
  
    def getStringTuple(o: Operation) : StringTuple = {

      var st : StringTuple = null
      for (operationStringTuple <- operationStringTuples) { //Fetch string tuple
        if (operationStringTuple.containsOld(o.getName))
          st = operationStringTuple
      }
      return st
    }
    createProvided(lst, newComponent, name, addToClassifier = false)
    val lst2 = new OperationSet(Buffer.empty[Operation])
    val requiredInterfaces = left.getBase_Component().getRequireds()
    requiredInterfaces.foreach(inf => lst2 ++= inf.getOwnedOperations())
    createRequired(lst2, newComponent, name)
    extractAndCopyAttributes(leftBaseComponent, newComponent, this)

    renamed
  }

  /**
   *  Create a restricted/hiding composition component
   */
  def addHidingComponent(
      left: rCOSComponent,
      operations : java.util.Set[String]) : rCOSComponent = {
    
    val leftBaseComponent = left.getBase_Component()
    val name = leftBaseComponent.getName() + "_restricted"
    checkHiding(leftBaseComponent, operations)
    val newComponent = createUMLComponent(name)
    createUMLDependency(newComponent, left.getBase_Component())

    //apply hiding stereotype on component
    val hidden = Stereotypes.applyHidingComponent(newComponent)
    hidden.setLeft(left)
    
    val innerOperations = leftBaseComponent.getOwnedOperations()
    val lst = new OperationSet(Buffer.empty[Operation])
    for (innerOperation <- innerOperations) {
    	if (operations.contains(innerOperation.getName())) {
    	   hidden.getMethod().add(innerOperation) //Hide this, put in list of hidden operations
    	}
    	else { //Not going to hide this operation, create an identical op in the composite component
    	   lst += innerOperation
    	}
    }
   
    //fields are untouched by hiding - so copy them over to the composite component
    extractAndCopyAttributes(leftBaseComponent, newComponent, this)
    createProvided(lst, newComponent, name)

    val lst2 = new OperationSet(Buffer.empty[Operation])
    val requiredInterfaces = left.getBase_Component().getRequireds()
    requiredInterfaces.foreach(inf => lst2 ++= inf.getOwnedOperations())
    createRequired(lst2, newComponent, name)
    hidden
  }
  
  /**
   * Create an UML component
   */
  def createUMLComponent(name: String) : Component = {
      rcos.getComponentModel()
      .getBase_Package()
      .createPackagedElement(name, UMLPackage.eINSTANCE getComponent).asInstanceOf[Component]
  }

  /**
   *  Get a top-level rCOSComponent
   */
  def getrCOSComponent(name : String, context: rCOSComponent, token: CommonTree) : rCOSComponent = {
    val contxt = if (context != null) context.getBase_Component() else null
    val component = getComponent(name, contxt, token)
    Stereotypes.isRCOSComponent(component) //this will return the rCOSComponent as its stereotype
  }
 
  /**
   * Set the name of a component
   */
  def setName(rco : rCOSComponent, name : String) {
    val baseComponent = rco.getBase_Component()
    baseComponent.setName(name)
    
    //Since this method is only called when we change the name of 
    //a top level component, its safe to add it to the set of topLevelComponents. 
    topLevelComponentsRegister.add(name)
  }
  
  /**
   * Set the context of a component
   */
  def setContext(rco: rCOSComponent, context: rCOSComponent) {
    if (context != null && !topLevelComponentsRegister.contains(rco.getBase_Component().getName())) {
    	context.getBase_Component.getNestedClassifiers.add(rco.getBase_Component())
    }
  }
  
  /**
   * Create StringTuple with two strings. Used when renaming.
   */
  def createStringTuple(oldName: String, newName : String) : StringTuple = {
    new StringTuple(oldName, newName)
  }

  /**
   * Create UML dependency between client and supplier
   */
  def createUMLDependency(client: Component, supplier: Component) : Dependency = {

    val dependency =  UMLFactory.eINSTANCE.createDependency
	dependency.getClients.add(client)
	dependency.getSuppliers.add(supplier)
	client.getPackagedElements().add(dependency) //Who should own the dependency?
	dependency
  }
  
  def calculateNewProvidedInterface(left: Component, right: Component) : OperationSet = { 
    //every owned operation is provided since we don't support privat internal interfaces
    //We also validate this through typechecking, since a component cant provide methods that is not declared in
    //one of its provided interfaces
    
    //or.. calculate based of the interface realizations..
    val lst = new OperationSet(Buffer.empty[Operation])
    lst ++= left.getOwnedOperations
    lst ++= right.getOwnedOperations
    lst
  }

  def calculateNewRequiredInterface(left: Component, right: Component) : OperationSet = { 
    //Fetch all required ops from the required interfaces, put them together and return.
    val requiredInterfaces = left.getRequireds union right.getRequireds()

    val lst = new OperationSet(Buffer.empty[Operation])
    requiredInterfaces.foreach(inf => lst ++= inf.getOwnedOperations())
    lst
  }
  
  def addProvidedOperationsToClassifier(operations: OperationSet, classifier: Classifier) {
       for (providedOp <- operations) {
    	extractAndCopyOperation(classifier, providedOp, providedOp.getName, this)
    }
  }
  
  def createProvided(providedOps: OperationSet, newComponent: Component, name: String, addToClassifier: Boolean = true) {
    if (providedOps.size == 0)
      return 

    if (addToClassifier)
    	addProvidedOperationsToClassifier(providedOps, newComponent) //add to component
    
    val pIF = addInterface("_provided")
    newComponent.getNestedClassifiers().add(pIF)
    addProvidedOperationsToClassifier(providedOps, pIF) //add to interface
    addProvidedInterface(pIF, newComponent) //create UML interface realization
  }
  
  def createRequired(requiredOps: OperationSet, newComponent: Component, name: String) {
    if (requiredOps.size == 0)
      return 

	val rIF = addInterface("_required")
    newComponent.getNestedClassifiers().add(rIF)
    addProvidedOperationsToClassifier(requiredOps, rIF)
    addRequiredInterface(rIF, newComponent, null, 0) //Create UML interface usage
  }
}
