package no.uio.ifi.rcos.algebraic.typechecking

import org.scalatest.FlatSpec
import org.scalatest.matchers.ShouldMatchers
import no.uio.ifi.rcos.algebraic.spec.CreateModel
import no.uio.ifi.rcos.algebraic.utility.exceptions.NoSuchOperationException
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import edu.unu.iist.rcos.spec.antlr.parser.SemanticError
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedInterfaceException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRequiredInterfaceException
import edu.unu.iist.rcos.spec.antlr.RCOSModelCheckException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.IncompleteProvidedBlockException
import no.uio.ifi.rcos.algebraic.utility.exceptions.CommonProvidedAndRequiredOperationsException
import no.uio.ifi.rcos.algebraic.utility.exceptions.MissingMainException
import org.antlr.runtime.MismatchedTokenException
import no.uio.ifi.rcos.algebraic.utility.exceptions.ProvidedUndeclaredOperationException

class AtomicComponentsTests extends FlatSpec with ShouldMatchers {

 "A component" should "not contain duplicate class definitions" in {
   val prog = "component K { class A{} class A{}}"
   var createModel = new CreateModel
   intercept [RCOSModelCheckException] {
  createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "A component" should "not contain duplicate variable definitions" in {
   val prog = "component K { int i; int i;}"
   var createModel = new CreateModel
   intercept [RCOSModelCheckException] {
  createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "The set of provided and required methods" should "be disjoint" in {
   val prog = "interface A {public foo(;int number);} interface B {public foo(;int number);} component K { provided A {public foo(;int number) {number := 2 }} required B; }"
   var createModel = new CreateModel
   intercept [CommonProvidedAndRequiredOperationsException] {
  createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }
 
 "A component" should "must provide an implementation for every method declared in the provided interface(s)" in {
   val prog = "interface A {public foo(); public bar(); public foobar();} component K { provided A {public foobar() {}} }"
   var createModel = new CreateModel
   intercept [IncompleteProvidedBlockException] {
  createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "A component" should "not provide a method that is not declared in the interface " in { /* DOES THIS REALLY MATTER? */
   val prog = "interface A {public foo();} component K { provided A {public foo(){} public foobar() {} public bar(){}} }"
   var createModel = new CreateModel
   intercept [ProvidedUndeclaredOperationException] {
  createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "A component" should "not provide the same interface twice" in {
   val prog = "interface A {public foo();}  component K { provided A {public foo() {}} provided A {public foo() {}}}"
   var createModel = new CreateModel
   intercept [DuplicateProvidedInterfaceException] {
  createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "A component" should "not provide same method twice" in {
   val prog = "interface A {public foo();} interface B {public foo(); public bar();} component K { provided A {public foo() {}} provided B {public foo() {} public bar(){}}}"
   var createModel = new CreateModel
   intercept [DuplicateProvidedOperationException] {
     createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "A component" should "not require the same interface twice (not using enum)" in {
   val prog = "interface A {public foobar();} component K { required A; required A; }"
   var createModel = new CreateModel
   intercept [DuplicateRequiredInterfaceException] {
  createModel.createModelFromFileSpec("twiceprovided.uml", "twiceprovided", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 //TODO: Handle main method in processes? not top prio now - maybe later
 //"A process component" should "have a main method" in {
 //  val prog = "process P {provided A {public foo(){}}}"
 //  var createModel = new CreateModel
 //  intercept [MissingMainException] {
 //    createModel.createModelFromFileSpec("twiceprovided.uml", "twiceprovided", new MyRCOSModel.RCOSFromString(prog))
 //  }
 //}

 "A process component" should "not provide methods" in {
   val prog = "interface A {public foo(); } process P {provided A {public foo(){}}}"
   var createModel = new CreateModel
   intercept [MismatchedTokenException] {
     createModel.createModelFromFileSpec("twiceprovided.uml", "twiceprovided", new MyRCOSModel.RCOSFromString(prog))
   }
 }

 "Duplicate operations in an interface" should " not parse" in {  
   val interfaceString ="interface Foo { public foo(); public foo(); }" 
   var createModel = new CreateModel
   // XXX: Should be DuplicatedMethodName
   intercept [SemanticError] {
	createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))	  
   }
 }
  
  "An implemented operation in the interface" should " not give type checking error when the interface is provided by a component and this component does not provide an impl of the method (overriding)" in {  
	val interfaceString ="interface Foo { public foo(); public bar() { Var int k; k := 2}} component K { provided Foo { public foo(){Var int j}}}" 
    var createModel = new CreateModel
	createModel.createModelFromFileSpec("loli.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))	  
	createModel.model.save
  }
  
   "Duplicate variable in a component" should " not tc" in {  
   val interfaceString ="class A {} component Foo {  A a; A a; }" 
   var createModel = new CreateModel
   createModel.createModelFromFileSpec("test.uml", "test", new MyRCOSModel.RCOSFromString(interfaceString))	  
 }
}