package no.uio.ifi.rcos.algebraic.typechecking

import org.eclipse.uml2.uml.Component
import no.uio.ifi.rcos.algebraic.utility.OperationSet
import scala.collection.mutable.Buffer
import org.eclipse.uml2.uml.Operation
import scala.collection.JavaConversions._
import no.uio.ifi.rcos.algebraic.utility.exceptions.CommonProvidedAndRequiredOperationsException
import no.uio.ifi.rcos.algebraic.utility.exceptions.IncompleteProvidedBlockException
import no.uio.ifi.rcos.algebraic.utility.exceptions.ProvidedUndeclaredOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedOperationException
import no.uio.ifi.rcos.algebraic.extractors.CopyOperation
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel

/**
 * Atomic component type checker (but can be used on all types of components for general checks)
 * Every check here is made during parsing.
 * @author martinhb
 */

class AtomicTypeChecker extends CopyOperation{

  val TypecheckOn = true //set to true or false depending on your desire of type checking

  /**
  * Disjoint required and provided set
  */
  def tcDisjointRequiredProvidedSet(component: Component) {
    val reqs = new OperationSet(Buffer.empty[Operation])
    val requiredInterfaces = component.getRequireds()
    requiredInterfaces.foreach(inf => reqs ++= inf.getOwnedOperations())

    val provideds = component.getOwnedOperations()
    
    for (provided <- provideds) {
      if (reqs.contains(provided)) {
        throw new CommonProvidedAndRequiredOperationsException("Operation " + provided.getName() + " is both required and provided in component " + component.getName )
      }
    }
  }

  /**
  * Check that we have an implementation in the component for every method in the declared interface(s)
  */
  def tcImplForEveryMethod(component: Component, model: MyRCOSModel) {
    //Fetch ops in provided interfaces
    val provs = new OperationSet(Buffer.empty[Operation])
    val providedInterfaces = component.getProvideds()
    providedInterfaces.foreach(inf => provs ++= inf.getOwnedOperations())

    //Fetch ops in component
    val owned = new OperationSet(Buffer.empty[Operation])
    owned ++= component.getOwnedOperations()
    
    //Compare
    for (p <- provs) {
      if (!owned.contains(p)) { //p is not explicitly provided by the component, but it might be an operation with an impl in the interface
    	if (p.getOwnedRules().size() > 0)
    	  extractAndCopyOperation(component, p, p.getName, model)
    	else 
          throw new IncompleteProvidedBlockException("The component " + component.getName + " does not contain every operation in the provided interface(s)")
      }
    }
  }

  /**
  * Check for duplicated provided operations
  */
  def tcDuplicateProvideds(component: Component) {
    
    val provs = new OperationSet(Buffer.empty[Operation])
    val providedInterfaces = component.getProvideds()
    
    //duplicates will be left out in this set
    providedInterfaces.foreach(inf => provs ++= inf.getOwnedOperations())

    var realCount = 0
    for (provided <- providedInterfaces) {
      realCount += provided.getOwnedOperations().size()
    }

    //If we get different values it means we have some duplicates in the provided interfaces
    if (realCount != provs.size) {
      throw new DuplicateProvidedOperationException("Duplicate provided operations in component " + component.getName)
    }
  }

  def tcNotProvideUndeclaredMethod(component: Component) {
    
    val provs = new OperationSet(Buffer.empty[Operation])
    val providedInterfaces = component.getProvideds()
    providedInterfaces.foreach(inf => provs ++= inf.getOwnedOperations())

    val owned = new OperationSet(Buffer.empty[Operation])
    owned ++= component.getOwnedOperations()
    
    for (o <- owned) {
      if (!provs.contains(o)) {
    	throw new ProvidedUndeclaredOperationException("The component " + component.getName + " provides method " + o.getName + " which doesn't exist in any of the provided interface(s)")
      }
    }
  }

  /**
   * Type check a service component during parsing
   */
  def tcServiceComponent(component: Component, model: MyRCOSModel) {
    
	
    if (TypecheckOn) {
      //not provide duplicates
      tcDuplicateProvideds(component)
    
      //should not provide a method that is not declared in the interface
      tcNotProvideUndeclaredMethod(component)

      //provide an impl for every method declared in the provided interface
      tcImplForEveryMethod(component, model)
    
	  //Disjoint required and provided set
      tcDisjointRequiredProvidedSet(component)
    }
  }
  
  /**
   * Type check a process component during parsing
   */
  def tcProcessComponent(component: Component) {
    //TODO Must have a main method (?)
  }
}