package no.uio.ifi.rcos.algebraic.typechecking

import org.eclipse.uml2.uml.Component
import no.uio.ifi.rcos.algebraic.utility.OperationSet
import scala.collection.mutable.Buffer
import org.eclipse.uml2.uml.Operation
import scala.collection.JavaConversions._
import no.uio.ifi.rcos.algebraic.spec.StringTuple
import no.uio.ifi.rcos.algebraic.utility.exceptions.NoSuchOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRenameException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateVariableException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateVariableException
import no.uio.ifi.rcos.algebraic.utility.OperationComparator
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRequiredOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.OverlappingProvidedAndRequiredOperationException
import org.eclipse.uml2.uml.Type

/**
 * Class with methods for type checking component composition expressions
 */
trait CompositionExpressionTypeChecker extends OperationComparator {

  val TypecheckOn = true //set to true or false depending on your desire of type checking

  /**
   * Check for duplicate variables
   */
  def noCommonVariables(left: Component, right: Component) {
    //fetch Attributes and associations
    val leftAttrs = left.getAllAttributes
    val leftAssocs = left.getAssociations 

    val rightAttrs = right.getAllAttributes
    val rightAssocs = right.getAssociations

    var leftSet = Set.empty[String]
    var rightSet = Set.empty[String]

    for (la <- leftAttrs) {
      leftSet = leftSet + la.getName
    }

    for (ra <- rightAttrs) {
      rightSet = rightSet + ra.getName
    }
   

    for (la <- leftAssocs) {
      for (lla <- la.getNavigableOwnedEnds()) {
    	  leftSet = leftSet + lla.getName
      }
    }

    for (ra <- rightAssocs) {
      for (rra <- ra.getNavigableOwnedEnds()) {
    	  rightSet = rightSet + rra.getName
      }
    }
 
    //TODO Handle associations
    //for (ra <- rightAssocs) {
    //  //rightSet = rightSet + n.substring(n.lastIndexOf("_")+1)
    //}


    val intersection = leftSet intersect rightSet
    if (intersection.size > 0)
      throw new DuplicateVariableException("Duplicate variables " + intersection.toString + " in composition with component " + left.getName + " and " + right.getName)
  }


  /**
   * Check for common provided operations
   */
  def noCommonProvideds(left: Component, right: Component) {
	val leftProvideds = left.getOwnedOperations
	val rightProvideds = right.getOwnedOperations
	
	for (lp <- leftProvideds) {
	  for (rp <- rightProvideds) {
		 if (equalOperations(lp, rp)) {
		   throw new DuplicateProvidedOperationException("Common provided operation " + lp.getName + 
		       " found while checking a composition expression with components " + left.getName + " and " + right.getName)
		 }
	  }
	}
    
  }

  /**
   * Check for common required operations
   */
  def noCommonRequireds(left: Component, right: Component) {

    val leftReq = new OperationSet(Buffer.empty[Operation])
    left.getRequireds.foreach(inf => leftReq ++= inf.getOwnedOperations())

    val rightReq = new OperationSet(Buffer.empty[Operation])
    left.getRequireds.foreach(inf => rightReq ++= inf.getOwnedOperations())
    
    if (leftReq.intersect(rightReq).size > 0)
      throw new DuplicateRequiredOperationException("Duplicate required operation during disjoint composition " +
      		"between components " + left.getName + " (left) and " + right.getName + " (right)")
  }
  
  /**
   * Check for overlapping provided and required operation(s)
   */
  def noProvidedRequiredOverlap(left: Component, right: Component) {

    val leftReq = new OperationSet(Buffer.empty[Operation])
    left.getRequireds.foreach(inf => leftReq ++= inf.getOwnedOperations())

    val rightReq = new OperationSet(Buffer.empty[Operation])
    right.getRequireds.foreach(inf => rightReq ++= inf.getOwnedOperations())
    
    val leftProv = new OperationSet(Buffer.empty[Operation])
    leftProv ++= left.getOwnedOperations()

    val rightProv = new OperationSet(Buffer.empty[Operation])
    rightProv ++= right.getOwnedOperations()
  
    if (leftReq.intersect(rightProv).size > 0)
      throw new OverlappingProvidedAndRequiredOperationException("Overlapping between required operation(s) of " + left.getName +
      		" and provided operation(s) of " + right.getName)

    if (rightReq.intersect(leftProv).size > 0)
      throw new OverlappingProvidedAndRequiredOperationException("Overlapping between required operation(s) of " + right.getName +
      		" and provided operation(s) of " + left.getName)
   
  }

  /**
   * Type check parallel expression
   */
  def checkParallel(left: Component, right: Component) {
    if (!TypecheckOn) return
    noCommonVariables(left, right)
    noCommonProvideds(left, right)
  }

  /**
   * Type check disjoint expression
   */
  def checkDisjoint(left: Component, right: Component) {
    if (!TypecheckOn) return
    checkParallel(left, right)
    noCommonRequireds(left, right)
    noProvidedRequiredOverlap(left, right)
    
    //one component do not provide something the other requires
  }

  /**
   * Type check plugging expression
   */
  def checkPlugging(left: Component, right: Component) {
    if (!TypecheckOn) return
    checkParallel(left, right)
  }

  /**
   * Type check renaming expression
   */
  def checkRenaming(component: Component, operationStringTuples : java.util.List[StringTuple]) {
    
    if (!TypecheckOn) return
    //no duplicate renamings
    var oldOps = Set.empty[String]
    var newOps = Set.empty[String]
    val name = component.getName + "_renamed"

    //checking for duplicate names - we don't want any of that please
    //also checking if the "old" operations exist - if not throw exception
    for	(operationStringTuple <- operationStringTuples) {
      val o = operationStringTuple.oldName
      val n = operationStringTuple.newName

      if (component.getOperation(o, null, null) == null)
        throw new NoSuchOperationException(
            "Operation " + o + " in component " + component.getName+ " during a renaming composition")
      if (!oldOps.contains(o))
        oldOps = oldOps + o
      else 
         throw new DuplicateRenameException(
            "Duplicate operation name during renaming for component " + name + ": " + o)
      if (!newOps.contains(n))
        newOps = newOps + n
      else 
         throw new DuplicateRenameException(
            "Duplicate operation name during renaming for component " + name + ": " + n)
    }
    
    
  }
  
  /**
   * Type check hiding expression
   */
  def checkHiding(component: Component, operations: java.util.Set[String]) {

    if (!TypecheckOn) return
    val name = component.getName() + "_restricted"
    for	(operation <- operations) {
       val op = component.getOperation(operation, null, null)
       if (op == null)
           throw new NoSuchOperationException(
               "Operator " + operation + " does not exist" +
               	" (restriction on component  " + name + ")")
    }
  }
}