package no.uio.ifi.rcos.algebraic.typechecking

import org.scalatest.FlatSpec
import org.scalatest.matchers.ShouldMatchers
import no.uio.ifi.rcos.algebraic.spec.CreateModel
import edu.unu.iist.rcos.mm.util.TypingException
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import no.uio.ifi.rcos.algebraic.utility.exceptions.NoSuchOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRenameException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateVariableException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateProvidedOperationException
import no.uio.ifi.rcos.algebraic.utility.exceptions.OverlappingProvidedAndRequiredOperationException
import edu.unu.iist.rcos.spec.antlr.RCOSModelCheckException
import no.uio.ifi.rcos.algebraic.utility.exceptions.DuplicateRequiredOperationException

class ParallelTCtests extends FlatSpec with ShouldMatchers {

  /* Parallel */
  "Duplicate variable (primitive)" should " not type-check" in {  
	val interfaceString ="component C { int k; } component D { int k; } Z = C || D;" 
    var createModel = new CreateModel
    intercept [DuplicateVariableException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }

  "Duplicate variable (class)" should " not type-check" in {  
	val interfaceString ="class E {} component C { E k; } component D { E k; } Z = C || D;" 
    var createModel = new CreateModel
    intercept [DuplicateVariableException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }

  "Duplicate class" should " not type-check" in {  
    //apparently there is only one class allowed no matter where it is declared
	val interfaceString ="component C { class E {} } component D { class E {} } Z = C || D;" 
    var createModel = new CreateModel
    intercept [RCOSModelCheckException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }

  "Duplicate provided operations" should " not type-check" in {  
	val interfaceString ="interface A{public foo(); } component C { provided A {public foo(){}} } component D { provided A {public foo(){}} } Z = C || D;" 
    var createModel = new CreateModel
    intercept [DuplicateProvidedOperationException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }
  
  /* Disjoint */
  "Duplicate required operations" should " not type-check" in {  
	val interfaceString ="interface A { public foo(); } component C { required A; } component D { required A; } Z = C o D;" 
    var createModel = new CreateModel
    intercept [DuplicateRequiredOperationException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }

  "(disjoint) Overlapping provided and required operation(s)" should " not type-check" in {  
	val interfaceString ="interface A { public foo(); } component C { provided A {public foo(){}}} component D { required A; } Z = C o D;" 
    var createModel = new CreateModel
    intercept [OverlappingProvidedAndRequiredOperationException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }
  
  /* Plugging */
  //No tests here - see parallel and restriction tests which plugging is a combination of
  
  /* Renaming */
  "The operations which are going to be renamed" should "exist" in {  
	val interfaceString ="interface Foo { public foo();} component C { provided Foo { public foo(){} }} D = C [foo <- newfoo, bar <- newbar];" 
    var createModel = new CreateModel
    intercept [NoSuchOperationException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }

  "Renaming" should "not rename an operation more than once" in {  
	val interfaceString ="interface Foo { public foo();} component C { provided Foo { public foo(){} }} D = C [foo <- newfoo, foo <- newfood];" 
    var createModel = new CreateModel
    intercept [DuplicateRenameException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }

  /* Hiding */
  "The operations which are going to be restricted" should "exist" in {  
	val interfaceString ="interface Foo { public foo();} component C { provided Foo { public foo(){} }} D = C \\ {bar};"  
    var createModel = new CreateModel
    intercept [NoSuchOperationException] {
		createModel.createModelFromFileSpec("add_operations.uml", "add_operations", new MyRCOSModel.RCOSFromString(interfaceString))
	}
  }
}