package edu.unu.iist.rcos.spec.antlr;

import java.util.*;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.*;
import org.eclipse.emf.edit.command.RemoveCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.DesignAbstraction;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.ExpAbstraction;

/**
 * This is the command wrapper for abstractions.
 * It has to be used in a StrictCompound, because not everything can be properly
 * wrapped in getters: we could get e.g. the design from an operation through a
 * command, but we also need to see (potentially) see many other model changes
 * so that the parsers can get the fields and attributes right.
 * @author cm
 * @author vs (EMF Commands)
 */
public class AbstractRCOSModel extends RCOSCommand<AbstractionParameters> {

	public AbstractRCOSModel(EditingDomain ed, List<Throwable> errs, final AbstractionParameters t) {
		super(ed,errs,t);
	}

	public AbstractRCOSModel(EditingDomain ed, final AbstractionParameters t) {
		super(ed,t);
	}

	public AbstractRCOSModel(){}

	private Command abstractMethod(Operation o, List<String> lfa, List<String> lpar) throws AbstractionException, WellformedException, VisitorException{
		assert(o != null);
		DesignOperation dop = Stereotypes.isDesignOperation(o);
		if (false && (dop == null)){ // TODO: review
			throw new AbstractionException("The operation " + o.getQualifiedName() + " is not stereotyped as an rCOS operation"); // WF
		} else {
			CompoundCommand cmp = new CompoundCommand();
			cmp.append(IdentityCommand.INSTANCE);
			try {
				DESIGN d = Parse.getDesign(o);
				if (d != null) {
					//			System.err.println("The design of "+ o.getLabel() + " before abstraction : " + DesignUtility.printDesign(d));
					DESIGN d2 = abstractDesign(d, lfa, lpar);
					//			System.err.println("The design "+ o.getLabel() + " after  abstraction : " + DesignUtility.printDesign(d2));
					// Remove all the skip;skip and true AND true
					d2 = simplify(d2,(Classifier) o.getOwner()); // TODO: inline
					cmp.append(new DesignsED(domain).update(o,designToString(o,d2)));
				}
			} catch (RecognitionException e) { // TODO: Recognition exceptions could would better be generated directly in Parse.getFoo?
				throw new AbstractionException("Syntax error in design of "+o.getQualifiedName()+": '"+Designs.iNSTANCE.get(o)+"'",e);
			}
			try {
				final EXP e = Parse.getGuard(o);
				if ( e != null) {
					final EXP e2 = abstractExp(e, lfa, lpar);
					// TODO: Simplify should be inside abstraction.
					final EXP e3 = simplifyExpression(e2);
					cmp.append(new GuardsED(domain).update(o,expToString(o,e3)));
				}
			} catch (RecognitionException e1) { // TODO: see above?
				throw new AbstractionException("Syntax error in guard of "+o.getQualifiedName()+": '"+Guards.iNSTANCE.get(o)+"'",e1);
			}
			return cmp.unwrap();
		}
	}

	/**
	 * Abstract unneeded things from a class.
	 * @param c Class to process.
	 * @param lfa List of variables to keep.
	 * TODO: What about any dependencies of fields? What about the superclass?
	 * TODO: What if there are entries in 't.getAttrs()' or 'lfa' that don't exist here?
	 * @throws AbstractionException 
	 * @author vs	  
	 */
	@Override
	public Command createCommand() {
		try {
			Classifier c = params.getClass_();
			if (c == null)
				throw new IllegalArgumentException("Abstraction parameters did not specify a class in "+params.getBase_Class().getQualifiedName());
			CompoundCommand cmp = new CompoundCommand("Apply abstraction on "+ c.getName());
			cmp.append(IdentityCommand.INSTANCE);
			// abstract all fields:
			for(Property p : c.getAttributes()) {
				boolean found = false;
				for (StringOrProperty st : params.getAttrs()) {
					found = st.equalsToProperty(p);
					if (found)
						break;
				}
				if (!found) {
					//				System.err.println("Scheduling '"+p.getName()+"' for deletion");
					cmp.append(RemoveCommand.create(domain, p));
				}
			}
			// create a copy for abstractMethodCall:
			List<String> lfa = new ArrayList<String>(params.getAttrs().size());
			SOPNameSwitch sw = new SOPNameSwitch();
			for(StringOrProperty sp : params.getAttrs()) {
				String k = sw.doSwitch(sp);
				assert (k != null);
				assert (!k.equals(""));
				lfa.add(k);
			}

			List<String> lpar = new ArrayList<String>(params.getParams().size());
			for (Parameter p : params.getParams())
				lpar.add(p.getLabel());

			// abstract all methods
			for (Operation op : c.getOperations()) {
				cmp.append(abstractMethod(op,lfa, lpar));
				// If we don't keep ALL the method parameters, we remove all the parameters apart from those we explicitly want to keep
				if (! params.isKeepParameters()) {
					for (Parameter p : op.getOwnedParameters()) {
						if (! params.getParams().contains(p))
							cmp.append(RemoveCommand.create(domain, op, UMLPackage.eINSTANCE.getParameter(), p));
					}
				}
			}
			return cmp.unwrap();
		} catch (Exception e) {
			return recordError(e);
		}
	}

/**
 * @author Charles Morisset
 * @param e an expression
 * @param lfa a list of identifiers for class attributes
 * @param lpar a list of identifiers for method parameters
 * @return a copy of e if e contains some references to elements in lfa, true otherwise
 * @throws VisitorException 
 */
static EXP abstractExp(EXP e, List<String> lfa, List<String> lpar) throws VisitorException {
	return new ExpAbstraction(lfa, null, lpar).run(e);
}

/**
 * @author Charles Morisset
 * @param d a design
 * @param lfa a list of identifiers for class attributes
 * @param lpar a list of identifiers for method parameters
 * @return  a copy of d with only the sub-elements containing references to elements in lfa, lvar and lpar
 * @throws VisitorException 
 */
static DESIGN abstractDesign(DESIGN d, List<String> lfa, List<String> lpar) throws VisitorException {
	return new DesignAbstraction(lfa, null, lpar).run(d);
}

public static class AbstractionException extends Exception {
	private static final long serialVersionUID = 1L;

	AbstractionException(String msg) { super(msg); }

	public AbstractionException(Throwable t) {
		super(t);
	}

	public AbstractionException(String m, Throwable t) {
		super(m,t);
	}
}
}
