package edu.unu.iist.rcos.spec.antlr;

import java.util.*;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CharStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Parameter;

import edu.unu.iist.rcos.mm.util.Designs;
import edu.unu.iist.rcos.mm.util.Guards;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.Stereotypes;
import edu.unu.iist.rcos.spec.antlr.Vardecls.Vardecl;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.COMMAND;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.SyntaxTree_ClassModelFactory;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSLexer;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSTreeToUML;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser.singleDesign_return;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser.singleExpression_return;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.RefinementException;

public class Parse {

	private static rCOSParser parse(CharStream cs) {
		rCOSLexer lex = new rCOSLexer(cs);
		CommonTokenStream tokens = new CommonTokenStream(lex);		
		rCOSParser parser = new rCOSParser(tokens);
		return parser;
	}	

	static CommonTree getCommonTree(CharStream cs) throws RecognitionException {
		rCOSParser parser = parse(cs);
		return (CommonTree) parser.rCOSSpec().getTree(); 
	}

	static CommonTree getCommonTreeOfExpFromString(String str) throws RecognitionException{
		CharStream cs = new ANTLRStringStream(str);
		rCOSParser parser = parse(cs);
		return (CommonTree) parser.singleExpression().getTree();
	}
	
	private static CommonTree getCommonTreeOfTypeFromString(String str) throws RecognitionException{
		CharStream cs = new ANTLRStringStream(str);
		rCOSParser parser = parse(cs);
		rCOSParser.type_return rret = parser.type();
		return (CommonTree) rret.getTree();
	}

	static CommonTree getCommonTreeOfDesignFromString(String str) throws RecognitionException{
		CharStream cs = new ANTLRStringStream(str);
		rCOSParser parser = parse(cs);
		singleDesign_return rret = parser.singleDesign();
		return (CommonTree) rret.getTree();
	}
	
	/**
	 * CAVEAT EMPTOR: if your context is an operation, you have to pass it as a context to get the parameters right!
	 * @see #getGuard(Operation)
	 * @see #getGuard(String, Operation)
	 */
	public static EXP getExpFromString(String str) throws RecognitionException{	
		CommonTree ret = getCommonTreeOfExpFromString(str);
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(ret);
		rCOSTreeToUML walker2 = new rCOSTreeToUML(nodes);
		return walker2.expression();
	}
	
	public static DESIGN getDesign(final Operation op) throws RecognitionException, WellformedException {
		assert(op != null);
		String cs = Designs.iNSTANCE.get(op);
		if (cs == null) return null;
		else return getDesign(cs, op);
	}
	
	/**
	 * Specialized version of getDesign(Operation op) if you're either already holding the body in your hands (micro-optimization)
	 * or if you want to check a design in a specific context.
	 * @param cs
	 * @param op
	 * @return
	 * @throws RecognitionException
	 * @throws WellformedException 
	 * @see {@link #getDesign(Operation)}
	 */
	public static DESIGN getDesign(final String cs, final Operation op) throws RecognitionException, WellformedException {
		assert(op != null);
		assert cs != null : op.getQualifiedName();
		rCOSParser parser = parse(new ANTLRStringStream(cs));
		putOpSymbols(op, parser);
		singleDesign_return rret = parser.singleDesign();
		CommonTree ret = (CommonTree) rret.getTree();
		return ctToUML(ret, new Simplify() {public DESIGN simplify(DESIGN d) throws RefinementException {return DesignUtility.simplify(d, (Classifier) op.getOwner());}}, true);
	}

	public static DESIGN getDesignChoke(final String cs, final Operation op) throws RecognitionException, WellformedException {
		assert(op != null);
		assert cs != null : op.getQualifiedName();
		rCOSParser parser = parse(new ANTLRStringStream(cs));
		putOpSymbols(op, parser);
		singleDesign_return rret = parser.singleDesign();
		CommonTree ret = (CommonTree) rret.getTree();
		return ctToUML(ret, new Simplify() {public DESIGN simplify(DESIGN d) throws RefinementException {return DesignUtility.simplify(d, (Classifier) op.getOwner());}}, false);
	}

	private static void putOpSymbols(final Operation op, rCOSParser parser)
			throws WellformedException, RecognitionException {
		// We set the parameters of the operation
		// TODO: don't allocate so many new ones.
		TypeHandler th = new TypeHandler(Stereotypes.isRCOSModel(Stereotypes.getRootModel(op)));
		
		for (Parameter para : op.getOwnedParameters()){
			String name = para.getName();
			if (para.getType() == null)
				throw new WellformedException("Parameter doesn't have a type: "+para.getQualifiedName());
			String type = TypeHandler.printTYPE(th.createTYPE(para.getType()));
			if (para.getUpper() == -1 && para.getLower() == 0)
				type = "Set<" + type + ">";
			parser.parSymbols.put(name, getCommonTreeOfTypeFromString(type));
		}
	}

	interface Simplify {
		DESIGN simplify(DESIGN d) throws VisitorException;
	}

	private static DESIGN ctToUML(CommonTree ret, Simplify s, final boolean choke) throws RecognitionException {
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(ret);
		rCOSTreeToUML walker2 = new rCOSTreeToUML(nodes) {
			@Override public COMMAND createChoke() {
				if (choke) return super.createChoke(); else return SyntaxTree_ClassModelFactory.eINSTANCE.createCHOKE();
			}
			};
		try {
			/* We normalize the AST, otherwise you will encounter some surprises with
			 * the SEQs arising from VARDECLS (they generate (SEQ l r)).
			 * TODO: Probably simplification does too much, so we really just want normalization here.		
			 */
			return s.simplify(walker2.seq_design());
		} catch (VisitorException e) {
			throw new RuntimeException(e); // This should not happen
		}
	}

	/**
	 * @param str
	 * @return
	 * @throws RecognitionException
	 * @deprecated WARNING: Usually you need the context (Operation) to get "this." right!
	 */
	public static DESIGN getDesignFromString(final String str) throws RecognitionException{
		CommonTree ret = getCommonTreeOfDesignFromString(str);	
		return ctToUML(ret, new Simplify() { public DESIGN simplify(DESIGN d) {return d;}}, true);
	}

	public static EXP getGuard(final Operation op) throws RecognitionException, WellformedException {
		assert(op != null);
		String cs = Guards.iNSTANCE.get(op);
		if (cs == null) return null;
		else return getGuard(cs, op);
	}

	/*
	 * Note that this will modify the passed vardecls, so bring your own copy
	 * if you want to reuse it!
	 */
	public static EXP getExp(final String cs, Vardecls ctx) throws RecognitionException, WellformedException {
		assert cs != null;
		assert ctx != null;
		rCOSParser parser = parse(new ANTLRStringStream(cs));
		/* We run in the order of the definition. That way, we simply
		 * overwrite older bindings instead of checking if the current
		 * item is shadowed.
		 */
		setupParserCtx(parser,ctx);
		singleExpression_return rret = parser.singleExpression();
		CommonTree ret = (CommonTree) rret.getTree();
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(ret);
		rCOSTreeToUML walker = new rCOSTreeToUML(nodes);
		return walker.expression();
	}

	// TODO: rename to getExp?
	public static EXP getGuard(final String cs, final Operation op) throws RecognitionException, WellformedException {
		assert(cs != null);
		assert(op != null);
		if ("".equals(cs)) {
			// TODO: Make validation warning?
			System.err.println("Guard of operation "+op.getQualifiedName()+" is the empty string");
			return null;
		}
		rCOSParser parser = parse(new ANTLRStringStream(cs));
		putOpSymbols(op,parser);
		singleExpression_return rret = parser.singleExpression();
		CommonTree ret = (CommonTree) rret.getTree();
		
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(ret);
		rCOSTreeToUML walker2 = new rCOSTreeToUML(nodes);
		return walker2.expression();
	}

	/**
	 * Primes parser identifier context from existing vardecls.
	 * Note that this will modify the passed vardecls, so bring your own copy
	 * if you want to reuse it!
	 */
	public static DESIGN getDesign(String cs, Vardecls ctx) throws RecognitionException {
		rCOSParser parser = parse(new ANTLRStringStream(cs));
		/* We run in the order of the definition. That way, we simply
		 * overwrite older bindings instead of checking if the current
		 * item is shadowed.
		 */
		setupParserCtx(parser,ctx);
		singleDesign_return rret = parser.singleDesign();
		CommonTree ret = (CommonTree) rret.getTree();
		return ctToUML(ret, new Simplify() {public DESIGN simplify(DESIGN d) throws RefinementException {return d;}}, true);
	}

	private static void setupParserCtx(rCOSParser parser, Vardecls ctx) throws RecognitionException {
		parser.mkLocalScope();
		parser.mkQuanScope();
		final Iterator<List<Vardecl>> i = ctx.vardecls.descendingIterator();
		while (i.hasNext()) {
			for (Vardecl v : i.next()) {
				final CommonTree cType = getCommonTreeOfTypeFromString(v.getValue().getName());  // XXX getName() correct?
				switch (v.pt) {
				case PAR: parser.parSymbols.put(v.getKey(), cType);
				case LOCAL: parser.putLocal(v.getKey(), cType);
				case QUAN: parser.putQuan(v.getKey(), cType);
				}
			}
		}
	}
}