package edu.unu.iist.rcos.spec.antlr;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.WrappedException;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;
import edu.unu.iist.rcos.spec.transformer.*;

/**
 * @author charles
 * @author stolz
 * TODO: there's still some overlap between RecognitionException and TypingException
 *
 */
public class RCOSModelUtility {

	public abstract class TranslatableResult {
		public boolean isOK() { return this instanceof TranslatableOK; }
	};
	public class TranslatableOK extends TranslatableResult{};
	public class TranslatableFail extends TranslatableResult {
		final DESIGN d;
		final Operation op;
		public TranslatableFail(Operation op, DESIGN d) { this.d = d; this.op = op; }
		@Override public String toString() { return DesignUtility.designToString(op,d); }
	};

	public TranslatableResult isTranslatable(RCOSModel mod) throws RCOSModelCheckException, WellformedException {
		ClassModel cm = mod.getClassModel();
		for (Element e : cm.getBase_Package().getOwnedElements()){ 
			if (e instanceof Class){
				Class c = (Class) e;
				// TODO: There must be an idiom for this test, or maybe use BasicDiagnostic?
				TranslatableResult r;
				if ((r = isTranslatable(c)) instanceof TranslatableFail)
					return r;
			}
		}
		
		// Some classes may be inside components:
		for (Element e : mod.getComponentModel().getBase_Package().getOwnedElements()) {
			if (e instanceof Component) {
				TranslatableResult r = isTranslatable((Component)e);
				if (r instanceof TranslatableFail)
					return r;
			}
		}
		return new TranslatableOK();
	}

	private TranslatableResult isTranslatable(Component co) throws WellformedException,
			RCOSModelCheckException {
		for (Element c : co.getOwnedElements()) {
			TranslatableResult r = null;
			if (c instanceof Component)
				r = isTranslatable((Component)c);
			if (r instanceof TranslatableFail)
				return r;

			if (c instanceof Class) {
				r = isTranslatable((Class)c);
			}
			if (r instanceof TranslatableFail)
				return r;
		}
		return new TranslatableOK();
	}

	public TranslatableResult isTranslatable(Class c) throws WellformedException,
			RCOSModelCheckException {
		for (Operation op : c.getOperations()) {
			try {
				DESIGN d = Parse.getDesign(op);
				// We don't permit empty designs here:
				if (d!= null
						&& ! DesignUtility.isDesignTranslatable(d))
					return new TranslatableFail(op,d);
			} catch (RecognitionException e1) { 
				throw new RCOSModelCheckException("In the class " + c.getQualifiedName()
						+ ", the design of the operation " + op.getLabel() + " is not a valid rCOS expression:\n" + e1.getMessage(),e1);
			}
		}
		return new TranslatableOK();
	}

	/**
	 * Transforms the design of every operation of the class using the design transformer.
	 * Runs the analysis 'dt' inside the command at execution time, so you should apply a CommandWrapper.
	 * @author Charles Morisset
	 * @param e a Class of a rCOS Model
	 * @param dt a Design transformer
	 * @throws RCOSModelCheckException
	 * @throws WellformedException 
	 */	
	public static Command transform(EditingDomain ed, Class e, DesignTransformer<?> dt) throws RCOSModelCheckException, WellformedException {
		CompoundCommand cmp = new StrictCompoundCommand("operations") {{ isPessimistic = true; }};
		cmp.append(IdentityCommand.INSTANCE);

		for (Operation op : e.getOperations())
			cmp.append(transform(ed, e, dt, op));
		return cmp.unwrap();
	}

	/**
	 * Runs the analysis 'dt' inside the command at execution time, so you should apply a CommandWrapper. Currently the only consumer is the Expert Pattern,
	 * but this may change in the future.
	 * TODO: what about guards and invariants?
	 * @author cm (logic)
	 * @author vs (Command, Designs)
	 */
	public static Command transform(final EditingDomain ed, final Classifier e, final DesignTransformer<?> dt, final Operation op) {
		return new CommandWrapper("class " + op.getQualifiedName()) {
			@Override public Command createCommand() {
				try {
					final DESIGN d = Parse.getDesign(op);
					if (d != null){
						CompoundCommand cmp = new StrictCompoundCommand("inside "+e.getQualifiedName()){{ isPessimistic = true;}};

						Command req = dt.getRequired(d, e);
						assert req != null;
						cmp.append(req);

						cmp.append(new CommandWrapper() {
							@Override public Command createCommand() {
								try {
									final DESIGN d2 = dt.transform(d,e); 
									return new DesignsED(ed).update(op, DesignUtility.designToString(op,d2));
								} catch (VisitorException e) {
									throw new WrappedException(e);
								}
							}});
						return cmp;

					} else {
						// No design attached to the operation -> nothing to do
						return new IdentityCommand();
					}
				} catch (RecognitionException re) {
					throw new WrappedException("In the class " + e.getQualifiedName()
							+ ", the design of the operation " + op.getLabel() + " is not a valid rCOS expression:\n" + re.getMessage(),re);
				} catch (WellformedException we) { throw new WrappedException(we); }
				catch (VisitorException ve) { throw new WrappedException(ve);}
			}};
	}

	/**
	 * Transforms every class of the Class Model of mod using the design transformer.
	 * @author Charles Morisset
	 * @param mod a  rCOS Model
	 * @param dt a Design transformer
	 * @throws RCOSModelCheckException
	 * @throws WellformedException 
	 * @deprecated Only used in test-cases.
	 */
	public static Command transformAllClasses(EditingDomain domain, RCOSModel mod, DesignTransformer<?> dt) throws RCOSModelCheckException, WellformedException{
		ClassModel cm = mod.getClassModel();
		CompoundCommand cmp = new StrictCompoundCommand() {{ isPessimistic = true; }};
		for (Element e : cm.getBase_Package().getOwnedElements()){ // TODO: deferred execution?
			if (e instanceof Class)
				cmp.append(transform(domain,(Class) e, dt));
		}
		return cmp;
	}
}
