package edu.unu.iist.rcos.spec.antlr;

import java.util.ArrayList;
import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.util.*;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.ModelValidation;
import edu.unu.iist.rcos.mm.ocl.StateMachine_OCL;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.mm.util.statemachine.STMHelper;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.visitor.check.*;
import edu.unu.iist.rcos.spec.tc.RuleTCMembers;
import edu.unu.iist.rcos.spec.tc.RuleTCMembersTH;
import edu.unu.iist.rcos.spec.tc.RuleTransitionGuard;
import edu.unu.iist.rcos.spec.tc.RuleTC.TypingWarningException;
import edu.unu.iist.rcos.spec.tc.RuleTCMembers.*;

/**
 * @author charles, vs
 * Sanity checks. Never return false but throw an exception instead.
 * Adapted to use the Diagnostics framework.
 * 
 * TODO: for the rtc.check-methods, the exceptions need to be properly dissected to find out if it is
 * a structural, syntax, or type checking error.
 */
public class RCOSModelValidationNoUML extends ModelValidation {

	private RuleTCMembers rtc;
	TypeHandler th;

	/**
	 * Entry point for topcased-based validation.
	 */
	public RCOSModelValidationNoUML() {
		this("rCOS");
	};
	
	protected RCOSModelValidationNoUML(String provider) {
		super(provider);
	};
	
	/**
	 * @param diag Diagnostic, or null.
	 * Entry point for manual calls.
	 */
	public RCOSModelValidationNoUML(DiagnosticChain diag) {
		this();
		this.diag = diag;
	}	

	public boolean validate(RCOSModel mod) throws RCOSModelCheckException, WellformedException {
		boolean valid = true;
		//COMMENTED OUT FOR algebraic project
		//th = new TypeHandler(mod);
		//rtc = new RuleTCMembersTH(th);
		//valid &= validateClasses(mod);
		//valid &= validateComponents(mod);

		return valid;
	}

	public static class InvalidComponentComposition extends WellformedException {
		private static final long serialVersionUID = 1L;
		public InvalidComponentComposition(String msg) {
			super(msg);
		}
	}

	protected boolean validateComponents(RCOSModel mod) throws RCOSModelCheckException, WellformedException {
		ComponentModel cm = mod.getComponentModel();
		if (cm == null)
			return false;
		for (Element e : cm.getBase_Package().allOwnedElements()){
			if ((e instanceof Class) && !(e instanceof Interaction)) {
				validate((Class)e); // do we want to have classes here?
			}
			if (e instanceof Component) {
				Component co = (Component)e;
				for (InterfaceRealization ir : co.getInterfaceRealizations()) {
					Interface i = ir.getContract();
					if (i != null) {
						ContractInterface ci = Stereotypes.isContractInterface(i);
						if (ci != null) { // Should be WF?
							final Class controllerClass = Utility.getControllerClass(co, ci);
							if (controllerClass != null)
								checkClassImplementsCtr(ci, controllerClass);
							/* If this CI is involved in a composition or delegation, the component
							 * must be hierarchical, i.e. owned by another one.
							 * RCOSTOOL-88
							 */
							for (Relationship r : i.getRelationships(UMLPackage.eINSTANCE.getDependency())) {
								Delegation d = Stereotypes.isDelegation((Dependency) r);
								Composition c = Stereotypes.isComposition((Dependency) r);
								if (d != null || c != null) {
									if (!(co.getOwner() instanceof Component)) {
										WellformedException we = new InvalidComponentComposition("Component composition/delegation only allowed within hierarchical components ("+co.getQualifiedName()+")!");
										if (diag == null)
											throw we;
										else
											dmsg(we,co);
							}
								}
							}
						}
					}
				}
				for (NamedElement d : co.getOwnedMembers()) {
					if (d instanceof Dependency) {
						Delegation del;
						if ((del = Stereotypes.isDelegation((Dependency) d)) != null) {
							try {
								ContractInterface ci = del.getClient();
								boolean isProvided = ci.getBase_Interface().getRelationships(UMLPackage.eINSTANCE.getInterfaceRealization()).size() > 0; //WF
								if (isProvided) {
									assert ci.getBase_Interface().getRelationships(UMLPackage.eINSTANCE.getUsage()).size() == 0;
									String result = del.getQualifiedProvidedInterfaceName();
									assert result != null;
								} else {
									assert ci.getBase_Interface().getRelationships(UMLPackage.eINSTANCE.getUsage()).size() > 0;
									/* TODO: Why is this not symmetric?! [vs] thinks this was just added to exercise getQualified..() a bit more. */
								}
							} catch (WellformedException we) {
								if (diag == null)
									throw we;
								else
									dmsg(we, d);
							}
						}
					}
				}
			} else if (e instanceof Interface){
				ContractInterface i = Stereotypes.isContractInterfaceNull((Interface) e);
				if (i != null) {
					validateContractInterface(i);
				} else
				{ /* NADA */ } 
			} else if (e instanceof Package){
				Protocol protocol = Stereotypes.isProtocol((Package) e);
				if(protocol != null){
					List<ContractInterface> cis = STMHelper.getContractInterface(protocol);
					if (cis.size() == 0)
						showException(ErrorType.WF, "Can't typecheck protocol: no owning contract defined.", (Package)e);
					else if (cis.size() > 1)
						showException(ErrorType.WF, "Can't typecheck protocol: too many contracts using protocol ("+cis+").", (Package)e);
					else {
						Collaboration sequencediagram = protocol.getSequencediagram();
						if (sequencediagram != null && sequencediagram.getOwnedBehaviors().size() > 0){
							Behavior behavior = sequencediagram.getOwnedBehaviors().get(0);
							if(behavior instanceof Interaction) {
								final RuleClassInvariant chk = getRtc().new RuleClassInvariant();
								for(InteractionFragment fragment : ((Interaction) behavior).getFragments())
									validateInteraction(cis.get(0).getBase_Interface(), chk,fragment);
							}
						};
					}
					validateStateMachine(protocol.getStatemachine());
				}
			}
		}
		return true;
	}

	private void dmsg(WellformedException we, NamedElement d) {
		dmsg(ErrorType.WF, we.getMessage(), we, d);
	}

	public static void checkImplements(ContractInterface ci, Class c) throws WellformedException {
		Interface i = ci.getBase_Interface();
    	for (Property p : i.getOwnedAttributes()){
    		if (c.getOwnedAttribute(p.getName(), p.getType()) == null) {
				// TODO: Nope. throw new RCOSModelCheckException("Attribute "+p.getName()+" not implemented");
			}
    	}
    	for (Operation o: i.getOwnedOperations()) {
    		if (Utility.implementsOperation(c, o) == null)
    			throw new WellformedException("Operation "+o.getName()+" not implemented");
    	}
	}
	
	private boolean checkClassImplementsCtr(ContractInterface ci, Class c) throws RCOSModelCheckException, WellformedException {
		if (ci == null) throw new IllegalArgumentException();
		if (c == null) throw new IllegalArgumentException();
		final String msg = "Class "+c.getQualifiedName()+" does not implement contract interface "+ci.getBase_Interface().getQualifiedName();
		try {
			checkImplements(ci, c);
		} catch (WellformedException e) {
			if (diag == null)
				throw e;
			else dmsg(ErrorType.WF, msg, e, c);
		}
		return true;
	}

	private void validateContractInterface(ContractInterface i) throws RCOSModelCheckException, WellformedException {
		// Stub. TODO
		if (i.getInterface() == null)
			if (diag == null)
				throw new WellformedException("No syntactic interface on "+i.getBase_Interface().getQualifiedName());
			else
				dmsg(ErrorType.WF, "No syntactic interface", i.getBase_Interface());

		validate(i.getBase_Interface());
	}

	private void validateStateMachine(StateMachine statemachine) throws WellformedException, RCOSModelCheckException{
		if(statemachine == null)
			return;
		if (!StateMachine_OCL.SM_EXACTLY_ONE_REGION_R.check(statemachine))
			throw new WellformedException(StateMachine_OCL.SM_EXACTLY_ONE_REGION_R.getDescription());
		EList<Region> regions = statemachine.getRegions();
		
		RuleTransitionGuard rule = new RuleTransitionGuard(null);
		for(Transition transition : regions.get(0).getTransitions()){
			try {
				rule.check(transition);
			} catch (RecognitionException e) {
				showException(ErrorType.SYNTAX, e.getMessage() == null ? e.toString() : e.getMessage(), transition);
			} catch (WellformedException e) {
				showException(ErrorType.WF, e.getMessage() == null ? e.toString() : e.getMessage(),transition);
			}
		}
	}

	private void showException(ErrorType type, String message, NamedElement element) throws WellformedException, RCOSModelCheckException{
		assert message != null;
		if (diag == null)
			if (type == ErrorType.SYNTAX)
				throw new RCOSModelCheckException(message);
			else
				throw new WellformedException(message);
		else
			dmsg(type, message.replace('\n', ' '), element);
	}

	private void validateInteraction(Classifier ctx, RuleClassInvariant chk, InteractionFragment fragment) throws WellformedException, RCOSModelCheckException {
		if(fragment instanceof CombinedFragment){
			final CombinedFragment cf = (CombinedFragment)fragment;
			for(InteractionOperand operand : cf.getOperands())
				validateInteraction(ctx, chk, operand);
		}else if(fragment instanceof InteractionOperand){
			InteractionOperand operand = (InteractionOperand) fragment;
			InteractionConstraint constraint = operand.getGuard();
			if(constraint != null){
				try {
					final ValueSpecification spec = constraint.getSpecification();
					if (spec instanceof LiteralSpecification) {
						LiteralSpecification s = (LiteralSpecification) spec; // WF cast
						if (s == null)
							throw new WellformedException("Incomplete specification");
						String g = s.stringValue();		
						if (g == null)
							throw new WellformedException("Incomplete specification");

						if (((CombinedFragment)operand.getOwner()).getInteractionOperator().equals(InteractionOperatorKind.ALT_LITERAL))
							try {
								chk.check(ctx, g);
							} catch (RecognitionException e) {
								showException(ErrorType.SYNTAX, e.getMessage(),s);
							}
					} else
						if (diag != null)
							dmsgW("Not typechecked (yet)", spec);
				} catch (WellformedException we) {
					showException(ErrorType.WF, we.getMessage(), constraint);
				}
			} else { /* NADA */ }
			for(InteractionFragment inner : operand.getFragments())
				validateInteraction(ctx, chk, inner);
		}
	}

	protected boolean validateClasses(RCOSModel mod) throws RCOSModelCheckException, WellformedException {
		ClassModel cm = mod.getClassModel();
		if (cm == null)
			return false;
		for (Element e : cm.getBase_Package().allOwnedElements()){
			if (e instanceof Class){
				validate((Class)e);
			}
		}
		return true;
	}

	/**
	 * Run all tests on a class
	 * @param e
	 * @return
	 * @throws RCOSModelCheckException
	 * @throws WellformedException 
	 * @throws RCOSModelCheckException 
	 */
	protected boolean validate(Classifier e) throws RCOSModelCheckException, WellformedException {
		if (th == null)
			throw new IllegalStateException();
		if (e instanceof Class && ! (e instanceof Component))
			checkImplements((Class) e);
		if ((e.getName() == null) || e.getName().equals("")) {
			final String prefix = "Unnamed class in package "; 
			if (diag == null)
				throw new WellformedException(prefix+e.getNearestPackage().getQualifiedName()+", "+e);
			else
				dmsg(ErrorType.WF,prefix,e);				
		}

		DiagnosticChain d = diag == null ? new BasicDiagnostic() : diag;
		if (! e.validateNoCyclesInGeneralization(d, null))
			if (diag == null) 
				throw new WellformedException("Generalization cycle in "+e.getNearestPackage().getQualifiedName()+", "+e);
			else
				dmsg(ErrorType.WF, "Generalization cycle", e);

		for (Operation op : e.getOperations()){
			if (op.getName() == null || "".equals(op.getName()))
				if (diag == null)
					throw new WellformedException("Operation doesn't have a name: "+op);
				else
					dmsg(ErrorType.WF,"Operation doesn't have a name!", op);
			else {
				// Check if it's a constructor (starting with an uppercase letter) and if it corresponds to the classifier name
				if (Character.isUpperCase(op.getName().charAt(0)) & ! op.getName().equals(e.getName()))
					if (diag == null)
						throw new WellformedException("In the classifier " + e.getQualifiedName() + ", the constructor " + op.getLabel() + " should have the classifier name");
					else
						dmsg(ErrorType.WF,"Bad constructor name ",op);
			}
			String prefix = null;
			String blob = null;
			try {
				{
					prefix = "Syntax error in design";
					RuleOperationDesign rod = getRtc().new RuleOperationDesign();
					blob = rod.rf.get(op);
					if (blob != null)
						if (blob.equals(""))
							dmsgW("Empty design",op);
						else
							rod.check(op);
				}
				{
					prefix = "Syntax error in guard";
					RuleOperationGuard rog = getRtc().new RuleOperationGuard();
					blob = rog.rf.get(op);
					if (blob != null)
						if (blob.equals(""))
							dmsgW("Empty guard",op);
						else
							rog.check(op);
				}
			} catch (RecognitionException e1) {
				assert blob != null; // otherwise we wouldn't be here.
				final String msg = prefix+" "+op.getQualifiedName()+":"+blob;
				if (diag == null)
					throw new WellformedException(msg,e1);
				else
					dmsg(ErrorType.SYNTAX,msg,e1,op);
			} catch (TypingWarningException ex) {
				dmsgW(ex.getMessage(), e);
			} catch (TypingException ex) {
				if (diag == null)
					throw ex;
				else
					dmsg(ErrorType.TYPING,"Type error",ex,op);
			} catch (WellformedException ex) {
				if (diag == null)
					throw ex;
				else
					dmsg(ErrorType.WF,"Not wellformed",ex,op);				
			}

			// check parameters and their direction
			{
				// UML superstructure 7.3.41: at most one return parameter, so we don't need to check it.
				// TODO: but somehow [vs] can't get this rule to fire, no even in the regular UML validator!

				final String oneMsgErr = "More than one parameter with direction 'return'";
				boolean oneReturnSeen = false;
				for (Parameter pa : op.getOwnedParameters()) {
					if (pa.getDirection().getValue() == ParameterDirectionKind.INOUT)
						dmsgW("Parameter with direction 'inout' will be treated as 'out'", pa);

					if (pa.getDirection().getValue() == ParameterDirectionKind.RETURN) {
						if (oneReturnSeen) {
							if (diag == null)
								throw new WellformedException(oneMsgErr+" on "+pa.getQualifiedName());
							else
								dmsg(ErrorType.WF,oneMsgErr,pa);
						} else
							oneReturnSeen = true;
					}
				}
			}
		}
		for (Property p : e.getAllAttributes()) {
			// Check that a default value is well-typed.
			// It has to be the right type, and no primed variables
			// TODO: This should eventually be hooked up to the GUI as well.
			
			{ // validate the name: mustn't be a keyword
				if (FreshNameGenerator.isReservedIdentifier(p.getName()))
					if (diag == null)
						throw new WellformedException("Property "+p.getQualifiedName()+" has an illegal name");
					else
						dmsg(ErrorType.SYNTAX,"Illegal name",p);
			}
			String def = p.getDefault();
			if (def != null) {
				EXP exp = null;
				try {
					exp = Parse.getExpFromString(def);
				} catch (RecognitionException e1) {
					if (diag == null)
						throw new WellformedException(e1);
					else
						dmsg(ErrorType.SYNTAX,"Syntax error in",e1,p);
				}

				if (exp != null) {
					if (DesignUtility.hasPrimedVariables(exp))
						if (diag == null)
							throw new PrimedVariableForbiddenException("The default value expression '" + def + "' contains dashed variables");
						else
							dmsg(ErrorType.DASHED,"The default value expression '" + def + "' contains dashed variables",p);

					{
						Type t;
						try {
							t = th.getType(exp, e);
							assert t != null;

							Type tp = p.getType();
							if (tp == null)
								if (diag == null)
									throw new WellformedException(p.getQualifiedName() + " does not have a type!");
								else
									dmsg(ErrorType.WF,p.getQualifiedName() + " does not have a type!",p);

							if (!th.isSubType(t, tp))
								if (diag == null)
									throw new TypingException(p.getQualifiedName() + " is declared as type "+tp.getQualifiedName() + " but expression has type "+t.getQualifiedName());
								else
									dmsg(ErrorType.TYPING,p.getQualifiedName() + " is declared as type "+tp.getQualifiedName() + " but expression has type "+t.getQualifiedName(),p);
						} catch (GetTypeException e1) {
							if (diag == null)
								throw new WellformedException(e1);
							else
								dmsg(ErrorType.TYPING,"Typing error",e1,p);
						}
					}
				}
			}
		}
		checkAssocAttr(e);
		try {
			getRtc().new RuleClassInvariant().check(e);
		} catch (RecognitionException e1) {
			if (diag == null)
				throw new WellformedException("Not an invariant on "+e.getQualifiedName(),e1);
			else
				dmsg(ErrorType.SYNTAX, "Not an invariant on "+e.getQualifiedName(), e1, e);
		} catch (TypingWarningException ex) {
			dmsgW(ex.getMessage(), e);
		} catch (TypingException e1) {
			if (diag == null)
				throw new WellformedException("Not an invariant on "+e.getQualifiedName(),e1);
			else
				dmsg(ErrorType.TYPING, "Not an invariant on "+e.getQualifiedName(), e1, e);
		}
		return true;
	}

	private boolean checkAssocAttr(Classifier e) throws WellformedException {
		// Check if a class has overlapping associations and attributes
		// Firstly, create a dictionary of all OWNED attributes.
		// Then, check all associations against this dictionary.
		// TODO: use OCL?
		List<String> names = new ArrayList<String>(5);
		for (Property p : e.getAttributes()) {
			names.add(p.getName());
		}
		boolean ret = true;
		for (Association a : e.getAssociations()) {
			for (Property p : a.getMemberEnds()) {
				if (names.contains(p.getName())) {
					if (diag == null)
						throw new WellformedException("Duplicate thing '"+p.getName()+"' in class:"+e.getQualifiedName());
					else {
						ret = dmsg(ErrorType.WF,"Duplicate thing '"+p.getName()+"' in class:"+e.getQualifiedName(),e);
					}
				}
			}
		}
		return ret;
	}

	/**
	 * @author Charles Morisset
	 * @param e
	 * @return true if e implements the interface it claims it does
	 * @throws WellformedException
	 * Package-level visibility for testcase. 
	 * @throws RCOSModelCheckException 
	 */
	boolean checkImplements(Class e) throws WellformedException, RCOSModelCheckException {
		try {
			return checkClassImplements(e);
		}
		catch (WellformedException we) {
			showException(ErrorType.WF, we.getMessage() == null ? we.toString() : we.getMessage(), e);
			return false;
		}
	}

	public static boolean checkClassImplements(Class e) throws WellformedException {
		boolean ret = true;
		for (InterfaceRealization ir : e.getInterfaceRealizations()) {
			Interface i = ir.getContract();
			if (i == null) {
				throw new WellformedException("No contract on interface realization "+ir.getQualifiedName());
			} else {

				for (Operation op : i.getAllOperations()) {
					if (Utility.implementsOperation(e, op) == null)
						throw new WellformedException("Operation '"+op.getName()+"' required by interface '"+i.getQualifiedName() + "' but not implemented by class '" +e.getQualifiedName() + "'");
				}
			}
		}
		return ret;
	}

	protected RuleTCMembers getRtc() {
		assert th != null;
		if (rtc == null)
			rtc = new RuleTCMembersTH(th);
		return rtc;
	}

	protected void setRtc(RuleTCMembers rtc) {
		this.rtc = rtc;
	}
}
