package edu.unu.iist.rcos.spec.antlr;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xmi.impl.BasicResourceHandler;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.ocl.Class_OCL;
import edu.unu.iist.rcos.mm.util.TypingException;
import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.DesignOperation;
import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.profile.Stereotypes;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.PrinterException;
import edu.unu.iist.rcos.spec.antlr.visitor.TypePrinter;
import edu.unu.iist.rcos.spec.antlr.visitor.check.ExpGetType;
import edu.unu.iist.rcos.spec.antlr.visitor.check.ExpGetTypeNoSet;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.uml.RCOSLoader;

/**
 * @author charles
 * A class which handles all possible cases of type transformation: 
 * String -> Interface, Class, etc
 * TYPE -> Interface, Class, etc
 * 
 * TODO: Move into mm.resources plugin, but that requires quite a bit of assembly.
 * Firstly define the interface in mm.resources, then setup the corresponding field
 * in RCOSModel with a cache.
 */
public class TypeHandler extends PrimitiveTypes {

	final RCOSModel rcos;
	final boolean incompleteClass;

	public TypeHandler(RCOSModel rm) {
		this(rm,false);
	}

	@SuppressWarnings("deprecation")
	public TypeHandler(RCOSModel rm, boolean ic) {
		super(rm);
		rcos = rm;
		incompleteClass = ic;
	}

	@Deprecated
	public TypeHandler(Classifier cl) {
		this(Stereotypes.isRCOSModel(Stereotypes.getRootModel(cl)));
	}

	/**
	 * @author Charles Morisset
	 * @param name The name of the type 
	 * @param ltype The list of parameters, only relevant when <b>name</b> is generic
	 * @param c The classifier containing <b>name</b>  
	 * @param incompleteClass <b>true</b> if any not found reference is assumed to be Object, <b>false</b> otherwise
	 * @return The UML Type corresponding to <b>name</b>, the Object type if not found and <b>incompleteClass</b> is true, {@code null} otherwise
	 * @throws GetTypeException 
	 */
	public Type getTypeFromString(CNAME n, List<Type> ltype, Classifier c) throws GetTypeException {
		assert n != null;
		final String name = n.getName();
		assert name != null;
		Type t = getTypeFromString(name);
		if (t != null)
			return t;
		else {
			try {
				// Quickly check if it's an Enum:
				t = rcos.getClassModel().getBase_Package().getOwnedType(name);
				if (t instanceof Enumeration)
					return t;
				// else use previous logic.
				t = (Type) Utility.getClassifierParam(name, c);
				if (t != null)
					return t;
			} catch (WellformedException e) {
				// Don't want to propagate the WF exception
				// System.err.println(e.getMessage());
				t = null; // TODO: Can this be confused with void?
			}
			try {
				final Component co = Utility.getScope(c);
				t = (Type) getClass(name, ltype, co);
				if (t != null)
					return t;
				t = (Type) rcos.getInterface(name, co);
				if (incompleteClass && (t == null)) {
					System.err.println("Type '"+name+"' not found, assuming Object");
					t = getObject();
				}
				return t;
			} catch (TypingException e1) {
				throw new GetTypeException(e1);
			}
		}
	}

	/**
	 * @see #checkType(TYPE, Classifier)
	 */
	public Type getTypeFromTYPE(TYPE t, Classifier c) throws GetTypeException {
		assert t != null;
		assert c != null;
		if (t instanceof GTYPE) {
			List<Type> ltype = new ArrayList<Type>();
			for (TYPE e : ((GTYPE) t).getPar()) {
				final Type t2 = getTypeFromTYPE(e, c);
				if (t2 == null)
					throw new GetTypeException("Type "+printTYPE(e)+" not found in "+c.getQualifiedName(),t);
				ltype.add(t2);
			}
			return getTypeFromString(((GTYPE) t).getName(), ltype, c);
		} else {
			if (t instanceof NAT) return getUnlimitedNatural();
			else if (t instanceof INT) return getInt();
			else if (t instanceof DOUBLE) return getDouble();
			else if (t instanceof LONG) return getLong();
			else if (t instanceof CHAR) return getChar();
			else if (t instanceof BOOLEAN) return getBoolean();
			else if (t instanceof STRING) return getString();
			else return null;
		}
	}

	public static String printTYPE(TYPE t) {
		assert t != null;
		try {
			return new TypePrinter<PrinterException>().run(t);
		} catch (PrinterException e) {
			throw new RuntimeException(e);
		}
	}

	public Interface getInterfaceFromTYPE(GTYPE t, Classifier c) throws RCOSModelCheckException {
		List<Type> ltype = new ArrayList<Type>();
		try {
			for (TYPE e : t.getPar()) {
				ltype.add(getTypeFromTYPE(e, c));
			}
			Component co;
			//			if (c.getOwner() instanceof Classifier)
			//				c = (Classifier) c.getOwner();
			Interface i = getInterface(t.getName().getName(), ltype, co = Utility.getScope(c));
			if (i == null)
				throw new RCOSModelCheckException("Interface " + printTYPE(t) + " not found, scope: "+c.getQualifiedName() +" / co: "+co, t);
			else 
				return i;
		} catch (GetTypeException e1) {
			throw new RCOSModelCheckException(e1,t);
		} catch (TypingException e1) {
			throw new RCOSModelCheckException(e1,t);
		}
	}

	/**
	 * Create an AST TYPE node for a UML type.
	 * @throws WellformedException if the type is not an rCOS primitive type, or defined in an rCOS model.
	 */
	public TYPE createTYPE(Type t) throws WellformedException {
		assert(t != null);
		if (t == getBoolean()) return MyDesign.createBoolean();
		else if (t == getInt()) return MyDesign.createInt();
		else if (t == getChar()) return MyDesign.createChar();
		else if (t == getDouble()) return MyDesign.createDouble();
		else if (t == getLong()) return MyDesign.createLong();
		else if (t == getString()) return MyDesign.createString();
		else if (t == getUnlimitedNatural()) return MyDesign.createNat();
		// TODO: this should be handled more nicely:
		else if (t == getObject()) return MyDesign.createGenericType(t.getName(), null, null);
		else {
			/* If this is not a primitive type, only accept types defined in an
			 * rCOS model (e.g. this one, or an imported one, but not e.g. UML types.
			 * We might want to turn this into Objects, very much like we do with the
			 * "incomplete" flag in the parser.
			 */
			if (Stereotypes.isRCOSModelUnsafe(Stereotypes.getRootModel(t)) == null)
				throw new WellformedException("Unknown type "+t.getQualifiedName()+". Types must be of rCOS primitive type, or defined in an rCOS model.");
			List<TYPE> lt = new ArrayList<TYPE>();
			Classifier c = (Classifier) t;
			TemplateSignature ts = c.getOwnedTemplateSignature();
			if (ts != null)
				for (TemplateParameter tp : ts.getOwnedParameters()) {
					ParameterableElement pe = tp.getOwnedParameteredElement();
					if (pe != null && pe instanceof Class) {
						lt.add(createTYPE((Type) pe));
					} else {
						throw new WellformedException("The template signature of the class " + c.getQualifiedName() + " is not well-formed");
					}
				}
			return MyDesign.createGenericType(t.getName(), lt, null);
		}
	}

	/**
	 * Helper which saves the AST with designs, and the type information in the EXP nodes.
	 */
	public void saveModelWithEXPTypes() throws IOException {
		Map<?, ?> options = RCOSLoader.optionsMapUpgrade(new HashMap<String,Object>(), new RCOSResourceHandler());
		rcos.eResource().save(options);
	}

	/**
	 * This is a copy of RCOSLoader.RCOSResourceHandler,
	 * which saves the AST, and the type information in the EXP nodes.
	 * @author stolz
	 * @see RCOSLoader.RCOSResourceHandler
	 */
	class RCOSResourceHandler extends BasicResourceHandler {
		@Override
		public void preSave(XMLResource r, OutputStream s, Map<?,?> options) {
			// Can't delete entries directly while in the TreeIterator.
			// Maybe with prune()/remove()?
			List<EObject> discard = new LinkedList<EObject>();
			{
				TreeIterator<EObject> tree = r.getAllContents();
				while (tree.hasNext()) {
					EObject e = tree.next();
					if (e instanceof Element) {
						Element el = (Element)e;
						if (el.getEAnnotation(Utility.BINDING_DNS) != null) {
							discard.add(e);
						}
					} else if (e instanceof EXP) {
						EXP exp = (EXP) e;
						Classifier owner = findClassifierFromEXP(exp);
						try {
							// TODO: this eventually needs to be rewritten to a non-static solution.
							exp.setUmlType(getType(exp, owner));
						} catch (GetTypeException e1) {
							throw new RuntimeException(e1); // XXX
						}
					}

					if (e instanceof DesignOperation) {
						DesignOperation d = (DesignOperation) e;
						// TODO: To be completed by Li Dan.
						d.setAst(null);
						// TODO: augment all EXPs in the DESIGN with type info. 
					}
				}
			}
			for (EObject e : discard) {
				EcoreUtil.delete(e);
			}
		}

		private Classifier findClassifierFromEXP(EXP exp) {
			/* Go "up" via eContainer() until you reach an Operation.
			 * Will later need to be extended to Guards/Invariants.
			 * To be completed by Li Dan.
			 */
			// TODO Auto-generated method stub
			return null;
		}
	}

	public RCOSModel getModel() {
		return rcos;
	}

	public Type getType(EXP e, Classifier cl) throws GetTypeException {
		return new ExpGetType(cl,this).run(e);
	}

	public interface VoidOK {
		boolean voidOK();
	}

	public Type getType(EXP e, Classifier cl, final boolean vok) throws GetTypeException {
		Type t = new ExpGetType(cl,this) {
			@Override
			public boolean voidOK() { return vok; };
		}.run(e);
		return t;
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @param cl the class in which e is defined 
	 * @return true if e is of boolean type, false otherwise and throws an exception if e is not well-typed
	 * @throws TypingException
	 * @throws GetTypeException
	 */
	public boolean checkBoolean(EXP e, Classifier cl) throws GetTypeException {
		ExpGetType egt = new ExpGetType(cl,this);
		egt.visit(e);
		return (egt.isBooleanType());
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @param cl the class in which e is defined 
	 * @return true if e is of boolean type, false otherwise
	 */
	public boolean isBooleanExp(EXP e,  Classifier cl){	
		try {
			return checkBoolean(e, cl); 
		} catch (GetTypeException exc) {
			return false;			
		}
	}

	public Type getTypeEXPNoSet(EXP e, Classifier cl) throws GetTypeException {
		return new ExpGetTypeNoSet(cl,this).run(e);
	}

	/**
	 * @author Charles Morisset
	 * @param name The 'Name' of the class
	 * @param ltype The list of type instantiating the formal parameters of the class 'Name' 
	 * @param co The component containing the class, null if it is a top-level class
	 * @param rm The RCOSModel containing the class 
	 * @return The corresponding binding 'name< ltype >', which is created if it does not exist,
	 * or null if the instantiation is not correct. 
	 */
	public Class getClass(String name, List<Type> ltype, Component co) {
		Class cl = rcos.getClass(name, co);
		if (cl == null) return null;
		TemplateSignature ts = cl.getOwnedTemplateSignature();
		if (ts == null && (ltype == null || ltype.size() == 0)) return cl;
		if (ts == null && ltype != null) return null;
		// TODO: throw an exception when the parameters are not correct
		// right now, it will return "Class Not Found"
		if (ts.getOwnedParameters().size() > 0 && ltype == null) return null;
		if (ts.getOwnedParameters().size() != ltype.size()) {
			return null;
		}
		else {
			return getBinding(cl, ltype, co);			
		}
	}

	/**
	 * @author Charles Morisset
	 * @param name The 'Name' of the interface
	 * @param ltype The list of type instantiating the formal parameters of the interface 'Name' 
	 * @param co The component containing the interface, null if it is a top-level class
	 * @param rm The RCOSModel containing the interface 
	 * @return The corresponding binding 'name< ltype >', which is created if it does not exist,
	 * or null if the instantiation is not correct. 
	 */
	private Interface getInterface(String name, List<Type> ltype, Component co) {
		Interface cl = rcos.getInterface(name, co);

		if (cl == null) return null;
		TemplateSignature ts = cl.getOwnedTemplateSignature();
		if (ts == null && (ltype == null || ltype.size() == 0)) return cl;
		if (ts == null && ltype != null) return null;


		// TODO: throw an exception when the parameters are not correct
		// right now, it will return "Class Not Found"
		if (ts.getOwnedParameters().size() > 0 && ltype == null) return null;
		if (ts.getOwnedParameters().size() != ltype.size()) {
			return null;
		}
		else {
			return getBinding(cl, ltype, co);			
		}
	}

	private Interface getBinding(Interface cl, List<Type> ltype, Component co) {
		String name = cl.getLabel();
		for (Type t : ltype) {
			name += "_"+t.getLabel();
		}
		// The loop ends because there is not a infinite set of classes and we change the name at 
		// each iteration. So eventually, either there is no class with this name or it's a correct binding. 
		while (true) {
			Interface bind = rcos.getInterface(name, co); 
			// If no class found, then no binding either, so we create it
			if (bind == null) 
				return createBinding(name, cl, ltype, co); 
			// Else, there is a class with the same name, so we need to check it's the correct binding
			else if (isCorrectBinding(bind, cl, ltype, co)) 
				return bind;
			else 
				name += "_";
		}
	}

	private Interface createBinding(String name, Interface cl, List<Type> ltype, Component co) {
		return createBindingHelper(name, cl, new BindingHelper<Interface>() {
			public Interface add(String name) throws WellformedException {
				assert name != null;
				if (name.equals(""))
					throw new WellformedException("Empty interface name not permitted!");
				final Package componentPackage = rcos.getComponentModel().getBase_Package();
				if (componentPackage.getOwnedMember(name, false, UMLPackage.eINSTANCE.getInterface()) != null
						&& Class_OCL.MEMBERS_UNIQUE_NAMES != null)
					throw new WellformedException("Duplicate interface '"+name+"' in component package");
				return componentPackage.createOwnedInterface(name);				
			}
		}, ltype, co);
	}

	/**
	 * @author Charles Morisset
	 */
	private boolean isCorrectBinding(Interface bind, Interface cl,
			List<Type> ltype, Component co) {
		// TODO Auto-generated method stub
		return false;
	}

	/**
	 * @author Charles Morisset
	 */
	private Class getBinding(Class cl, List<Type> ltype, Component co) {
		String name = cl.getLabel();
		for (Type t : ltype) {
			name += "_"+t.getLabel();
		}
		// The loop ends because there is not a infinite set of classes and we change the name at 
		// each iteration. So eventually, either there is no class with this name or it's a correct binding. 
		while (true) {
			Class bind = rcos.getClass(name, co); 
			// If no class found, then no binding either, so we create it
			if (bind == null) 
				return createBinding(name, cl, ltype, co); 
			// Else, there is a class with the same name, so we need to check it's the correct binding
			else if (isCorrectBinding(bind, cl, ltype, co)) 
				return bind;
			else 
				name += "_";
		}
	}

	interface BindingHelper<T extends Classifier> {
		T add(String name) throws WellformedException;
	}

	/**
	 * @author Charles Morisset
	 */
	private <T extends Classifier> T createBindingHelper(String name, T cl, BindingHelper<T> bh,  List<Type> ltype, Component co) {
		T bind;
		try {
			bind = bh.add(name);
		} catch (WellformedException e) {
			// we now know that the class doesn't exist.
			throw new RuntimeException("This should not have happened",e);
		}
		TemplateBinding tb = UMLFactory.eINSTANCE.createTemplateBinding();
		tb.setBoundElement(bind);
		tb.setSignature(cl.getOwnedTemplateSignature());
		for (int i = 0; i < ltype.size(); i++) {
			TemplateParameterSubstitution ps = tb.createParameterSubstitution();
			ps.setFormal((TemplateParameter) cl.getOwnedTemplateSignature().getOwnedParameters().get(i));
			Utility.setActuals(ps,ltype.get(i));		
		}
		bind.createGeneralization(cl);
		// We don't want to keep Sets bindings, because they are handled using multiplicity
		if (cl.getLabel().equals("Set")) { // XXX literal
			bind.createEAnnotation(Utility.BINDING_DNS);
			bind.createOwnedComment().setBody("Automatically generated on-the-fly by type checker; will not be saved in the model.");
		}
		return bind;
	}

	/**
	 * @author Charles Morisset
	 */
	private Class createBinding(String name, Class cl, List<Type> ltype, Component co) {
		return createBindingHelper(name, cl, new BindingHelper<Class>() {
			public Class add(String name) throws WellformedException {
				assert name != null;
				assert !name.equals("");
				final Package classPackage = rcos.getClassModel().getBase_Package();
				if (classPackage.getOwnedMember(name, false, UMLPackage.eINSTANCE.getClass_()) != null
						&& Class_OCL.MEMBERS_UNIQUE_NAMES != null)
					throw new WellformedException("Duplicate class '"+name+"' in class package");
				Class c = classPackage.createOwnedClass(name, false);
				// rCOS: add generalization of Object
				return c;
			}
		}, ltype, co);
	}

	/**
	 * @author Charles Morisset
	 */
	private static boolean isCorrectBinding(Class bind, Class cl, List<Type> ltype, Component co) {
		// Firstly, bind has to inherits from cl
		if (bind.getGeneralization(cl) == null) return false;

		// The template binding of bind should correspond to the template signature of cl
		TemplateBinding tb = bind.getTemplateBinding(cl.getOwnedTemplateSignature());
		if (tb == null) return false;

		// There should be exactly as many parameter substitutions as types in ltype
		if (tb.getParameterSubstitutions().size() != ltype.size()) return false; 

		// And each of the actual parameters should be in ltype, in the correct order
		for (int i = 0; i < ltype.size(); i ++) {
			ParameterableElement p1 = Utility.getActuals(tb.getParameterSubstitutions().get(i));
			if (p1 != ltype.get(i)) return false;
		}
		return true;
	}

	public Type findSetType(Type type, Component co) {
		assert type != null;
		List<Type> ltype = Collections.singletonList(type);
		return (Type) getClass("Set", ltype, co);
	}
	
	public EnumerationLiteral findEnumLiteral(String c) {
		for (Type t : getModel().getClassModel().getBase_Package().getOwnedTypes()) {
			if (t instanceof Enumeration) {
				Enumeration e = (Enumeration) t;
				EnumerationLiteral lit = e.getOwnedLiteral(c);
				if (lit != null) {
					return lit;
				}
			}
		}
		return null;
	}
}
