package edu.unu.iist.rcos.spec.antlr;

import java.util.*;
import java.util.Map.Entry;

import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.FOR;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PREDICATE;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.TYPE;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.VARDECL;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.VARNAME;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

public abstract class Vardecls {

	public enum VarType { QUAN,PAR,LOCAL };
	final Deque<List<Vardecl>> vardecls;

	public abstract Type getType(TYPE t) throws GetTypeException;

	public Vardecls() {
		vardecls = new ArrayDeque<List<Vardecl>>();
		/* empty intial frame of declarations */
		pushDeclFrame();
	}

	public class Vardecl implements Entry<String,Type> {

		/* Declarations in current frame (most recent first).
		 * A new declaration will add itself to the frame.
		 */
		final private String name;
		private Type type;
		final public VarType pt; // type in parser

		public Vardecl(Vardecl v) {
			this(v.name, v.type, v.pt);
		}

		public Vardecl(String n, Type t, VarType vt) {
			name = n;
			type = t;
			pt = vt;
			top().add(0,this);
		}

		private Vardecl(VARNAME v, TYPE t, VarType vt) throws GetTypeException {
			this(v.getName(), getType(t), vt);
		}

		public Vardecl(VARDECL r) throws GetTypeException {
			this(r.getVar(),r.getType(), VarType.LOCAL);
		}

		public Vardecl(FOR d) throws GetTypeException {
			this(d.getVar(), d.getType(), VarType.QUAN);
		}

		public Vardecl(PREDICATE d) throws GetTypeException {
			this(d.getVar(), d.getType(), VarType.QUAN);
		}

		public String getKey() {
			return name;
		}

		public Type getValue() {
			return type;
		}

		public Type setValue(Type value) {
			Type old = getValue();
			type = value;
			return old;
		}
	}

	/**
	 * Remove a variable declaration from the current scope.
	 * @throws GetTypeException if the variable was never declared. 
	 */
	public void remove(VARNAME v) throws GetTypeException {
		List<Vardecl> cur = top();			
		for (Vardecl o : cur) {
			if (o.getKey().equals(v.getName())) {
				cur.remove(o);
				return;
			}
		}
		throw new GetTypeException("There's no variable named '"+v.getName()+"' in the current scope to end!",v);
	}

	private List<Vardecl> top() {
		return vardecls.getFirst(); // "top"
	}

	public void remove(Vardecl v) {
		if (!top().remove(v))
			throw new IllegalStateException("Variable "+v.name+" no longer declared?!");
	}

	/**
	 * Looks up the type of a local variable.
	 * @param n
	 * @return The type, which may be null.
	 * @see #find(String)
	 * @see #isLocal(String)
	 */
	public Type find(VARNAME n) {
		return find(n.getName());
	}

	/**
	 * Looks up the type of a local variable.
	 * @param n
	 * @return The type, which may be null.
	 * @see #find(VARNAME)
	 * @see #isLocal(String)
	 */
	protected Type find(String n) {
		for (List<Vardecl> cur : vardecls) {
			for (Vardecl o : cur) {
				if (o.getKey().equals(n))
					return o.getValue();
			}
		}
		return null;
	}

	/**
	 * Checks if a variable is locally declared.
	 * @param n
	 * @see #find(VARNAME)
	 * @see #find(String)
	 */
	public boolean isLocal(VARNAME n) {
		return isLocal(n.getName());
	}

	public boolean isLocal(String n) {
		for (List<Vardecl> cur : vardecls) {
			for (Vardecl o : cur) {
				// Note that o.getValue() might yield null!
				if (o.getKey().equals(n))
					return true;
			}
		}
		return false;
	}

	public void popDeclFrame() {
		vardecls.pop();
	}

	public void pushDeclFrame() {
		vardecls.push(new ArrayList<Vardecl>());
	}
}
