package edu.unu.iist.rcos.spec.antlr.commands;

import java.lang.ref.WeakReference;
import java.util.*;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.*;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.*;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.associationtorequiredinterface.AssocToReqIfExpVisitor;

public class AssociationToRequiredInterfaceCommand extends RCOSCommand<AssociationToRequiredInterfaceParameter> implements IRCOSCommand {

	public Property candidate;
	public Component co;
	public boolean cancelled = false;
	private NamedElement m_owner; // For memento, since we loose this information upon deletion
	private String reqIfName;

	public AssociationToRequiredInterfaceCommand(){}

	public AssociationToRequiredInterfaceCommand(EditingDomain domain, List<Throwable> errors, AssociationToRequiredInterfaceParameter params) {
		super(domain,errors,params);
	}

	static class AssocProvider extends EMFDeleteProvider implements AssociationToRequiredInterfaceProvider<Command> {
		public AssocProvider(EditingDomain domain, Command cmd) {
			super(domain, cmd);
		}

		public void createEdge(Interface iface, Usage u, Component co) {
			// Nothing to do in EMF-land.
		}

		public void deleteAssoc(Association a, CompoundCommand cmd) {
			add(cmd);
			delete(a);
		}

		public void undoExecuted() {
			// Only needed in GEF-land
		}
	}

	@Override
	protected void initCommand(EditingDomain ed, List<Throwable> errors, AssociationToRequiredInterfaceParameter p){
		super.initCommand(ed, errors, p);
		final AssocProvider dp = new AssocProvider(ed, IdentityCommand.INSTANCE);
		setProvider(dp);
	}

	@Override protected Command createCommand() {
		try {
			if (params.getAssociation() == null)
				return recordError(new IllegalArgumentException(params.toString()));
			Element e = params.getAssociation().getElement();
			if (!(e instanceof Association)) {
				String msg = e == null ? "nothing." : e.eClass().getName();
				throw new TransformationFailedException("Element has wrong type, expected Association, got "+msg);
			}
			Association a = (Association) e;
			candidate = getCandidate(a);
			if (candidate == null)
				throw new TransformationFailedException(a.getName() + ": doesn't have a suitable navigation end.");
			m_owner = (NamedElement) a.getOwner();
			co = getTargetComponent(candidate);
			assert co != null : candidate.getQualifiedName();

			reqIfName = params.getName(); // May call into GUI!
			if (reqIfName == null) {
				cancelled = true;
				return UnexecutableCommand.INSTANCE; // cancelled
			}
			checkContractInterfaceName(reqIfName);
			final rCOSComponent rco = Stereotypes.isRCOSComponent(co);
			getTargetNameFree(reqIfName, rco);

			// All preconditions checked, let's do some transforming:

			final Interface i = UMLFactory.eINSTANCE.createInterface();
			i.setName(reqIfName);

			CompoundCommand cmd = new StrictCompoundCommand() {{isPessimistic = true; }};
			// copy operations
			for (Operation o : ((Class)candidate.getType()).getOperations()) {
				final EList<Type> pts = new BasicEList<Type>();
				EList<String> pns = new BasicEList<String>();
				for (Parameter p : o.getOwnedParameters()) {
					pts.add(p.getType());
					pns.add(p.getName());
				}
				// TODO: use CreateOperationCommand
				final Operation opOut = i.createOwnedOperation(o.getName(), pns, pts);
				{
					// fix parameter direction. Don't we have a helper for this?
					Iterator<Parameter> opPs = o.getOwnedParameters().iterator();
					opPs.hasNext();
					for (Parameter p : opOut.getOwnedParameters()) {
						Parameter opP = opPs.next();
						assert p.getName().equals(opP.getName());
						p.setDirection(opP.getDirection());
					}
				}
			}
			Usage u = UMLFactory.eINSTANCE.createUsage();
			{ // create the Usage
				u.setName(co.getName()+" requires "+i.getName()); // unique name doesn't matter so much
				Collection<EObject> is = new ArrayList<EObject>();
				is.add(i);
				cmd.append(AddCommand.create(domain, u, UMLPackage.Literals.DEPENDENCY__SUPPLIER, is));
				Collection<EObject> cs = new ArrayList<EObject>();
				cs.add(co);
				cmd.append(AddCommand.create(domain, u, UMLPackage.Literals.DEPENDENCY__CLIENT, cs));
				Collection<EObject> us = new ArrayList<EObject>();
				us.add(u);
				cmd.append(AddCommand.create(domain, co, UMLPackage.Literals.COMPONENT__PACKAGED_ELEMENT, us));
			}
			cmd.append(AddCommand.create(domain, getOwner(co), UMLPackage.Literals.PACKAGE__OWNED_TYPE, Collections.singleton(i)));
			final ApplyStereotypeCommand stCmd;
			cmd.append(stCmd = new ApplyStereotypeCommand(i, Stereotypes.getSt(ContractInterface.class,co)));
			// "loop back" interface:
			cmd.append(new AbstractCommand(){ // TODO: Buglet? Using SetCommand in the CommandWrapper suddenly stopped working!
				@Override protected boolean prepare() { return true; };
				public void execute() {
					final ContractInterface ci = getCI(stCmd);
					ci.setInterface(i);
					params.setRequired(ci);
				}
				/**
				 * Fetch the redone object.
				 * @see ApplyStereotypeCommand#getResult()
				 */
				public void redo() { execute(); }
				@Override public void undo() { /* Don't care, object will be destroyed anyway */ };
			});
			final Class usingClass = (Class)(candidate.getOtherEnd().getType());
			assert usingClass != null : candidate.getQualifiedName();
			assert usingClass.eResource() != null : "Class not contained in model: "+usingClass.getName(); // TODO

			// TODO Pattern: forall Ops;Guards;Invs
			/* We now rewrite the designs. But first, we will "discharge" all
			 * non-null requirements in designs (most notably in preconditions), guards, and the invariant.
			 * If we do it in this order, we can make AssocToReqIfDesignVisitor stop the process if it
			 * encounters something it doesn't "like" more easily.
			 */
			cmd.append(AssertNotNullCommand.mkAssertNotNullCommand(domain, errors, candidate.getName(), usingClass));
			/* Now the designs: */
			{
				final AssocToReqIfExpVisitor av = new AssocToReqIfExpVisitor(candidate,reqIfName);
				final DesignsED desEd = new DesignsED(domain);

				for (final Operation op : usingClass.getOwnedOperations()) {
					/* Delay evaluation so that we have chance to pick up the designs rewritten in above through AssertNotNullCommand. */
					cmd.append(new CommandWrapper() {
						@Override protected Command createCommand() {
							try {
								DESIGN designStr = Parse.getDesign(op);
								if (designStr != null) { // TODO: pattern
									av.setExceptionContex(op.getName()+"()");
									DESIGN v = av.new AssocToReqIfDesignVisitor(usingClass).run(designStr);
									assert v != null;
									return desEd.update(op, DesignUtility.designToString(op, v));
								} else
									return IdentityCommand.INSTANCE;
							} catch (Exception e) {
								return recordError(e);
							}
						}
					});
				}
			}

			// Save rCOS model for validation later:
			RCOSModel rm = Stereotypes.isRCOSModel(Stereotypes.getRootModel(a));

			getProvider().deleteAssoc(a,cmd);
			// ^Mustn't use cmd anymore, get the current result form the provider instead!
			// Stub for creating edge in diagram. May need changes:			
			getProvider().createEdge(i,u,co);
			/* ^Since we know that this will happen in the GEF side-channel, we don't even bother to
			 * fetch a new command, unlike the deleteAssoc command.
			 */

			/* Run validation in-line, and collect result. */
			getProvider().add(new ValidateModelCommand(rm) {
				@Override protected Command handleException(Throwable t) {
					getProvider().undoExecuted(); return recordError(new TransformationValidationException(t));
				}
			});
			return getProvider().get();
		} catch (TransformationFailedException e) {
			return recordError(e);
		}
	}

	@Override public void undo() {
		super.undo();
	}

	public Parameters setData(Class elC) {
		elC.setName("Association to req. interface "+candidate.getName());
		ExecutionList el = Stereotypes.apply(ExecutionList.class, elC);
		Class c;
		c = ((Package)elC.getOwner()).createOwnedClass("Query "+elC.getName(), false);
		c.createEAnnotation(Utility.BINDING_DNS);
		final Query q = Stereotypes.apply(Query.class, c);

		SOE_Element soe_owner = ProfileFactory.eINSTANCE.createSOE_Element();
		soe_owner.setElement(m_owner); // TODO: Listener, might need a memento himself

		q.setObject(soe_owner);
		/* XXX Experimental and incomplete: even though we have removed the element from the model, it is not entirely dead!
		 * The object still lives on the undo stack (TODO: how long? and then?).
		 * We can attach a notifier onto it, and when the command is undo, and the element edited, we can track
		 * the live changes here. Unfortunately, this breaks down at the model: if you have saved this memento and
		 * loaded the model again, of course the listener will no longer exist. We would have to reinstantiate them.
		 * For deletion, we would have to distinguish between forward-engineering (the element got transformed, so we
		 * have to keep the memento) and model revision (the element gets deleted from the initial model, and thus this
		 * command becomes garbage. 
		 */
		candidate.eAdapters().add(new AdapterImpl() {
			final WeakReference<Query> wr = new WeakReference<Query>(q);
			public final void notifyChanged(Notification msg) {
				Query q = wr.get();
				if (q == null) {
					System.err.println("Wow, please tell Volker!");
					getTarget().eAdapters().remove(this);
					return;
				}
				if (msg.getEventType() == Notification.SET && msg.getFeature() == UMLPackage.eINSTANCE.getNamedElement_Name())
					q.setQuery("ownedMember->select(name = '"+msg.getNewStringValue()+"')->asSequence()->at(1)");
			}
		});
		q.setQuery("ownedMember->select(name = '"+candidate.getName()+"')->asSequence()->at(1)");

		c = ((Package)elC.getOwner()).createOwnedClass("Cmd "+elC.getName(), false);
		c.createEAnnotation(Utility.BINDING_DNS);
		AssociationToRequiredInterfaceParameter st = Stereotypes.apply(AssociationToRequiredInterfaceParameter.class, c);
		st.setName(reqIfName);
		SOE_Query soe = ProfileFactory.eINSTANCE.createSOE_Query();
		soe.setMetaclass(UMLPackage.eINSTANCE.getAssociation());
		soe.setQuery(q);
		st.setAssociation(soe);

		el.getList().add(q);
		el.getList().add(st);
		return el;
	}

	public interface NameProvider {
		String getName(Component inCo) throws InterruptedException;
	}

	public static void checkContractInterfaceName(String name) throws TransformationFailedException {
		final char ch = name.toCharArray()[0];
		if (ch == Character.toLowerCase(ch))
			throw new TransformationFailedException("Contract interface name must start with an uppercase letter: "+name);
	}

	/**
	 * Find out in which component we actually are, that is, which component will receive the new required interface.
	 */
	private static Component getTargetComponent(Property candidate) throws TransformationFailedException {
		assert candidate != null;
		final Class candidateOwner = (Class) candidate.getOtherEnd().getType(); // TODO: review...Property.getClass_() doesn't work, but maybe we're doing things the wrong way
		if (candidateOwner == null)
			throw new TransformationFailedException("Property "+candidate.getOtherEnd().getName()+" does not seem to have a type.");
		try {
			assert !(candidateOwner instanceof Component); // WF
			return Utility.getScope(candidateOwner);
		} catch (TypingException e) {
			throw new TransformationFailedException(e);
		}
	}

	private static Property getCandidate(Association a) throws TransformationFailedException {
		Property candidate = null;
		for (Property me : a.getNavigableOwnedEnds()) {
			if (me.getLower() == 1 && me.getUpper() == 1) {
				if (candidate != null) {
					throw new TransformationFailedException(a.getName() + ": doesn't have a unique navigation end.");
				}
				candidate = me;
			}
		}
		return candidate;
	}

	public static void getTargetNameFree(String name, rCOSComponent co) throws TransformationFailedException {
		assert co != null;
		String msg;
		if ((msg = MyRCOSModel.canAddInterface(name, co)) != null)
			throw new TransformationFailedException(msg);
	}

	/**
	 * Small wrapper that should somehow relate to MyRCOSModel and indicate where the
	 * new contract interface should be created.
	 * TODO: Refactor with MyRCOSModel.addContract/addInterface
	 */
	private static Package getOwner(Component co) {
		return co.getNearestPackage();
	}

	/* The Provider interface to handle EMF or GEF: */
	private AssociationToRequiredInterfaceProvider<Command> provider;

	protected AssociationToRequiredInterfaceProvider<Command> getProvider() {
		return provider;
	}

	public void setProvider(AssociationToRequiredInterfaceProvider<Command> provider) {
		this.provider = provider;
	}

	public Component getComponent() {
		return this.co;
	}

	public static ContractInterface getCI(final ApplyStereotypeCommand stCmd) {
		final Collection<?> result = stCmd.getResult();
		assert result.size() == 1;
		return (ContractInterface) result.toArray()[0];
	}
}