package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.*;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.TypingException;
import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.profile.*;

public class ClassToComponentCommand extends RCOSCommand<ClassToComponentParameter> {

	private ClassToComponentProvider prov;

	// TODO: We might want to access the resulting component from the command later. 
	public ClassToComponentCommand(){}

	public ClassToComponentCommand(EditingDomain ed, List<Throwable> errors, ClassToComponentParameter p) {
		super(ed, errors, p);
	}

	public interface ClassToComponentProvider {
		void addComponent(Package pkg, Component parent, Component co);
	}

	public static class Provider implements ClassToComponentProvider {

		public void addComponent(Package pkg, Component parent, Component co) {
			// Nothing to do in EMF
		}
	}

	@Override
	public Command createCommand() {
		if (params.getSourceClass() == null)
			return recordError(new IllegalArgumentException("Parameters not configured correctly on "+params.getBase_Class().getQualifiedName()));

		Class clazz = params.getSourceClass().getQueryClass();
		assert clazz != null; // TODO
		try {
			if (Utility.isControllerClass(clazz)) // TODO: This check should probably be recorded "semantically", e.g. on the ClassQ interface.
				return recordError( new IllegalArgumentException("Can't create a new component from an existing controller class."));
		} catch (TypingException e) {
			return recordError(e);
		}
		String componentName = params.getTargetComponentName();
		assert componentName != null; // TODO
		CompoundCommand resultCommand = new StrictCompoundCommand() {{isPessimistic = true; }};
		
		Component newComponent = UMLFactory.eINSTANCE.createComponent();
		newComponent.setName(componentName);
		/* TODO: Check for duplicate name? */

		/* Find out where to put the component.
		 * TODO: We need to find a suitable place later, it doesn't ALWAYS go directly into ComponentModel,
		 * it might be hierarchical as well. */
		final RCOSModel rm = Stereotypes.isRCOSModel(Stereotypes.getRootModel(clazz));
		assert rm != null;
		ComponentModel coM = rm.getComponentModel();
		assert coM != null;
		Package targetPackage = (Package) coM.getBase_Package();
		
		resultCommand.append(AddCommand.create(domain, targetPackage, UMLPackage.Literals.PACKAGE__OWNED_TYPE, Collections.singleton(newComponent)));
		resultCommand.append(ApplyStereotypeCommand.create(newComponent, ServiceComponent.class));
		prov.addComponent(targetPackage, null /* XXX */, newComponent);

		final Interface iface = createInterfaceFromClass(clazz);
		/* We pull the usual trick of reusing the syntactic interface for the contract interface */
		final Interface contract = iface;

		resultCommand.append(AddCommand.create(domain, newComponent, UMLPackage.Literals.COMPONENT__PACKAGED_ELEMENT, Collections.singleton(contract)));
		resultCommand.append(new MakeContractInterfaceCommand(contract, Stereotypes.getSt(ContractInterface.class, rm.getBase_Model()),iface){
			@Override public void execute() {
				super.execute();
				assert result instanceof ContractInterface;
				params.setProvided((ContractInterface) result);
			}
		});
 
		newComponent.createInterfaceRealization(clazz.getName()+"_IR", contract);
		return resultCommand;
	}

	private static Interface createInterfaceFromClass(Class clazz2) {
		Interface result = UMLFactory.eINSTANCE.createInterface();
		result.setName("Interface_From_" + clazz2.getName());
		List<Operation> publicOperation = getPublicOperationOfClass(clazz2);
		for(Operation operation : publicOperation){
			EList<String> names = new BasicEList<String>();
			EList<Type> types = new BasicEList<Type>();
			for(Parameter parameter : operation.getOwnedParameters()){
				names.add(parameter.getName());
				types.add(parameter.getType());
			}
			result.createOwnedOperation(operation.getName(), names, types/*, operation.getReturnResult().getType()*/);
		}
		return result;
	}

	private static List<Operation> getPublicOperationOfClass(Class clazz2) {
		List<Operation> result = new ArrayList<Operation>();
		for(Operation operation : clazz2.getAllOperations()){
			if(operation.getVisibility().equals(VisibilityKind.PUBLIC_LITERAL))
				result.add(operation);
		}
		return result;
	}

	public void setProvider(ClassToComponentProvider prov) {
		this.prov = prov;
	}
}
