package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.RemoveCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.*;

public class ComposeRequiredProvidedCommand extends RCOSCommand<ComposeRequiredProvidedParameter> {

	public ComposeRequiredProvidedCommand(){};
	public ComposeRequiredProvidedCommand(EditingDomain domain, List<Throwable> errors, ComposeRequiredProvidedParameter params) {
		super(domain,errors,params);
	}

	@SuppressWarnings("serial")
	@Override protected Command createCommand() {
		/* Hm, we can't tell from the error if the provider is failing or the parameter is not configured.
		 * Hopefully, validation will take care of the parameter.
		 */
		if (params.getProvided() == null || params.getProvided().getProvided() == null)
			return recordError(new IllegalArgumentException("No provided interface!"));
		if (params.getRequired() == null || params.getRequired().getRequired() == null)
			return recordError(new IllegalArgumentException("No required interface!"));

		final ContractInterface prov = params.getProvided().getProvided();
		final ContractInterface req = params.getRequired().getRequired();
		/* TODO: Consistency -- the two interfaces must be otherwise unused. */
		/* XXX: Refactor WF! */
		for (DirectedRelationship dr : prov.getBase_Interface().getSourceDirectedRelationships(UMLPackage.eINSTANCE.getDependency())) { // getTarget...? XXX We need a convention for composition, req -> source?
			if (dr instanceof Usage)
				return recordError(new IllegalArgumentException("Provided contract interface "+prov.getBase_Interface().getQualifiedName()+" already used by "+((Usage)dr).getQualifiedName()));
			if (dr instanceof Dependency) {
				Composition c = Stereotypes.isComposition((Dependency) dr);
				if (c != null)
					return recordError(new IllegalArgumentException("Provided contract interface "+prov.getBase_Interface().getQualifiedName()+" already used by composition "+c.getBase_Dependency().getQualifiedName()));
				Delegation d = Stereotypes.isDelegation((Dependency) dr);
				if (d != null)
					return recordError(new IllegalArgumentException("Provided contract interface "+prov.getBase_Interface().getQualifiedName()+" already used by delegation "+d.getBase_Dependency().getQualifiedName()));
			}
		}

		final Component cocomp = UMLFactory.eINSTANCE.createComponent();
		cocomp.setName(params.getComponentName());

		CompoundCommand cmd = new StrictCompoundCommand() {{ isPessimistic = true; }};
		/* Record result for subsequent transformations */
		final RCOSModel rcosModel = Stereotypes.isRCOSModel(Stereotypes.getRootModel(prov.getBase_Interface()));
		ComponentModel como = rcosModel.getComponentModel();
		Command addToModelCmd = AddCommand.create(domain, como.getBase_Package(), UMLPackage.Literals.PACKAGE__PACKAGED_ELEMENT, new ArrayList<Component>(){{ add(cocomp);}});
		cmd.append(addToModelCmd);
		cmd.append(SetCommand.create(domain,params,ProfilePackage.Literals.COMPOSE_REQUIRED_PROVIDED_PARAMETER__COMPONENT,cocomp));


		/* TODO: Calculate resulting unmatched provided/required interfaces from the two involved components, and synthesize.
		 * These should then be read-only?
		 */

		try {
			final Component provComponent = Utility.getInterfaceComponent(prov);
			if (provComponent == null)
				return recordError(new IllegalArgumentException()); // XXX shouldn't happen
			final Component reqComponent = Utility.getInterfaceComponent(req);
			if (reqComponent == null)
				return recordError(new IllegalArgumentException()); // XXX shouldn't happen

			/* XXX WF: can't compose already owned components, has to be on top-level */
			if (provComponent.getOwner() instanceof Component)
				return recordError(new IllegalArgumentException("Component "+provComponent.getName()+" already owned by "+((Namespace)provComponent.getOwner()).getName()));
			if (reqComponent.getOwner() instanceof Component)
				return recordError(new IllegalArgumentException("Component "+reqComponent.getName()+" already owned by "+((Namespace)reqComponent.getOwner()).getName()));

			final Stereotype st = Stereotypes.getSt(ServiceComponent.class, rcosModel.getBase_Model());
			cmd.append(new ApplyStereotypeCommand(cocomp,st));

			/* Relocate both components "inside" the new one.
			 * Do things explicitly, as otherwise a plain Add or Set without the remove will not undo properly.
			 */
			cmd.append(RemoveCommand.create(domain, reqComponent.getOwner(), reqComponent.eContainingFeature(), Collections.singleton(reqComponent)));
			cmd.append(RemoveCommand.create(domain, provComponent.getOwner(), provComponent.eContainingFeature(), Collections.singleton(provComponent)));
			cmd.append(AddCommand.create(domain, cocomp, UMLPackage.Literals.CLASS__NESTED_CLASSIFIER, new ArrayList<Component>(2){{ add(provComponent); add(reqComponent); }}));
			cmd.append(CompositionCommand.create(domain, rcosModel, cocomp, req, prov));
			/* Now propagate all free interfaces from the composed components to the outer component */
			cmd.append(new CommandWrapper() {
				@Override
				protected Command createCommand() {
					CompoundCommand cmd = new StrictCompoundCommand() {{ isPessimistic = true; }};
					try {
						ProviderHelper ph = new ProviderHelper(cocomp,cmd,"p");
						ph.propagate(reqComponent);
						ph.propagate(provComponent);
					} catch (WellformedException e) {
						return recordError(e);
					}
					cmd.append(IdentityCommand.INSTANCE);
					return cmd.unwrap();
				}
			});
		} catch (WellformedException e) {
			return recordError(e);
		}

		/* TODO: update diagrams.
		 * We could get the EMFGraphNode, and fire a RestoreConnectionsRequest.
		 */
		return cmd;
	}

	interface ContractInterfaceIsFree<T extends Dependency> {
		EList<Interface> getInterfaces();
		boolean isFree(Dependency r);
		T createDependency(Interface oI);
	}

	class ProviderHelper {
		/* With hindsight, I would do this again, differently:
		 * comppute the set of free prov/req interfaces by simply enumerating them, and eliding
		 * the two passed as parameters for the composition.
		 * 
		 * TODO:
		 * - one GUI glitch: outer provided interface doesn't show interface realization(!?)
		 */
		Port port = null; /* Only create a single port */
		final private String portName;
		final private CompoundCommand cmd;
		final private Component outer;

		public ProviderHelper(Component outer, CompoundCommand cmd, String portName) {
			this.portName = portName;
			this.cmd = cmd;
			this.outer = outer;
		}

		public synchronized Port getPort() {
			if (port == null) {
				port = UMLFactory.eINSTANCE.createPort();
				port.setName(portName);
				port.setAggregation(AggregationKind.COMPOSITE_LITERAL);
				cmd.append(AddCommand.create(domain, outer, UMLPackage.Literals.STRUCTURED_CLASSIFIER__OWNED_ATTRIBUTE, Collections.singleton(port)));
			}
			return port;
		}

		public void propagate(final Component co) throws WellformedException {
			ContractInterfaceIsFree<Usage> usageProvider = new ContractInterfaceIsFree<Usage>() {
				public EList<Interface> getInterfaces() { return co.getRequireds(); }

				public boolean isFree(Dependency d) {
					return !(d instanceof InterfaceRealization || Stereotypes.isComposition(d) != null || Stereotypes.isDelegation(d) != null);
				}

				public Usage createDependency(Interface oI) {
					Usage u = UMLFactory.eINSTANCE.createUsage();
					cmd.append(AddCommand.create(domain, u, UMLPackage.Literals.DEPENDENCY__CLIENT, Collections.singleton(outer)));
					cmd.append(AddCommand.create(domain, u, UMLPackage.Literals.DEPENDENCY__SUPPLIER, Collections.singleton(oI)));
					cmd.append(AddCommand.create(domain, outer, UMLPackage.Literals.COMPONENT__PACKAGED_ELEMENT, Collections.singleton(u)));
					return u;
				}
			};
			
			ContractInterfaceIsFree<InterfaceRealization> realizationProvider = new ContractInterfaceIsFree<InterfaceRealization>() {
				public EList<Interface> getInterfaces() { return co.getProvideds(); }

				public boolean isFree(Dependency d) {
					return !(d instanceof Usage || Stereotypes.isComposition(d) != null || Stereotypes.isDelegation(d) != null);
				}

				@SuppressWarnings("serial")
				public InterfaceRealization createDependency(Interface oI) {
					InterfaceRealization ir = UMLFactory.eINSTANCE.createInterfaceRealization();
					cmd.append(AddCommand.create(domain, ir, UMLPackage.Literals.DEPENDENCY__CLIENT, new ArrayList<Element>(2){{ add(getPort()); add(outer); }}));
					cmd.append(AddCommand.create(domain, ir, UMLPackage.Literals.DEPENDENCY__SUPPLIER, Collections.singleton(oI)));
					cmd.append(SetCommand.create(domain, ir, UMLPackage.Literals.INTERFACE_REALIZATION__CONTRACT, oI));
					cmd.append(AddCommand.create(domain, outer, UMLPackage.Literals.BEHAVIORED_CLASSIFIER__INTERFACE_REALIZATION, Collections.singleton(ir)));
					return ir;
				}
			};
			propagate2(usageProvider, co);
			propagate2(realizationProvider, co);
		}

		private void propagate2(ContractInterfaceIsFree<?> ciP, Component co) throws WellformedException {
			final Stereotype ciSt = Stereotypes.getSt(ContractInterface.class, co);
			for (Interface i : ciP.getInterfaces()) {
				ContractInterface ci = Stereotypes.isContractInterface(i);			
				if (ci != null) {
					boolean free = true;
					EList<Relationship> deps = i.getRelationships();
					for (Relationship r : deps) {
						/* Not 100% sure why this one may happen: */
						if (r.getOwner() == null || !(r instanceof Dependency))
							continue;
						Dependency d = (Dependency) r;
						free &= ciP.isFree(d);
						if (!free)
							break; // break into outer for()
					}
					if (free) {
						/* create another CI, wire synt. IF, delegate */
						Interface oI = UMLFactory.eINSTANCE.createInterface();
						oI.setName(i.getName());
						/* TODO: Review: here we create the CI inside the component. Maybe it should be outside. */
						cmd.append(AddCommand.create(domain, outer, UMLPackage.Literals.COMPONENT__PACKAGED_ELEMENT, Collections.singleton(oI)));
						final MakeContractInterfaceCommand oCICmd;
						cmd.append(oCICmd = new MakeContractInterfaceCommand(oI, ciSt, ci.getInterface()));

						ciP.createDependency(oI);

						Dependency delegation = UMLFactory.eINSTANCE.createDependency();
						delegation.setName(outer.getName()+" to "+i.getName());
						cmd.append(AddCommand.create(domain, outer, UMLPackage.Literals.COMPONENT__PACKAGED_ELEMENT, Collections.singleton(delegation)));
						final ApplyStereotypeCommand delStCmd;
						cmd.append(delStCmd = new ApplyStereotypeCommand(delegation, Stereotypes.getSt(Delegation.class, co)));
						cmd.append(new CommandWrapper() {
							@Override protected Command createCommand() {
								final Delegation del = (Delegation) delStCmd.getResult().iterator().next();
								final Object o = oCICmd.getResult().iterator().next();
								assert o != null;
								assert del != null;
								return SetCommand.create(domain, del, ProfilePackage.Literals.DELEGATION__INTERFACE, o);
							}

							/* Hack */
							@Override public void redo() {
								final Delegation del = (Delegation) delStCmd.getResult().iterator().next();
								final Object o = oCICmd.getResult().iterator().next();
								assert o != null;
								assert del != null;
								del.setInterface((ContractInterface) o);
							}
						});
						Port port = getPort();
						cmd.append(AddCommand.create(domain,delegation,UMLPackage.Literals.DEPENDENCY__SUPPLIER, Collections.singleton(port)));
						cmd.append(AddCommand.create(domain,delegation,UMLPackage.Literals.DEPENDENCY__CLIENT, Collections.singleton(i)));
					}
				}
			}
		}
	}

	static class CompositionCommand {

		public static Command create(EditingDomain domain, RCOSModel rm, Component cocomp,
				final ContractInterface req, final ContractInterface prov) {
			Stereotype st = Stereotypes.getSt(Composition.class, rm.getBase_Model());
			final Dependency dep = UMLFactory.eINSTANCE.createDependency();
			dep.setName("composition req/prov "+cocomp.getName());
			dep.getClients().add(req.getBase_Interface());
			dep.getSuppliers().add(prov.getBase_Interface());
			CompoundCommand cmd = new StrictCompoundCommand() {{ isPessimistic = true; }};
			/* We can't just cocomp.getPackagedElements().add(dep) here, as we don't know who owns the component in general---
			 * it might have come from another command.
			 */
			cmd.append(AddCommand.create(domain, cocomp, UMLPackage.Literals.COMPONENT__PACKAGED_ELEMENT, Collections.singleton(dep)));
			cmd.append(new ApplyStereotypeCommand(dep, st));
			return cmd;
		}
	}
}
