package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.*;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.*;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;

/**
 * TODO: convenience access to guards?
 * @author stolz
 */
public class CreateOperationCommand extends RCOSCommand<CreateOperationParameter> {

	public CreateOperationCommand(EditingDomain ed, List<Throwable> errors,
			CreateOperationParameter p) {
		super(ed, errors, p);
	}

	/**
	 * There is something unorthogonal about the arguments: you can pass the name and the class
	 * by parameter, but the signature comes from the overridden methods. However, this just
	 * reflects the current use cases for this class.
	 */
	public CreateOperationCommand(EditingDomain ed, List<Throwable> errors,
			String mName, Classifier c) {
		this(ed, errors, ProfileFactory.eINSTANCE.createCreateOperationParameter());
		this.params.setClass_(c);
		this.params.setMethodName(mName);
	}

	public CreateOperationCommand(){}

	@Override
	public Command createCommand() {
		assert params != null;
		assert params.getBase_Class() == null || params.getBase_Class() != null; // !!!
		final String cmdQualifiedName = params.getBase_Class() == null ? "<internal>" : params.getBase_Class().getQualifiedName();
		if (getClass_() == null)
			return recordError(new IllegalArgumentException("No class set for command "+cmdQualifiedName));
		if (getMethodName() == null || getMethodName().equals(""))
			return recordError(new IllegalArgumentException("No method name set for command "+cmdQualifiedName));
		final String classQualifiedName = getClass_().getQualifiedName();
		// In rCOS, we never care about the parameter names.
		if (getClass_().getOperation(getMethodName(), null, getTypes()) != null)
			return recordError(new Exception("Operation "+getMethodName()+" already exists in class "+classQualifiedName));

		// Don't hide existing methods (TODO: PO)
		if (getTypes() != null && params.isCheckOverrides()) {
			assert getClass_().eResource() != null : this.toString(); // ick!
			RCOSModel rm = Stereotypes.isRCOSModel(Stereotypes.getRootModel(getClass_()));
			PrimitiveTypes pt = rm.getPrimitiveTypes();
			outer:
				for (Operation op : getClass_().getAllOperations()) {
					// XXX Should use ExpGetType.getOperation!
					if (op.getOwnedParameters().size() == getTypes().size()
							&& op.getName().equals(getMethodName())) {
						Iterator<Type> it = getTypes().iterator();
						for (Parameter p : op.getOwnedParameters()) {
							if (!(it.hasNext() && pt.isSubType(it.next(), p.getType())))
								continue outer;
						}
						return recordError(new Exception("Operation "+getMethodName()+" must not override "+qName(op)));
					}
				}
		}

		Operation op = UMLFactory.eINSTANCE.createOperation();
		op.setName(getMethodName());
		if (getNames() != null && getTypes() != null) {
			if (getNames().size() != getTypes().size())
				return recordError(new IllegalArgumentException());
			Iterator<Type> it = getTypes().iterator();
			for (String s : getNames()) {
				op.createOwnedParameter(s, it.next());
			}
		}
		if (getDesign() != null)
			try {
				Designs.iNSTANCE.create(op, getDesign());
			} catch (WellformedException e) {
				return recordError(e);
			}
			return AddCommand.create(domain, getClass_(), UMLPackage.Literals.CLASS__OWNED_OPERATION, op)
			// record the created operation:
			.chain(SetCommand.create(domain, params, ProfilePackage.Literals.CREATE_OPERATION_PARAMETER__CREATED_OPERATION, op));
	}

	protected Classifier getClass_() {
		return params.getClass_();
	}

	protected String getMethodName() {
		return params.getMethodName();
	}

	private String qName(Operation op) {
		return op.getQualifiedName();
	}

	/* Hooks to inject parameters */
	/**
	 * The design to add as a body (or null).
	 * @see #createDesign(DesignProviderI)
	 */
	protected String getDesign() { return null; }
	protected EList<String> getNames() {
		return null;
	}

	/**
	 * For advanced signature check. Note well the difference between 'null' (no check),
	 * and an empty list (the empty signature).
	 */
	protected EList<Type> getTypes() {
		return null;
	}

	public Operation getCreatedOperation() {
		return (Operation) getResult().toArray()[0];
	}

	public interface DesignProviderI {
		/**
		 * Must not return null.
		 */
		String getDesign(Operation op);
	}

	/**
	 * Creates a new command that will set the design according to the current provider.
	 * The operation passed to the DesignProvider will be part of the model by the time it is called.
	 * This is a convenience method for legacy code. Setting the design directly through
	 * {@link #getDesign()} is more light-weight, but usually requires more preparation by the caller.
	 */
	public CompoundCommand createDesign(final DesignProviderI dp) {
		CompoundCommand cmp = new StrictCompoundCommand() {{ isPessimistic = true; }};
		cmp.append(this);
		cmp.append(new CommandWrapper() {
			@Override protected Command createCommand() { return new DesignsED(domain).create(getCreatedOperation(), dp.getDesign(getCreatedOperation())); };
		});
		return cmp;
	}
}
