package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Operation;

import edu.unu.iist.rcos.mm.IRCOSCommand;
import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.ChokeException;
import edu.unu.iist.rcos.spec.antlr.visitor.check.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation.*;

public class DesignDelegationCommand extends RCOSCommand<DesignDelegationParameter> implements IRCOSCommand {

	private String designWithChoke;
	private String targetOperationName;
	private String targetPath;
	private Operation op;
	
	public String getTargetOpName() {
		return targetOperationName;
	}
	
	public String getTargetPath() {
		return targetPath;
	}
	public Operation getOp() {
		return op;
	}
	
	public DesignDelegationCommand() {};
	public DesignDelegationCommand(EditingDomain domain, List<Throwable> errors, DesignDelegationParameter desPar) {
		super(domain,errors,desPar);
	}

	public static class SelectionInvalidException extends Exception {
		public SelectionInvalidException(Exception e) {
			super("Selection invalid",e);
		}

		public SelectionInvalidException(String msg) {
			super(msg);
		}

		public SelectionInvalidException(String msg, Exception e) {
			super(msg,e);
		}

		private static final long serialVersionUID = 1L;
	}

	/**
	 * @return The complement of the design of the operation and the part matched by CHOKE in dChoke.
	 * @throws WellformedException 
	 * @throws SelectionInvalidException 
	 */
	public static String getSelection(String fullDesign, String dChoke) throws WellformedException, SelectionInvalidException {
		int chokeStart = dChoke.indexOf("CHOKE");
		if (chokeStart == -1)
			throw new WellformedException("No CHOKE in command?!");
		String trailer = dChoke.substring(chokeStart+"CHOKE".length());
		System.err.println(trailer);
		final String res;
		if ("".equals(trailer))
			res = fullDesign.substring(chokeStart-1);
		else {
			int chokeEnd = fullDesign.indexOf(trailer);
			if (chokeEnd == -1)
				throw new SelectionInvalidException("Selection invalid; can't find trailer in design:\n[trailer]\n"+trailer+"*\n[design]\n"+fullDesign+"*");
			res = fullDesign.substring(chokeStart-1, chokeEnd);
		}
		return res;
	}

	/**
	 * Hooks for overloading in GUI.
	 * @throws InterruptedException 
	 */
	protected String getTargetPath(final Operation op, final VardeclsFIS ctx, DESIGN dSelection) throws InterruptedException { return params.getTargetPath(); };
	protected String getTargetOperationName() throws InterruptedException { return params.getTargetOperationName(); };

	public static class Context {

		final public VardeclsFIS ctx;
		public DESIGN designFull = null;
		final public CHOKE choke;

		public Context(String designWithChoke, Operation op) throws SelectionInvalidException {
			try {
				designFull = Parse.getDesignChoke(designWithChoke, op);
				/* We use an exception to dump the typing context here,
				 * without having to run the visitor to completion :-)
				 * This can only be the only CHOKE in the code, due to previously parsing 'd'.
				 * 
				 * But first, we check if the resulting design would be valid.
				 * We could make CHOKE record the state, and go on.  But we would
				 * have to make a deep copy of the context then.
				 */
				new DesignCheckType(op){
					@Override
					protected void visit(CHOKE d) throws ChokeException {
						// NADA
					}
				}.run(designFull);
				new DesignCheckType(op){
					@Override
					protected void visit(CHOKE d) throws ChokeException {
						throw new ChokeException(this.ev.vdecl,d);
					}
				}.run(designFull);
				throw new RuntimeException(); // NOT REACHED
			}
			/* We expect to CHOKE!
			 * The others either mean the remaining fragment would not be valid,
			 * or that the fragment you selected is inconsistent (e.g. if you included
			 * an ENDDECL without a matching VARDECL).
			 */
			catch (RecognitionException ex) {
				throw new SelectionInvalidException("Outer context invalid; can't parse.");
			} catch(WellformedException ex) {
				throw new RuntimeException("this should not have happened:" + designWithChoke,ex);
			} catch (ChokeException ex) {
				ctx = (VardeclsFIS) ex.vdecls; // Trust me.
				choke = (CHOKE) ex.token;
			} catch (GetTypeException ex) {
				throw new SelectionInvalidException("Outer context invalid.",ex);
			}
		}
	}

	@Override
	public Command createCommand() {
		if (params.getSourceOperation() == null)
			return recordError(new IllegalArgumentException("Parameters not configured correctly on "+params.getBase_Class().getQualifiedName()));
		try {
			/* We severely split the exceptions so that we can pinpoint what went wrong precisely. */			
			op = params.getSourceOperation().getOperation();
			final String t;
			DESIGN dSelection;
			try {
				final String fullDesign = new DesignsED(domain).get(op);
				t = getSelection(fullDesign, params.getDesignWithChoke());
			} catch (WellformedException e) {
				return recordError(new SelectionInvalidException(e));
			} catch (IndexOutOfBoundsException e) {
				return recordError(new SelectionInvalidException(e));
			} catch (SelectionInvalidException e) {
				return recordError(e);
			}

			try {
				designWithChoke = params.getDesignWithChoke();

				// first try to parse designs:
				dSelection = Parse.getDesign(t, op);
			} catch (WellformedException e) {
				return recordError(new SelectionInvalidException("Selection invalid in operation "+ op.getName()+".",e));
			} catch (RecognitionException e) {
				return recordError(new SelectionInvalidException("Selection invalid in operation "+ op.getName()+"; can't parse:\n"+t));
			}
			assert dSelection != null;

			/* Next, we try TCing without the chunk---it should still work out */
			{
				try {
					final Context context = new Context(designWithChoke,op);
					assert context.ctx != null;
					assert context.designFull != null;
					assert context.choke != null;

					final PREFIX choice;

					/* Reparse design with proper context. However, that will spoil the context...*/
					dSelection = Parse.getDesign(t, new VardeclsFIS(context.ctx));
					targetPath = getTargetPath(op, new VardeclsFIS(context.ctx), dSelection);
					final VardeclsFIS vctx = new VardeclsFIS(context.ctx);
					try {
						choice = (PREFIX) Parse.getExp(targetPath, vctx);
					} catch (RecognitionException e) {
						return recordError(new SelectionInvalidException("Target path invalid.",e));
					} catch (WellformedException e) {
						return recordError(new SelectionInvalidException("Target path invalid.",e));
					}

					/* Re-parse selection with obtained context */
					final DESIGN dReparsed;
					try {
						dReparsed = Parse.getDesign(t, vctx);
					} catch (RecognitionException e) {
						return recordError(new SelectionInvalidException("Target design invalid.",e));
					}

					/* Delegate. Calculate target method name, replace CHOKE with a new MCALL. */
					// we will destroy designFull by patching an MCALL into the place of the CHOKE
					DesignDelegator dd = new DesignDelegator(domain, errors, op, (REF) choice, context.ctx) {
						@Override
						protected Command log(Throwable e) {
							return recordError(e); // TODO: get rid of?
						}
					};
					// May call into GUI:
					return dd.delegateDesign(targetOperationName = getTargetOperationName(), context.choke, dReparsed, context.designFull);
				} catch (DelegationException e) {
					return recordError(e);
				} catch (InterruptedException e) {
					// interactively cancelled
					return recordError(e);
				} catch (SelectionInvalidException e) {
					return recordError(e);
				} catch (RecognitionException e) {
					// Shouldn't happen
					return recordError(e);
				}
			}
		} catch (OperationQException e) {
			return recordError(e);
		}
	}

	public Parameters setData(Class c) {
		c.setName(getLabel());
		DesignDelegationParameter p = Stereotypes.apply(DesignDelegationParameter.class, c);
		p.setDesignWithChoke(designWithChoke);
		p.setTargetOperationName(targetOperationName);
		p.setTargetPath(targetPath);
		OperationQ_Operation sop = ProfileFactory.eINSTANCE.createOperationQ_Operation();
		sop.setOperation(op); // TODO: Listener/Query
		p.setSourceOperation(sop);
		p.setMine(sop);
		return p;
	}
}