	package edu.unu.iist.rcos.spec.antlr.visitor;

import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.Classifier;

import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

public class ExpPrinter extends ExpVisitor<String,String, PrinterException, TypePrinter<PrinterException>> {
	
	final protected FieldIsShadowed fis;

	// TODO: can we override the literals in LogOps e.g. for the UTF8 printer?
	protected String AND = LogOps.AND.getLiteral();
	protected String OR = LogOps.OR.getLiteral();
	protected String IMP = LogOps.IMP.getLiteral();
	protected String EQUIV = LogOps.EQUIV.getLiteral();
	protected String EXISTS = "exists";
	protected String FORALL = "forall"; 
	protected String ELEM = "in"; 
	protected String NEG = "not";		

	public interface FieldIsShadowed {
		/**
		 * Decides whether the field 'fa' is currently shadowed by a local declaration
		 * or parameters.
		 * If you want to be on the safe side, always return 'true', see AlwaysShadowed below.
		 * That way, the pretty printer will prefix all field accesses with "this.". 
		 */
		boolean isShadowed(VARNAME fa) throws GetTypeException;
	}

	public static class AlwaysShadowed implements FieldIsShadowed {
		public boolean isShadowed(VARNAME fa) throws GetTypeException {
			return true;
		}
	}

	/**
	 * @deprecated Should provide FieldIsShadowed interface, which requires some form of context (Classifier or Operation).
	 */
	@Deprecated
	public ExpPrinter(){
		this(new AlwaysShadowed());
	}

	public ExpPrinter(Classifier c) {
		this(new VardeclsFIS(c));
	}

	public ExpPrinter(FieldIsShadowed fis){
		super(new TypePrinter<PrinterException>());
		assert fis != null;
		state = "";
		this.fis = fis;
	}

	@Override
	protected void visit(MCALL d) throws PrinterException{
		// Print m() instead of this.m():
		final String owner;
		if (!(d.getOwner() instanceof THIS))
			owner = run(d.getOwner())+".";
		else
			owner = "";

		String name = d.getName().getName();
		String in = getString(d.getIn());
		String out = "";
		if (d.getOut().size() > 0) {
			out = "; " + getString(d.getOut());
		}
		set(owner + name + "(" + in + out + ")");
	}

	private String getString(EList<?> l) throws PrinterException{
		boolean first = true;
		StringBuffer res = new StringBuffer("");
		for (Object o : l){
			if (first) {
				first = false;
			} else {
				res.append(", ");
			}
			res.append(run((EXP) o));
		}
		return res.toString();
	}

	@Override
	protected void visit(AOP d) throws PrinterException {
		// TODO: should use OO here, but the EMF codegen is in the way.
		String left = b(d.getLeft()) ;
		String sop = ((AOP)d).getOp().toString();
		String right = b(d.getRight());
		set(left + " " + sop + " " + right);
	}

	String b(EXP e) throws PrinterException {
		String s = run(e);
		if (e instanceof CONST || e instanceof NULL || e instanceof THIS || e instanceof VAR || e instanceof PAR || e instanceof FA || e instanceof MCALL || e instanceof NEW)
			return s;
		else
			return "("+s+")";
	}

	@Override
	protected void visit(LOGOP d) throws PrinterException {
		// TODO: should use OO here, but the EMF codegen is in the way.
		String left = b(d.getLeft()) ;
		String sop = null;
		final LogOps dOp = ((LOGOP)d).getOp();
		if (dOp.equals(LogOps.AND))
			sop = AND;
		else if (dOp.equals(LogOps.OR))
			sop = OR;
		else if (dOp.equals(LogOps.IMP))
			sop = IMP;
		else if (dOp.equals(LogOps.EQUIV))
			sop = EQUIV;
		else
			throw new RuntimeException("not reached");
		String right = b(d.getRight());
		set(left + " " + sop + " " + right);
	}

	@Override
	protected void visit(RELOP d) throws PrinterException {
		// TODO: should use OO here, but the EMF codegen is in the way.
		String left = b(d.getLeft()) ;
		String sop = ((RELOP)d).getOp().toString();
		String right = b(d.getRight());
		set(left + " " + sop + " " + right);
	}

	@Override
	protected void visit(INSTANCEOF d) throws PrinterException {
		set(run(d.getRef()) + " instanceof " + d.getClass_().getName());
	}

	@Override
	protected void visit(NEG d) throws PrinterException {
		set(NEG +  " " + b(d.getE()));
	}

	@Override
	protected void visit(FA d) throws PrinterException {
		String name = d.getName().getName();
		boolean notShadowed;
		try {
			notShadowed = d.getOwner() instanceof THIS && !fis.isShadowed(d.getName());
		} catch (GetTypeException e) {
			throw new PrinterException(e);
		}
		final String owner;
		if (notShadowed)
			owner = ""; // no shadowing, no 'this.' required.
		else
			owner = run(d.getOwner());
		String dash = d.isDash()?"\'":"";
		set((notShadowed ? "" : (owner + ".")) + name + dash);
	}

	@Override protected void visit(PAR p) throws PrinterException {
		String dash = p.isDash()?"\'":"";
		set(p.getName().getName() + dash);
	}

	@Override protected void visit(VAR v) throws PrinterException {
		String dash = v.isDash()?"\'":"";
		set(v.getName().getName() + dash);
	}

	@Override
	protected void visit(THIS d) throws PrinterException {
		set("this");
	}
	
	@Override
	protected void visit(SUPER d) throws PrinterException {
		set("super");
	}

	@Override
	protected void visit(NULL d) {
		set("null");
	}

	@Override
	protected void visit(NEW d) throws PrinterException {
		String name = tv.run(d.getName());
		String in = getString(d.getIn());
		String out = "";
		if (d.getOut().size() > 0) {
			out = "; " + getString(d.getOut());
		}
		set(name + ".new(" + in + out + ")");
	}

	@Override
	protected void visit(PREDICATE d) throws PrinterException {
		if (d instanceof ITERATOR) {
			visit((ITERATOR)d);
			return;
		}
		String quant = "";
		if (d instanceof FORALL)
			quant = FORALL;
		else if (d instanceof EXIST)
			quant = EXISTS;
		set(quant + " " + tv.run(d.getType()) + " " + d.getVar().getName() + ":" + run(d.getE()));
	}

	@Override
	protected void visit(ITERATOR d) throws PrinterException {
		String quant = "";
		quant = FORALL;
		set(quant + " " + tv.run(d.getType()) + " " + d.getVar().getName() + ":" + run(d.getRange()) + "|" + run(d.getE()));
	}
	
	@Override
	protected void visit(CAST d) throws PrinterException {
		set("(("+  tv.run(d.getType()) + ") "+ run(d.getVal()) + ")");
	}
	
	@Override
	protected void visit(CNAME c) {
		set(c.getName());
	}

	@Override
	protected void visit(VARNAME var) throws PrinterException {
		set(var.getName());
	}

	@Override
	protected void visit(BASEBOOL d) throws PrinterException {
		set(d.getVal());
	}

	@Override
	protected void visit(BASEINT d) throws PrinterException {
		set(d.getVal());
	}

	@Override
	protected void visit(BASESTRING d) throws PrinterException {
		set(d.getVal());
	}

	@Override
	protected void visit(BASEDOUBLE d) throws PrinterException {
		set(d.getVal());
	}

	@Override
	protected void visit(BASEENUM d) throws PrinterException {
		set(d.getVal());
	}
}