package edu.unu.iist.rcos.spec.antlr.visitor.expprefix;

import java.util.Collection;
import java.util.Iterator;

import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.DesignVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.TypeVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.ExpPrefixFilterVisitor.PrefixFilterSoftFail;

/**
 * This visitor weeds out dashed targets, and those that are not classes. This is mostly
 * because the only consumer at the moment is the Delegation-pattern. Maybe that's a bit
 * too much reuse...
 * 
 * @author stolz
 */
public class DesignPrefixFilterVisitor extends DesignVisitor<Collection<FA>, Collection<FA>, Void, PrefixFilterException, TypeVisitor<Void, PrefixFilterException>, ExpPrefixFilterVisitor> implements PrefixChoice {

	/* We're tricking a bit here. The Design visitor does not really have its own state,
	 * but only asks the Exp visitor. That's also why here you can first call
	 * run() AND THEN getChoice(), which you can't do in the Exp variant. Eeek!
	 * 
	 * A word of warning: this visitor will produce garbage for type-invalid input:
	 * it only handles scope frames ad hoc in FOR.
	 */

	public DesignPrefixFilterVisitor(RCOSModel rm, Operation op) {
		this(new VardeclsFIS(op, new TypeHandler(rm)));
	}

	public DesignPrefixFilterVisitor(final VardeclsFIS vdecl) {
		super(new ExpPrefixFilterVisitor(vdecl){
			@Override protected boolean check(FA fa) throws PrefixFilterException {
				if (fa.isDash())
					return false;
				Type tt;
				try {
					tt = vdecl.getType(fa);
				} catch (GetTypeException e) {
					throw new PrefixFilterException(e);
				}
				assert tt != null : fa;
				if (!(tt instanceof org.eclipse.uml2.uml.Class))
					return false;
				if (!((Class)tt).getTemplateBindings().isEmpty())
					return false; // Generic type
				return true;
			}
		});
	}


	@Override
	public void visit(PP d) throws PrefixFilterException {
		super.visit(d);
	}

	@Override
	public void visit(FOR d) throws PrefixFilterException {
		try {
			ev.vds.pushDeclFrame();
			ev.vds.new Vardecl(d);
			visit(d.getDo());
			ev.vds.popDeclFrame();
		} catch (GetTypeException e) {
			throw new PrefixFilterException(e);
		}
	}

	@Override
	public void visit(VARDECL d) throws PrefixFilterException {
		try {
			// TODO: should be a new frame for proper shadowing?
			ev.vds.new Vardecl(d);
		} catch (GetTypeException e) {
			throw new PrefixFilterException(e);
		}
	}

	@Override
	public void visit(ENDDECL d) throws PrefixFilterException {
		try {
			ev.vds.remove(d.getVar());
		} catch (GetTypeException e) {
			throw new PrefixFilterException(e);
		}
	}

	public static class ReturnPrefixFilterException extends PrefixFilterSoftFail {
		private static final long serialVersionUID = 1L;
		public ReturnPrefixFilterException() {
			super("Can't cover \'return\' statements.");
		}
	}

	@Override
	public void visit(RET d) throws PrefixFilterException {
		/* Can't cover RETURN. Well, we could, in the visitor,
		 * but then, what? TODO: Revisit later if we have
		 * refactoring actually working. Bail out:
		 */
		throw new ReturnPrefixFilterException();
	}

	public PREFIX getChoice() throws PrefixFilterException {
		return ev.getChoice();
	}

	public Collection<FA> getResult() throws PrefixFilterException {
		return ev.getResult();
	}

	/**
	 * @deprecated Only for testing
	 */
	@Deprecated
	public void print() throws PrefixFilterException {
		ev.print();
	}

	@Override
	protected void visit(CHOKE d) throws PrefixFilterException {
		// NADA
	}

	public static REF analyseTgt(VardeclsFIS vfis, DESIGN d) throws PrefixFilterException {
		/* For an assignment, the expert is obvious:
		 */
		if (d instanceof ASSIGN) {
			REF ch = ((ASSIGN)d).getVal();
			if (ch instanceof FA) {
				PREFIX o = ((FA) ch).getOwner();
				if (o instanceof DASHABLE)
					return (DASHABLE)o;
			}
		}
		/* We first try a quick-n-dirty hack and check if there's a single primed
		 * variable in it. Its owner is the obvious candidate in any case.
		 */
		{
			Collection<DASHABLE> choices = DesignUtility.getPrimedVariables(d);
			if (choices.size() == 1) {
				DASHABLE ch = choices.iterator().next();
				if (ch instanceof FA) {
					FA fa = (FA) ch;
					/* Is the root a variable defined OUTSIDE of the design?
					 * A variable introduced by a quantifier doesn't help us.
					 */
					PREFIX o = fa.getRootOwner();
					if ((!(o instanceof VAR) || (o instanceof VAR && vfis.isLocal(((VAR) o).getName())) && fa.getOwner() instanceof REF) && !(fa.getOwner() instanceof THIS)) {
						return (REF) fa.getOwner();
					}
				}
			}
		}
		/* Okay, the high road, start looking... */
		DesignPrefixFilterVisitor dpf = new DesignPrefixFilterVisitor(vfis);
		Collection<FA> tgt = dpf.run(d);
		PREFIX choice = dpf.getChoice();
		FA t;
		if (choice instanceof FA)
			t = (FA) choice;
		else { // No unique choice, so let's see if we have any alternatives
			// TODO: sort by size?
			t = null;
			final Iterator<FA> it = tgt.iterator();
			do {
				switch (tgt.size()) {
				case 0:
					return null;
				case 1:
					t = it.next();
					break;
				default:
					t = it.next();
					System.err.println("many targets, picking first: "+vfis.expToString(t)); // TODO
				}
			} while (t == null && it.hasNext());
		}
		return t;
	}
}
