package edu.unu.iist.rcos.spec.antlr.visitor.expprefix;

import java.util.*;
import org.eclipse.uml2.uml.Operation;

import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.SoftFail;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.TypeVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * This visitor only does syntactic filtering.
 *
 * @author stolz
 *
 */
public class ExpPrefixFilterVisitor extends ExpVisitor<Collection<FA>, Void, PrefixFilterException, TypeVisitor<Void, PrefixFilterException>> implements PrefixChoice {

	/* Actually the state should be a tuple/new type combining the FAs and the getChoice() computation. */

	/* TODO:
	 * - note that this.x can have two reasons: it's either the true FA, or it could be
	 *		from this.x.m(..), where we prune the method call.
	 * - find out who's doing set(null);
	 */

	/**
	 * Used to filter out definitions local to a larger fragment.
	 * The {@link DesignPrefixFilterVisitor} can push declarations onto here.
	 */
	final Vardecls vds;

	public ExpPrefixFilterVisitor(RCOSModel rm, Operation op) {
		this(new VardeclsFIS(op,rm));
	}

	@SuppressWarnings("serial")
	public static class PrefixFilterSoftFail extends PrefixFilterException implements SoftFail {
		public PrefixFilterSoftFail(String m) {
			super(m);
		}
	}

	@Override
	public void set(Collection<FA> v) {
		assert v != null;
		super.set(v);
	}

	public ExpPrefixFilterVisitor(Vardecls vardecls) {
		super(new TypeVisitor<Void, PrefixFilterException>());
		assert !init;
		vds = vardecls;
		set(new HashSet<FA>());
		init = true;
	}

	/**
	 * We'll return a REF since we know we're not doing CNAMEs at the moment.
	 */
	public REF getChoice() throws PrefixFilterException {
		if (getResult() == null || getResult().size() == 0)
			return null;
		final List<FA> in = new ArrayList<FA>(getResult());
		PREFIX out = in.remove(0);
		// find prefix of FAs:
		for (FA fa : in) {
			out = new PrefixVisitor(out).run(fa);
			if (out == null || out instanceof THIS)
				return null;
		}
		return (REF) out;
	}

	/**
	 * A poor man's check if two expressions have a common prefix
	 * @author stolz
	 */
	public static class PrefixVisitor extends ExpVisitor<PREFIX, Void, PrefixFilterException, TypeVisitor<Void, PrefixFilterException>> {

		final private PREFIX exp;

		public PrefixVisitor(PREFIX out) {
			super(null);
			exp = out;
		}

		@Override
		protected void visit(THIS d) {
			if (exp instanceof THIS)
				set(d);
		}

		@Override
		protected void visit(MCALL d) throws PrefixFilterException {
			PREFIX prefix1 = new PrefixVisitor(exp).run(d.getOwner());

			if (exp instanceof MCALL) {
				PREFIX prefix2 = new PrefixVisitor(((MCALL)exp).getOwner()).run(d);
				PREFIX prefix = new PrefixVisitor(((MCALL)exp).getOwner()).run(d.getOwner());
				// compare with the other solutions:
				PREFIX best1 = bestOf(prefix,prefix1);
				PREFIX best2 = bestOf(prefix,prefix2);
				set(bestOf(best1,best2));
				return;
			}
			set(prefix1);
		}

		@Override
		protected void visit(FA d) throws PrefixFilterException {
			// TODO: delay inline prefix1/2 into branches for efficiency
			PREFIX prefix1 = new PrefixVisitor(exp).run(d.getOwner());
			if (exp instanceof FA) {
				PREFIX prefix2 = new PrefixVisitor(((FA)exp).getOwner()).run(d);
				PREFIX prefix = new PrefixVisitor(((FA)exp).getOwner()).run(d.getOwner());
				if (prefix != null)
					// check for 'dashity'
//					if ((d.isDash() == ((FA)exp).isDash()) && d.getName().equals(((FA)exp).getName())) {
					if (d.getName().equals(((FA)exp).getName())) {
						set(d); return; // best solution!
					} else {
						// compare with the other solutions:
						PREFIX best1 = bestOf(prefix,prefix1);
						PREFIX best2 = bestOf(prefix,prefix2);
						set(bestOf(best1,best2));
						return;
					}
				assert prefix == null;
				/* We know the 'prefix' is useless. So let's find the longer of
				 * prefix1 and prefix2.
				 */
				if (prefix2 != null) {
					set(bestOf(prefix1,prefix2));
					return;
				}
			}
			set(prefix1);
		}

		@Override
		protected void visit(PAR p) {
			if (exp instanceof PAR && p.getName().equals(((PAR)exp).getName()))
				set(p);
			else
				set(null);
		}

		@Override
		protected void visit(INSTANCEOF p) throws PrefixFilterException {
			throw new PrefixFilterSoftFail("Refactoring not supported: INSTANCEOF");
		}

		@Override
		protected void visit(VAR p) throws PrefixFilterException {
			set(null);
		}

		@Override
		protected void visit(CAST p) throws PrefixFilterException {
			throw new PrefixFilterSoftFail("Refactoring not supported: CAST");
		}
		
		@Override
		protected void visit(NEW p) throws PrefixFilterException {
			throw new PrefixFilterSoftFail("Refactoring not supported: NEW");
		}

		@Override
		protected void visit(SUPER p) throws PrefixFilterException {
			throw new PrefixFilterSoftFail("Refactoring not supported: SUPER");
		}

		@Override
		protected void visit(CNAME c) throws PrefixFilterException {
			if (exp instanceof CNAME && c.equals((CNAME)exp))
				set(exp);
		}

		@Override
		protected void visit(VARNAME var) throws PrefixFilterException {
			if (exp instanceof VARNAME && var.equals((VARNAME)exp))
				set(exp);
		}

		@Override protected void visit(BASEBOOL d) throws PrefixFilterException { set(null); }
		@Override protected void visit(BASEINT d) throws PrefixFilterException { set(null); }
		@Override protected void visit(BASESTRING d) throws PrefixFilterException { set(null); }
		@Override protected void visit(BASEDOUBLE d) throws PrefixFilterException { set(null); }
		@Override protected void visit(BASEENUM d) throws PrefixFilterException { set(null); }
	}

	public static PREFIX bestOf(PREFIX prefix1, PREFIX prefix2) throws PrefixFilterException {
		if (prefix1 == null) {
			return prefix2;
		}
		if (prefix2 instanceof THIS) {
			assert prefix1 != null;
			return prefix1;
		} else {
			/* We calculate the shared prefix of both.
			 * If that prefix == prefix1, we know that
			 * prefix2 must be equal OR LONGER!
			 */
			PREFIX both = new PrefixVisitor(prefix1).run(prefix2);
			if (both == null) {
				return null;
			} else if (both instanceof THIS) {
				return both;
			} else {
				if (DesignUtility.isExpEqual(both, prefix1)) {
					return prefix2;
				} else {
					return prefix1;
				}
			}	
		}
	}

	protected static <T extends EXP> void print(Collection<T> cs) {
		for(T p : cs) {
			System.err.println( DesignUtility.expToString(p));
		}
	}

	protected void add(FA fa) throws PrefixFilterException {
		/* Bail out if the root is a local definition in the current fragment:
		 * only happens within larger designs.
		 * XXX We need this for the single-variant to work!
		 */
//		if (fa.getOwner() instanceof VAR && vds.isLocal(((VAR) fa.getOwner()).getName()))
//			return;
		if  (!check(fa)) // customized checks
			return;
		// check for dupes:
		for (FA a : getResult()) {
			if (DesignUtility.isExpEqual(a,fa))
				return;
		};
		getResult().add(fa);
	}

	/**
	 * Hook to plug in another validator, e.g. by checking types.
	 */
	protected boolean check(FA fa) throws PrefixFilterException {
		return true;
	}

	@Override
	protected void visit(FA fa) throws PrefixFilterException {
		if (fa.getOwner() instanceof DASHABLE || fa.getOwner() instanceof THIS)
			add(fa);
		// let's look at all prefixes as well:
		visit(fa.getOwner());
	}

	@Override
	protected void visit(PREDICATE p) throws PrefixFilterException {
		vds.pushDeclFrame();
		try {
			vds.new Vardecl(p);
			super.visit(p);
		} catch (GetTypeException e) {
			throw new PrefixFilterException(e);
		} finally {
			vds.popDeclFrame();
		}
	}

	/**
	 * We rely in getChoice() on not having any CNAMEs in the set!
	 */
	@Override protected void visit(CNAME c) throws PrefixFilterException {}
	@Override protected void visit(VARNAME var) throws PrefixFilterException {}
	@Override protected void visit(BASEBOOL d) throws PrefixFilterException {}
	@Override protected void visit(BASEINT d) throws PrefixFilterException {}
	@Override protected void visit(BASESTRING d) throws PrefixFilterException {}
	@Override protected void visit(BASEDOUBLE d) throws PrefixFilterException {}
	@Override protected void visit(BASEENUM d) throws PrefixFilterException {}

	/**
	 * @deprecated Only for testing
	 */
	@Deprecated
	void print() throws PrefixFilterException {
		print(getResult());
	}
}
