package edu.unu.iist.rcos.spec.antlr.visitor.requiredProtocol;

import java.util.*;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.ExpGetType;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

public class RequiredProtocolEXP extends ExpVisitor<State, Void, RequiredProtocolVisitorException, TypeVisitor<Void, RequiredProtocolVisitorException>> {

	private State last;
	private final TypeHandler th;
	private final Classifier cl;
	private final Collection<State> shortcircuit = new ArrayList<State>();

	public RequiredProtocolEXP(RCOSModel rm, Classifier cl) {
		super(new TypeVisitor<Void, RequiredProtocolVisitorException>());
		this.th = new TypeHandler(rm);
		this.cl = cl;
		
		this.last = UMLFactory.eINSTANCE.createState();
		set(last);
	}

	@Override protected void visit(final MCALL m) throws RequiredProtocolVisitorException {
		/* No need for recursive calls, strictly speaking:
		 * we know that INSIDE an mcall there can't be any more component calls.
		 * But we visit() nonetheless for completeness. 
		 */
		
		final EObject mContainer = m.eContainer();
		final EStructuralFeature mFeature = m.eContainingFeature();
		try {
			ExpGetType ev = new ExpGetType(cl, th) {public boolean voidOK() {
				return (
						mContainer == null
						|| (mContainer instanceof WHILE && mFeature == SyntaxTree_ClassModelPackage.eINSTANCE.getWHILE_Do())
						|| (mContainer instanceof FOR && mFeature == SyntaxTree_ClassModelPackage.eINSTANCE.getFOR_Do())
						|| (mContainer instanceof COMP && (mFeature == SyntaxTree_ClassModelPackage.eINSTANCE.getCOMP_Left() || mFeature == SyntaxTree_ClassModelPackage.eINSTANCE.getCOMP_Right()))
				);
			};};
			Operation op = ev.getOperationFromMcall(m);
			assert op != null : m.getOwner().toString() + m;
			if (m.getOwner() instanceof CNAME) {
				CNAME c = (CNAME) m.getOwner();
				State out = UMLFactory.eINSTANCE.createState();
				createTransition(op, c.getName()+"."+op.getName(), last, out);
				last = out;
				super.visit(m);
			} else {
				if (!op.isQuery())
					throw new RequiredProtocolVisitorException("Can't handle method call:"+m);
				else
					super.visit(m);
			}
		} catch (GetTypeException e) {
			throw new RequiredProtocolVisitorException(e);
		}
	}

	private Transition createTransition(Operation op, String name, State in, State out) {
		Transition t = UMLFactory.eINSTANCE.createTransition();
		t.setSource(in);
		t.setTarget(out);
		t.setName(name);
		Trigger tr = t.createTrigger("Trigger_"+t.getName()); // could do with less triggers.
		CallEvent ce = UMLFactory.eINSTANCE.createCallEvent();
		if (op != null)
			ce.setName(op.getName());
		else
			ce.setName("tau");
		ce.setOperation(op);
		tr.setEvent(ce);
		return t;
	}

	private Transition createTau(State in, State out) {
		return createTransition(null, "tau", in, out);
	}

	public class RequiredProtocolDesignVisitor extends DesignVisitor<State, State, Void, RequiredProtocolVisitorException, TypeVisitor<Void, RequiredProtocolVisitorException>, RequiredProtocolEXP> {

		public RequiredProtocolDesignVisitor() {
			super(RequiredProtocolEXP.this);
		}

		@Override public State run(DESIGN d) throws RequiredProtocolVisitorException {
			State res = super.run(d);
			for (State s : shortcircuit) {
				if (s != last)
					createTau(s, last);
			}
			return res;
		}
		@Override public State getResult() throws RequiredProtocolVisitorException {
			return ev.getResult();
		}

		@Override public void set(State s) {
			throw new UnsupportedOperationException();
		}

		@Override protected void visit(FOR f) throws RequiredProtocolVisitorException {
			State in = last;
			super.visit(f);
			State out = last;
			last = UMLFactory.eINSTANCE.createState();
			createTau(out, last);
			createTau(out, in);
			createTau(in, last);
		}
		
		@Override protected void visit(WHILE e) throws RequiredProtocolVisitorException {
			State in = last;
			super.visit(e);
			State out = last;
			last = UMLFactory.eINSTANCE.createState();
			createTau(out, last);
			createTau(out, in);
			createTau(in, last);
		}

		@Override protected void visit(IF ite) throws RequiredProtocolVisitorException {
			visit(ite.getB());
			State prev = last;
			visit(ite.getLeft());
			State outLeft = last;
			last = prev;
			visit(ite.getRight());
			State outRight = last;
			// merge branches into a new node:
			last = UMLFactory.eINSTANCE.createState();
			if (outLeft != prev)
				for (Transition t : outLeft.getIncomings()) {
					t.setTarget(last);
				}
			if (outRight != prev)
				for (Transition t : outRight.getIncomings()) {
					t.setTarget(last);
				}
		}

		@Override protected void visit(CHOICE c) throws RequiredProtocolVisitorException {
			State prev = last;
			visit(c.getLeft());
			State outLeft = last;
			last = prev;
			visit(c.getRight());
			State outRight = last;
			last = UMLFactory.eINSTANCE.createState();
			if (outLeft != prev)
				for (Transition t : outLeft.getIncomings()) {
					t.setTarget(last);
				}
			if (outRight != prev)
				for (Transition t : outRight.getIncomings()) {
					t.setTarget(last);
				}
		}

		@Override protected void visit(RET r) throws RequiredProtocolVisitorException {
			State prev = last;
			super.visit(r);
			shortcircuit.add(last);
			last = prev;
		}

		@Override protected void visit(PP pp) throws RequiredProtocolVisitorException {
			// Actually we know that there are no component calls in pre, but...
			visit(pp.getPre());
			// Allow to skip post in the protocol, in case the precondition failed:
			shortcircuit.add(last);
			visit(pp.getPost());
		}

		@Override protected void visit(MAGIC m) throws RequiredProtocolVisitorException {
			throw new RequiredProtocolVisitorException("Can't handle MAGIC");
		}
		
		@Override protected void visit(CHAOS c) throws RequiredProtocolVisitorException {
			throw new RequiredProtocolVisitorException("Can't handle CHAOS");
		}

		@Override protected void visit(CHOKE d) throws RequiredProtocolVisitorException {
			throw new RuntimeException();
		}
	}

	@Override protected void visit(VARNAME var) throws RequiredProtocolVisitorException {}
	@Override protected void visit(BASEBOOL d) throws RequiredProtocolVisitorException {}
	@Override protected void visit(BASEINT d) throws RequiredProtocolVisitorException {}
	@Override protected void visit(BASESTRING d) throws RequiredProtocolVisitorException {}
	@Override protected void visit(BASEDOUBLE d) throws RequiredProtocolVisitorException {}
	@Override protected void visit(BASEENUM d) throws RequiredProtocolVisitorException {}

	@Override protected void visit(CNAME c) throws RequiredProtocolVisitorException {}
}