package edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation;

import java.util.*;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.*;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.UMLPlugin;

import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.SoftFail;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.*;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.DesignPrefixFilterVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.ExpPrefixFilterVisitor.PrefixFilterSoftFail;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.PrefixFilterException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation.ExpDelegationVisitor.DelegationExceptionSoftFail;

/**
 * Find the first pre/post in an operation that can be successfully delegated.
 * 
 * TODO: Report "nothing to do" better...currently the IdentityCommand still
 * 		ends up on the command stack.
 * TODO: Avoid duplicates for, uhm, duplicate delegates. Should be a simply syntactic check.
 * TODO: Turn into RCOSCommand.
 * 
 * @author stolz
 */
public class DelegatePrePost {

	/* Accumulate the changes in the operation. As long as we only look at
	 * pre/posts, that'll be fine, since there can't be any nesting of a
	 * transformation target within a transformation target.
	 * But we still need to make things strict, since we introduce new operations
	 * into the model that need to be visible to subsequent (small) steps.
	 */
	final CompoundCommand result = new StrictCompoundCommand() {{isPessimistic = true;}};

	@SuppressWarnings("serial")
	static class Finished extends GetTypeException{
		public Finished(Exception t) { super("Couldn't transform Pre/Post.",t); UMLPlugin.INSTANCE.log(t);}
	}

	class RewritingCheckTypeVisitor extends DesignCheckType {
		public RewritingCheckTypeVisitor(Operation op) throws GetTypeException {
			super(op);
		}
	}

	public Command delegate(final EditingDomain ed, final List<Throwable> errors, final Operation op) throws RecognitionException, WellformedException, GetTypeException {
		final DESIGN d = Parse.getDesign(op);
		try {
			/* We call the visitor, but there isn't really a result.
			 * Either we apply the delegation and bail out, or we run
			 * through the visitor, knowing there'e nothing to do.
			 */
			new RewritingCheckTypeVisitor(op){
				private boolean canceled = false; /* For aborting interactive mode, the pending command have to be aborted. */
				@Override
				protected void visit(final PP pp) throws Finished {
					final VardeclsFIS vfis = (VardeclsFIS) ev.vdecl;
					final REF tgt;
					try {
						tgt = DesignPrefixFilterVisitor.analyseTgt(vfis,pp);
					} catch (PrefixFilterSoftFail e) {
						return;
					} catch (PrefixFilterException e) {
						throw new Finished(e);
					}
					if (tgt != null) {
						assert (!(tgt instanceof DASHABLE)) || (tgt instanceof DASHABLE && !((DASHABLE)tgt).isDash());
							final String tgtStr = vfis.expToString(tgt);
							/* Defer execution so that later commands can see the new operations
							 * introduced by visiting earlier statements.
							 */
							result.append(new CommandWrapper() {

								@Override
								protected Command createCommand() {
									try {
										if (canceled)
											return IdentityCommand.INSTANCE;
										DesignDelegator dd = new DesignDelegator(ed,errors,op,tgt,vfis) {
											@Override protected Command log(Throwable e) {
												/* we're in batch-mode, so don't stop */
												if (e instanceof SoftFail)
													return IdentityCommand.INSTANCE;
												errors.add(e);
												return UnexecutableCommand.INSTANCE;
											}
										};
										// TODO: check if tgtOp free here?
										return dd.delegateDesign(getTargetOperationName(tgtStr,pp), pp, pp, d);
									} catch (DelegationExceptionSoftFail e) {
										return IdentityCommand.INSTANCE; /* ! */
									} catch (DelegationException e) {
										errors.add(e);
										return UnexecutableCommand.INSTANCE;
									} catch (InterruptedException e) {
										errors.add(e);
										canceled = true;
										return IdentityCommand.INSTANCE;
									}
								}
							});
					} else {
						// nada: no target, nothing to do here, so proceed with visitor.
					}
				}
			}.run(d);
			result.append(IdentityCommand.INSTANCE);
			return new CommandWrapper("auto-delegate "+op.getQualifiedName(),null,result.unwrap());
		} catch (Finished e) {
			errors.add(e);
			return UnexecutableCommand.INSTANCE;
		}
	}

	/* GUI hook */
	protected String getTargetOperationName(String tgt, DESIGN d) throws InterruptedException {
		return tgt.replace('.', '_');
	}
}