package edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation;

import java.util.*;

import org.antlr.runtime.RecognitionException;
import org.apache.commons.collections.*;
import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.spec.antlr.*;
import edu.unu.iist.rcos.spec.antlr.Vardecls.VarType;
import edu.unu.iist.rcos.spec.antlr.commands.CreateOperationCommand;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation.ExpDelegationVisitor.DelegationExceptionSoftFail;

interface ArgTable {
	PAR lookupOrCreate(VAR v) throws GetTypeException;
	PAR lookupOrCreate(PAR v) throws GetTypeException;
	PAR lookupOrCreate(FA fa) throws GetTypeException, PrinterException;
	PAR lookupOrCreate(THIS owner) throws GetTypeException;
	Map<String,Arg> getTable();
}

/**
 * The main Design Delegator to (syntactically) create a delegation.
 * 
 * @author stolz
 */
public class DesignDelegator {

	class ArgTableImpl implements ArgTable {
		Map<String,Arg> table = new HashMap<String,Arg>();

		void insert(String r, Arg a) {
			Arg old = table.put(r, a);
			assert old == null;
		}

		public PAR lookupOrCreate(VAR v) throws GetTypeException {
			return lookupOrCreate(v,v.getName().getName(),vdcls.getType(v.getType())); // XXX getType()?!
		}

		public PAR lookupOrCreate(PAR v) throws GetTypeException {
			return lookupOrCreate(v,v.getName().getName(),vdcls.getType(v.getType()));
		}

		@SuppressWarnings("deprecation")
		public PAR lookupOrCreate(FA fa) throws GetTypeException, PrinterException {
			/* Hack: we render the FA as a string for the mapping.
			 * We have to consider whether it is dashed or not! For the
			 * string, we need to remove the dash first. The dash still goes
			 * into the argtable (because the FA does), so later when we generate the
			 * actual parameters, we need to nuke the dash as well, since it goes into
			 * delegated target.
			 */
			boolean wasDash = fa.isDash();
			fa.setDash(false);
			String key =  new ExpPrinter().run(fa);
			fa.setDash(wasDash);
			assert key.startsWith("this.");
			assert !key.contains("\'");
			return lookupOrCreate(fa,key,vdcls.getType(fa));
		}

		public PAR lookupOrCreate(THIS owner) {
			PAR r = lookupOrCreate(owner, "this", cl);
			r.getName().setName("that");
			return r;
		}

		PAR lookupOrCreate(REF r, String s, Type t) {
			assert t != null;
			Arg res = table.get(s);
			if (res == null) {
				res = new Arg(r,t);
				insert(s,res);
			} else {
				// NADA
			}
			PAR f = SyntaxTree_ClassModelFactory.eINSTANCE.createPAR(res.formal);
			// If the actual expression was dashed, also dash the placeholder.
			if (res.actual instanceof DASHABLE)
				f.setDash(((DASHABLE)res.actual).isDash()); 
			return f;
		}

		public Map<String, Arg> getTable() {
			return table;
		}
	}

	final ArgTable inTable = new ArgTableImpl();
	final ArgTable outTable = new ArgTableImpl();
	private final VardeclsFIS vdcls; /* context */
	private final REF tgt;
	private DesignDelegationVisitor dv;
	private final Classifier cl;
	private final EditingDomain ed;
	private final Operation op;
	protected Class dT;
	private final List<Throwable> errors;

	DesignDelegator(EditingDomain ed, Operation op, REF tgt, VardeclsFIS vdcls) throws DelegationException {
		this(ed, new ArrayList<Throwable>(), op, tgt, vdcls);
	}

	public DesignDelegator(EditingDomain ed, List<Throwable> errors, Operation op, REF tgt, VardeclsFIS vdcls) throws DelegationException {
		assert ed != null;
		assert op != null;
		assert tgt != null;
		assert vdcls != null;
		this.ed = ed;
		this.op = op;
		this.cl = (Classifier) op.getOwner();
		this.tgt = tgt;
		this.vdcls = vdcls;
		this.errors = errors;
		/* Get the static type of tgt, and save it for later. */
		try {
			final Type t = vdcls.getType(tgt);
			final String tgtStr = vdcls.expToString(tgt);
			if (!(t instanceof Class))
				throw new DelegationException("Expected a class for delegation target "+tgtStr+", but got: "+t.getQualifiedName());
			dT = dT != null ? dT : (Class) t;
		} catch (GetTypeException e) {
			throw new DelegationException(e);
		}
	}

	DESIGN run(DESIGN d) throws DelegationException {
		/* Only allow to run once */
		if (dv != null)
			throw new IllegalStateException();
		/* work */
		dv = new DesignDelegationVisitor(tgt, vdcls, inTable, outTable);
		return dv.run(d);
	}

	/** 
	 * @throws DelegationExceptionSoftFail if the name/signature combo is already taken either here
	 * or upwards in the inheritance tree.
	 */
	Command mkOperation(String name, final DESIGN d2) throws DelegationExceptionSoftFail {
		/* vdecls for pretty-printing the operation while it's not part of the model.
		 * Might need to be wrapped? */
		final VardeclsFIS ctx = new VardeclsFIS(dT);

		final EList<Type> types = new BasicEList<Type>();
		final EList<String> formalParamNames = new BasicEList<String>();
		for (Arg e : getInArgs()) {
			types.add(e.type);
			boolean unchanged = formalParamNames.add(e.formal);
			assert unchanged : e.formal;
			ctx.new Vardecl(e.formal, e.type, VarType.PAR);
		}
		for (Arg e : getOutArgs()) {
			types.add(e.type);
			boolean unchanged = formalParamNames.add(e.formal);
			assert unchanged : e.formal;
			ctx.new Vardecl(e.formal, e.type, VarType.PAR);
		}

		String suffix = "";
		{ /* Find free operation. */
			// TODO: how does this work with in/out parameters?!
			// TODO: avoid duplicates
			Operation opOut;
			int i = 0;
			do {
				opOut = dT.getOperation(name+suffix, null, types);
				if (opOut != null)
					suffix = Integer.toString(i++);
			} while (opOut != null);
		}

		return new CreateOperationCommand(ed, errors, name+suffix, dT){
			@Override protected EList<Type> getTypes() { return types;}
			@Override protected EList<String> getNames() { return formalParamNames; }
			@Override protected String getDesign() { return ctx.designToString(d2); }
			/* Override for GUI. Maybe we should just poke the message from getErrors() later....*/
			@Override protected Command recordError(Throwable t) { return DesignDelegator.this.log(t); }
			@Override public void execute() {
				super.execute();
				Operation op = getCreatedOperation();
				for (Arg e : getOutArgs()) {
					Parameter p = op.getOwnedParameter(e.formal, e.type);
					p.setDirection(ParameterDirectionKind.OUT_LITERAL);
				}
				try {
					Collection<DASHABLE> das = new GetPARVisitor().run(Parse.getDesign(op));
					for (final Parameter p : op.getOwnedParameters()) {
						assert (CollectionUtils.find(das, new Predicate(){
							public boolean evaluate(Object object) {
								DASHABLE da = (DASHABLE) object;
								/* If it's an FA, go "left" */
								if (da instanceof FA) {
									PREFIX root = ((FA)da).getRootOwner();
									if (root instanceof PAR)
										da = (DASHABLE) root;
									else
										return false;
								}
								if (da instanceof PAR)
									return ((PAR) da).getName().equals(p.getName());
								else
									return false;
							}}) != null)
							: "Created unused parameter: "+p.getName(); // SHOULDN'T HAPPEN!
					}
				} catch (VisitorException e1) {
					log(e1);
				} catch (RecognitionException e1) {
					log(e1);
				} catch (WellformedException e1) {
					log(e1);
				}
			}
		};
	}

	// TODO: Pattern?
	static class GetPARVisitor extends DesignVisitor<Collection<DASHABLE>,Collection<DASHABLE>, Void, VisitorException,TypeVisitor<Void, VisitorException>, ExpGetVariables> {
		protected GetPARVisitor() {
			super(new ExpGetVariables());
		}

		@Override protected void visit(CHOKE d) throws VisitorException {
			throw new RuntimeException(); // NOT REACHED
		}

		@Override public Collection<DASHABLE> getResult() throws VisitorException {
			return ev.getResult();
		}
	}

	/**
	 * Override those guys for error reporting and GUI interaction!
	 */
	protected Command log(Throwable t) {
		UMLPlugin.getPlugin().log(t);
		return UnexecutableCommand.INSTANCE;				
	}

	public Command delegateDesign(final String tgtName, final TOKEN choke, final DESIGN dSelection, final DESIGN dFull) {
		// we will destroy designFull by patching an MCALL into the place of the CHOKE
		assert choke != null;
		assert tgtName != null && !tgtName.equals("");
		assert dSelection != null;
		assert dFull != null;
		assert dT != null;

		return new CommandWrapper("delegate to "+dT.getQualifiedName()+"::"+tgtName) {
			@Override
			public Command createCommand() {
				try {
					DESIGN newD = run(dSelection);
					/* Remember that the command hasn't been executed, so you can't do op.getQualifiedName()! */
					CompoundCommand cmd = new StrictCompoundCommand() {{ isPessimistic = true; }};
					cmd.append(mkOperation(tgtName, newD));
					List<EXP> in = new ArrayList<EXP>();
					for (Arg e : getInArgs()) {
						// Don't put dashes into actual parameters. They go into the delegation.
						if ((e.actual instanceof DASHABLE) && ((DASHABLE)e.actual).isDash())
							((DASHABLE)e.actual).setDash(false);
						in.add(e.actual);
					}
					List<REF> out = new ArrayList<REF>();
					for (Arg e : getOutArgs()) {
						// Don't put dashes into actual parameters. They go into the delegation.
						if ((e.actual instanceof DASHABLE) && ((DASHABLE)e.actual).isDash())
							((DASHABLE)e.actual).setDash(false);
						out.add(e.actual);
					}

					MCALL mc = MyDesign.createMcall(tgt, tgtName, in, out);
					/* replace CHOKE with the MCALL by hacking the structure, unless we're at the root */
					final DESIGN dModify;
					if (choke.eContainer() == null)
						dModify = mc;
					else {
						choke.eContainer().eSet(choke.eContainingFeature(), mc);
						dModify = dFull;
					}
					cmd.append(update(op, vdcls.designToString(dModify)));
					return cmd;
				} catch (DelegationExceptionSoftFail e) {
					return log(e);
				} catch (DelegationException e) {
					return log(new DelegationException(e.getMessage()+"\nWhile processing "+op.getQualifiedName(), e));
				}
			}
		};
	}

	private Command update(Operation op, String d) {
		assert op != null;
		assert d != null;
		// will create a missing design
		return new DesignsED(ed).update(op, d);
	}

	private Collection<Arg> getInArgs() {
		return inTable.getTable().values();
	}

	private Collection<Arg> getOutArgs() {
		return outTable.getTable().values();
	}
}