package edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation;

import org.eclipse.emf.ecore.EObject;

import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.SoftFail;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpCopy;
import edu.unu.iist.rcos.spec.antlr.visitor.PrinterException;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.*;

public class ExpDelegationVisitor extends ExpCopy<DelegationException> {

	final ArgTable inTable;
	final ArgTable outTable;
	final VardeclsFIS vdcls;
	final REF tgt;

	/* XXX:
	 * Should push VAR,PAR onto vdcls?!
	 */
	public ExpDelegationVisitor(REF tgt, VardeclsFIS vdcls, ArgTable inTable, ArgTable outTable) {
		this.vdcls = vdcls;
		this.inTable = inTable;
		this.tgt = tgt;
		this.outTable = outTable;
	}

	@Override
	protected void visit(FA fa) throws DelegationException {
		/* If fa in prefix: this else find prefix-match.
		 * If prefix-match: rewrite to this.remainder.
		 */
		try {
			PREFIX prefix = ExpPrefixFilterVisitor.bestOf(fa, tgt);
			/* Don't use pointer equality here, as we want know if they are the same expression,
			 * but with different "dashity", so that we can bail out.
			 */
			if (DesignUtility.isExpWeakEqual(prefix,tgt)) {
				if (fa instanceof DASHABLE && ((DASHABLE)fa).isDash())
					throw new DelegationException("Can't delegate to dashed target: "+vdcls.expToString(fa)); // XXX wrong context?
				setThis(fa);
			} else {
				/* If we are at the 'head' of the term, we are looking
				 * at an attribute, so turn it into a parameter.
				 * This will e.g. delegate 'store.sales.m' to 'PAR x.sales.m'.
				 * If you want 'PAR x.m' with x = store.sales instead, here is the place
				 * to turn the screws.
				 * Also remember to pass dashed variables by passing the container
				 * ---in general there might be an extension to output parameters necessary,
				 * so we avoid this issue for now.
				 * If we are delegating to THIS, then of course we don't need to turn
				 * a field access into a parameter.
				 * XXX: This may look entirely different for predicates!
				 */
				if (fa.getOwner() instanceof THIS && !(tgt instanceof THIS)) {
					final EXP newFa;
					if (fa.isDash()) {
						PAR f = inTable.lookupOrCreate((THIS)fa.getOwner());
						// COPY the name!
						newFa = MyDesign.createFA(f, fa.getName().getName(), fa.isDash(), null);
					} else {
						newFa = inTable.lookupOrCreate(fa);
					}
					set(newFa);
				} else {			
					super.visit(fa); // let the visitor do the work!
				}
			}
		} catch (PrefixFilterException e) {
			throw new DelegationException(e);
		} catch (GetTypeException e) {
			throw new DelegationException(e);
		} catch (PrinterException e) {
			throw new DelegationException(e);
		}
	}

	@Override
	protected void visit(PAR v) throws DelegationException {
		/* All parameters are passed explicitly.
		 * Just be careful when our delegation target was actually a parameter!
		 */
		if (tgt instanceof PAR && ((PAR)tgt).getName().equals(v.getName())) {
			if (v.isDash())
				throw new DelegationException("Can't delegate to dashed target.");
			else
				setThis(v);
		} else {
			try {
				if (v.isDash()) {
					PAR arg;
					set(arg = outTable.lookupOrCreate(v));
					assert arg.isDash();
				} else
					set(inTable.lookupOrCreate(v));
			} catch (GetTypeException e) {
				throw new DelegationException(e);
			}
		}
	}

	@Override
	protected void visit(VAR v) throws DelegationException {
		/* Add v to parameters if coming from vdcls, unless it's the target.
		 * (local definition outside the selection). Then, use the new variable.
		 */
		if (vdcls.isLocal(v.getName())) {
			/* If the variable is dashed,
			 * we can't help you at the moment. We might wrap it somehow in the future.
			 * The eternal fight of CBV <-> CBN.
			 */
			try {
				if (tgt instanceof VAR && (((VAR)tgt).getName().equals(v.getName())))
					if (v.isDash())
						throw new DelegationException("Can't delegate to dashed target.");
					else
						setThis(v);
				else {
					if (v.isDash())
						set(outTable.lookupOrCreate(v)); // Can be 'this'?
					else
						set(inTable.lookupOrCreate(v)); // Can be 'this'?
					assert !(state instanceof THIS);
				}
			} catch (GetTypeException e) {
				throw new DelegationException(e);
			}
		} else
			super.visit(v);
	}

	private void setThis(DASHABLE fa) throws DelegationException {
		if (fa.isDash())
			throw new DelegationException("Can't delegate dashed target.");
		/* This is mostly the stopper which prevents turning 
		 *   [ sale != null |- .. ]
		 * into
		 *   [this != null |- ..]
		 * While not wrong per se, it means you probably should do some shuffling around.
		 */
		if (insidePreconditionOrInvariant(fa)) {
			String d = DesignUtility.dexpToString(vdcls, (DESIGNEXP) fa.eContainer());
			throw new DelegationExceptionSoftFail("Cowardly refusing to create an expression which does not make much sense from:\n"+d);
		}
		set(MyDesign.createThis());
	}

	private static boolean insidePreconditionOrInvariant(EXP d) {
		EObject c = d.eContainer();
		if (c == null || !(c instanceof DESIGNEXP))
			// XXX What about MCALL?
			return d instanceof EXP; // Guard or invariant
		else {
			if (c instanceof DESIGN)
				return c instanceof PP && d.eContainingFeature() == SyntaxTree_ClassModelPackage.eINSTANCE.getPP_Pre();
			else
				return !(c instanceof FA) && (c instanceof EXP && insidePreconditionOrInvariant((EXP)c));
		}
	}

	@SuppressWarnings("serial")
	public static class DelegationExceptionSoftFail extends DelegationException implements SoftFail {
		public DelegationExceptionSoftFail(String m) {
			super(m);
		}
	}
}