package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;

import java.util.*;
import java.util.Map.Entry;

import org.eclipse.uml2.uml.Classifier;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.ExpSubstitutePredicate;

/**
 * @author Charles Morisset
 *  Ensure that each pre-condition is either simple or a quantifier
 *  Transforms [ e1 * e2 * ... * en |- true ] into    (where * stands for any logical connector)
 *   var boolean b_0; b_0 := false; [ Pred0 |- b_0' = true];
 *   var boolean b_1; b_1 := false; [ Pred1 |- b_1' = true];
 *    ...
 *   [j1 * j2 * .... * jn |- true] (where ji = ei if ei is simple or ji = b_k if ei = Predk])
 *   end b_n; ...; end b_1; end b_0
 *  
 *  The operation is similar for conditional statements
 */
public class DesignExtractPredicate extends DesignRefinement<RefinementException, ExpSubstitutePredicate> {

	public DesignExtractPredicate(Classifier cl) {
		super(new ExpSubstitutePredicate(), cl);
	}

	@Override
	protected void visit(PP d) throws RefinementException { 
		assert(currentClass  != null);
		ev.setFng(fng);

		if (! isTrue(d.getPost()) || isSimpleExpression(d.getPre()) || d.getPre() instanceof PREDICATE) {
			set(copyDesign(d));
			return;
		}

		// We need to first clear the map
		ev.reset();
		EXP pre = copyExpression(d.getPre());
		EXP new_pre = ev.run(pre);

		// new_pre is pre where each predicate has been substituted by a fresh boolean variable
		// Now we have to: declare these variables, initialise them to false and create the corresponding PP
		
		Map<String, PREDICATE> m = ev.getListNewVariables();
		if (m.size() == 0) {
			set(copyDesign(d));
			return;
		}
		
		DESIGN body = createSkip();

		for (Entry<String,PREDICATE> entry : m.entrySet()) {
			final String name = entry.getKey();
			VAR ok = createVar(name, createBoolean()); 
			VARDECL vd = createVarDecl(createBoolean(), name);
			ASSIGN a = createAssign(ok, createFalse());

			VAR ok2 = (VAR) copyExpression(ok);
			ok2.setDash(true);
			RELOP rel = createRelOp(RelOps.EQ, ok2, createTrue()); // ok' = true 

			PP pp1 = createPP(entry.getValue(), rel);
			body = createSeq(vd, createSeq(a, createSeq(pp1, body)));
		}
		
		// Now we can append new_exp and then undeclare all the variables
		body = createSeq(body, createPP(new_pre, createTrue()));
		
		for (String name : m.keySet()) {
			body = createSeq(body, createEndDecl(name));
		}

		setModified(body);
	}

	@Override
	protected void visit(IF d) throws RefinementException { 
		assert(currentClass  != null);
		ev.setFng(fng);

		if (! isSimpleExpression(d.getB()) || d.getB() instanceof PREDICATE) {
			set(copyDesign(d));
			return;
		}

		// We need to first clear the map
		ev.reset();
		EXP pre = copyExpression(d.getB());
		EXP new_pre = ev.run(pre);

		// new_pre is pre where each predicate has been substituted by a fresh boolean variable
		// Now we have to: declare these variables, initialise them to false and create the corresponding PP
		
		Map<String, PREDICATE> m = ev.getListNewVariables();
		if (m.size() == 0) {
			set(copyDesign(d));
			return;
		}
		
		DESIGN body = createSkip();
		
		for (Entry<String,PREDICATE> entry : m.entrySet()) {
			final String name = entry.getKey();
			VAR ok = createVar(name, createBoolean()); 
			VARDECL vd = createVarDecl(createBoolean(), name);
			ASSIGN a = createAssign(ok, createFalse());

			VAR ok2 = (VAR) copyExpression(ok);
			ok2.setDash(true);
			RELOP rel = createRelOp(RelOps.EQ, ok2, createTrue()); // ok' = true 

			PP pp1 = createPP(entry.getValue(), rel);
			body = createSeq(vd, createSeq(a, createSeq(pp1, body)));
		}
		
		// Now we can append new_exp and then undeclare all the variables
		body = createSeq(body, createPP(new_pre, createTrue()));
		
		for (String name : m.keySet()) {
			body = createSeq(body, createEndDecl(name));
		}

		setModified(body);
	}
}
