package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;

import java.util.List;

import org.eclipse.uml2.uml.Classifier;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * @author Charles Morisset
 *  Transform: 
 *  "[e |- true]" into "if (e) then skip else chaos" if e doesn't contain any forall of exist
 *  
 *  "[\forall T x: o.contains(x) => P |- true]" into 
 *      "for (T x : o ) { body }" where body = run([P |- true])
 *       
 *  "[\exist T x: o.contains(x) AND P |- true]" into 
 *      "var boolean b; b := false; for (T x) { body }; if (b) then skip else chaos" where body run([P |- b' = true])
 *      
 */
public class DesignRefineAssert extends DesignRefinement<RefinementException, ExpRefinement<RefinementException>> {

	public DesignRefineAssert(Classifier cl) {
		super(new ExpRefinement<RefinementException>(), cl);
	}

	protected void visit(PP d) {
		if (isTrue(d.getPost())) {
			IF i = createIf(copyExpression(d.getPre()), createSkip(), createChaos());
			set(i);
		} else {
			set(copyDesign(d));
		}
	}
	
	protected void visit(IF d) throws RefinementException { 
		EXP b = copyExpression(d.getB());
		if (hasPrimedVariables(b)) {
			throw new RefinementException("The condition of '" + designToString(d) + "' contains primed variables");
		}
		
		if (isSimpleExpression(b)) {
			set(copyDesign(d));
			return;
		} else if (b instanceof PREDICATE) {
			String name = fng.getBoolean();
			DESIGN d1; 
			if (b instanceof EXIST)
				d1 = getForExistPre((EXIST) b, name);
			else
				d1 = getForForallPre((FORALL) b, name);
			if (d1 != null) {
				VARDECL vd = createVarDecl(createBoolean(), name);
				VAR ok = createVar(name, createBoolean());
				IF i = createIf(ok, copyDesign(d.getLeft()), copyDesign(d.getRight()));
				ENDDECL ed = createEndDecl(name);
				setModified(createSeq(vd, createSeq(d1, createSeq(i, ed))));
				return;
			} else {
				set(copyDesign(d));
				return;
			}
		}  else {
			set(copyDesign(d));
			return;
		}
	}

	/**
	 * @author Charles Morisset
	 * @param p a predicate "forall T x : o.contains(x) => P"
	 * @param name the name of the variable standing for the predicate
	 * @return null if p is not of the correct form, "name := true; for (T x : o) { if (P) then skip else name := false}" otherwise
	 * @throws RefinementException
	 */
	private DESIGN getForForallPre(FORALL p, String name) throws RefinementException {
		assert(currentClass != null);
		EXP e = copyExpression(p.getE());
		TYPE t = ev.run(p.getType());

		List<EXP> l_imp = split(e, LogOps.IMP);

		MCALL mc;
		final String pName = p.getVar().getName();
		try {
			mc = extractContains(l_imp, pName, t);
		} catch (GetTypeException e1) {
			throw new RefinementException(e1);
		}
		// If there is no "o.contains(x)", we return null
		if (mc == null) 
			return null;

		// If the "o.contains(x)" is the last element, we can't rewrite it, so we return null
		if (l_imp.indexOf(mc) == l_imp.size() - 1)
			return null;

		l_imp.remove(mc);

		// Variables "name" 
		VAR ok = createVar(name, createBoolean()); 
		VAR ok2 = (VAR) ev.run(ok);

		ASSIGN a = createAssign(ok, createFalse());  // name := false
		ASSIGN a2 = createAssign(ok2, createTrue()); // name := true

		// creating "if (P) then skip else name := false"
		DESIGN body = createIf(merge(l_imp, LogOps.IMP, createTrue()), createSkip(), a);

		// The for loop
		FOR f = MyDesign.createFor(ev.run(p.getType()), pName, (VAL) ev.run(mc.getOwner()), body);

		return createSeq(a2, f);
	} 

	/**
	 * @author Charles Morisset
	 * @param p a predicate "exist T x : o.contains(x) AND P"
	 * @param name the name of the variable standing for the predicate
	 * @return null if p is not of the correct form, "name := false; for (T x : o) { if (P) then name := true else skip }" otherwise
	 * @throws RefinementException
	 */
	private DESIGN getForExistPre(EXIST p, String name) throws RefinementException {
		assert(currentClass != null);

		List<EXP> l_and = split(ev.run(p.getE()), LogOps.AND);
		MCALL mc;
		final String pName = p.getVar().getName();
		try {
			mc = extractContains(l_and, pName, ev.run(p.getType()));
		} catch (GetTypeException e) {
			throw new RefinementException(e);
		}

		// If there is no "o.contains(x)", we return null
		if (mc == null) return null;

		l_and.remove(mc);

		// Variables "name"
		VAR ok = createVar(name, createBoolean()); 
		VAR ok2 = (VAR) ev.run(ok);

		ASSIGN a = createAssign(ok, createFalse());  // name := false
		ASSIGN a2 = createAssign(ok2, createTrue()); // name := true

		// Creating "if (P) then name := true else skip"
		DESIGN body = createIf(merge(l_and, LogOps.AND, createTrue()), a2, createSkip());

		// The for loop
		FOR f = MyDesign.createFor(ev.run(p.getType()), pName, (VAL) ev.run(mc.getOwner()), body);

		return createSeq(a, f);
	}
}
