package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.DesignGenCopy;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * Generic design visitor for refinement.
 * @author Charles Morisset
 */
public abstract class DesignRefinement<RE extends RefinementException, ER extends ExpRefinement<RE>> extends DesignGenCopy<RE, ER> {

	// Indicates if the design has been modified
	protected boolean modified; 
	final protected Classifier currentClass;
	protected FreshNameGenerator fng;
	final protected GetType gt;

	interface GetType {
		Type getType(EXP e) throws GetTypeException;
		PrimitiveTypes getPrimitiveTypes();
		boolean isBooleanExp(EXP e) throws GetTypeException;
	}

	@Override
	public void visit(DESIGN d) throws RE {
		if (! init && (currentClass != null))
			fng = new FreshNameGenerator(gt.getPrimitiveTypes(), currentClass, d);
		super.visit(d);
	}
	
	/**
	 * @param cl The classifier containing the design, needed for type checking. If set to null, trying 
	 * to rewrite a forall, exist or a method call may lead to throw GetTypeException
	 */
	public DesignRefinement(ER e, final Classifier cl) {
		super(e);
		if (cl == null)
			throw new IllegalArgumentException();
		modified = false;
		currentClass = cl;
		gt = new GetType() {
			TypeHandler th = new TypeHandler(cl);
			public Type getType(EXP e) throws GetTypeException { return th.getType(e, cl); }
			public PrimitiveTypes getPrimitiveTypes() { return th; }
			public boolean isBooleanExp(EXP e) throws GetTypeException { return th.checkBoolean(e, cl); }
		};
	}	

	/**
	 * @deprecated Try to pass a classifier, please.
	 * @see DesignRefinement#DesignRefinement(ExpRefinement, Classifier)
	 */
	@Deprecated
	public DesignRefinement(ER e) {
		super(e);
		modified = false;
		currentClass = null;
		gt = new GetType() {
			// TODO: I wonder how this is working...
			public Type getType(EXP e) throws GetTypeException { return DesignUtility.getTypeExp(e, null); }
			public PrimitiveTypes getPrimitiveTypes() { return null; }
			public boolean isBooleanExp(EXP e) throws GetTypeException {
				// TODO Auto-generated method stub
				return false;
			}
		};
	}

	private void setModified() {
		modified = true;
	}
	
	protected void setModified(DESIGN d) {
		modified = true;
		set(d);
	}
	
	public void unsetModified() {
		modified = false;
	}

	public boolean isModified() {
		return modified;
	}
	
	protected void visit(EXP e) throws RE {
		ev.visit(e);
		if (ev.isModified()) setModified();
	}
	
	/**
	 * @author Charles Morisset
	 * @param l a list of expression [e1, e2, ..., en]
	 * @param str LogOp
	 * @param def is the default expression (when the list is empty)
	 * @return e1 str (e2 str ... str en), null if some expressions cannot be copied
	 * @throws RE 
	 */
	EXP merge(List<EXP> l, LogOps str, EXP def) throws RE {
		if (l.isEmpty())
			return def;
		else {
			int size = l.size();
			EXP ret = ev.run(l.get(size - 1 ));
			for (int i = size - 2; i >= 0 ; i--) {
				ret = createLogOp(str, ev.run(l.get(i)), ret);
			}
			return ret;
		}
	}

	EXP mergeAnd(List<EXP> l) throws RE {
		return merge(l, LogOps.AND, createTrue());
	}
	
	EXP mergeOr(List<EXP> l) throws RE {
		return merge(l, LogOps.OR, createFalse());
	}
	
	/**
	 * @author Charles Morisset
	 * @param l a list [e1; e2; ... ; en]
	 * @param varname 
	 * @param type 
	 * @return ei  = o.contains(VAR(varname, type)) if it exists, null otherwise
	 * @throws GetTypeException 
	 */
	MCALL extractContains(List<EXP> l, String varname, TYPE type) throws GetTypeException {
		for (EXP e : l) {
			if (e instanceof MCALL) {
				MCALL m = (MCALL) e;
				VAR v = createVar(varname, type);
				if (Utility.isSetType(gt.getType(m.getOwner()))
						&& m.getName().equals("contains") // XXX string literal
						&& DesignUtility.isExpEqual(m.getIn().get(0), v)) {
					return m;
				}
			}
		}
		return null;
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression under the form e1 str e2 str .... str en
	 * @param str : "AND", "IMP", etc
	 * @return [e1, e2, ... , en]
	 * @throws VisitorException 
	 */
	List<EXP> split(EXP e, LogOps str)  {
		if (e instanceof LOGOP && ((LOGOP)e).getOp().equals(str)) {
			LOGOP lop= (LOGOP) e;
			List<EXP> l1 = split(lop.getLeft(), str);
			l1.addAll(split(lop.getRight(), str));
			return l1;
		} else {
			List<EXP> l = new ArrayList<EXP>();
			l.add(e);
			return l;
		}
	}
}
