package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Type;


import edu.unu.iist.rcos.mm.util.WellformedException;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * @author Saeed
 *  Transforms the design 
 *  [true |- r_1 && r_2 && ... && r_n], where each r_i is a RELOP 
 *  into
 *  [true |- p_k1 ; p_k2 ; ... ; p_kn], where each p_ki represents an r_i
 *  
 *   You need to create an instance of the class and then called postToSEQ.
 *   ! If you want to call postToSEQ for the second time, you must call resetVars before the second call
 *   ! First make a copy of the expression that you want to translate to sequential compositions, then 
 *     call postToSEQ(with the copy you have made).    
 *   
 *   By calling isSuccess it is possible to check weather the input is transformed to sequential compositions.
 *   Sometimes it is not possible, such as: [ |- x'>5 and x'<7] (it needs some theorem proving techniques)  
 */

public class EXPToComposition extends ExpRefinement<RefinementException> {

	private boolean success = true; 
	private List<RELOP> exprsDefined = new ArrayList<RELOP>();// the predicates that all its primed variables can be calculated
	private List<RELOP> exprsUndefined = new ArrayList<RELOP>();// the predicates that NOT all its primed variables can be calculated
	private List<DASHABLE> defPrimedVars = new ArrayList<DASHABLE>();
	private List<VARDECL>  varsDecl= new ArrayList<VARDECL>();
	private List<ENDDECL>  varsEnd= new ArrayList<ENDDECL>();
	private Classifier cl; 
	private TypeHandler th;

	public EXPToComposition(Classifier aCl) {
		super();
		assert aCl != null;
		cl = aCl;
		th = new TypeHandler(aCl);
	}

	@Deprecated
	public EXPToComposition(){
		super();
		cl = null;
	}

	//set the classifier
	/**
	 * @deprecated pass into constructor
	 */
	@Deprecated
	public void setCl(Classifier cl) {
		this.cl = cl;
	}

	public boolean isSuccess() {
		return success;
	}

	 // Traversing the tree using the visitor pattern
	protected void visit(LOGOP lop) throws RefinementException {
		if (! lop.getOp().equals(LogOps.AND)) {
			success = false;
			return;
		}

		EXP new_right = lop.getRight();
		EXP new_left = lop.getLeft();
		if ( (new_left instanceof LOGOP) && ( ((LOGOP) new_left).getOp().equals(LogOps.AND) ) ){
			run(new_left);
		}else if (new_left instanceof RELOP){
			if (unsupportedEXP(new_left)){
				success = false;
				return;				
			}
			addRelOp((RELOP) new_left);
		}else{
			success = false;
			return;
		}

		if ( (new_right instanceof LOGOP) && ( ((LOGOP) new_right).getOp().equals(LogOps.AND) ) ){
			run(new_right);
		}else if (new_right instanceof RELOP){
			if (unsupportedEXP(new_right)){
				success = false;
				return;				
			}			
			addRelOp((RELOP) new_right);
		}else{
			success = false;
			return;
		}
	}

	private boolean unsupportedEXP(EXP e) {
		if (e instanceof BINEXP) 
			return unsupportedEXP(((BINEXP) e).getLeft()) || unsupportedEXP(((BINEXP) e).getRight());
		if (e instanceof INSTANCEOF) return unsupportedEXP(((INSTANCEOF) e).getRef());
		if (e instanceof NEG) return unsupportedEXP(((NEG) e).getE());
		if (e instanceof CAST) return unsupportedEXP(((CAST) e).getVal());
		if (e instanceof FA) return unsupportedEXP(((FA) e).getOwner());
		if ((e instanceof MCALL)||(e instanceof PREDICATE)||(e instanceof NEW)) return true;
		return false;
	}

	// Add each RELOP in AST to the arrays exprsDefined or exprsUndefined
	private boolean addRelOp(RELOP rp) throws RefinementException{  
		boolean addToDef = allVarsDefined(rp);
		if (addToDef) {
			exprsDefined.add(rp);
		} else {
			if ( exprsUndefined.indexOf(rp)== -1 ) exprsUndefined.add(rp);
		}
		return addToDef;
	}

	//If all primed variables, for the input, can be calculated return true otherwise false
	private boolean allVarsDefined(RELOP rp) throws RefinementException{ 
		List<DASHABLE> primedVars = new ArrayList<DASHABLE>(getPrimedVariables(rp));
		if (primedVars.size()==0)	return true;
		int i=0;//No. of undefined primed variables
		int j = -1;//index of undefined primed variable (the last one)
		for (int k=0;k<primedVars.size() && i<=1 ;k++) {
			if (! isVarDefined( primedVars.get(k)) ) {
				i++;
				j = k;
			}//if (! isVarDefined((VAR) r))
		}//for (REF r : primedVars)
		if (i==1){
			if (defPrimedVars.indexOf(primedVars.get(j)) == -1){
				defPrimedVars.add(primedVars.get(j));
				return true;
			}else{
				success = false;
			}
		}
		//i != 1 or a primed variable is defined two time
		return false;
	}

	//check if the value of a primed variable is already calculated, if this is the case it must be in the defPrimedVars arrays
	private boolean isVarDefined(DASHABLE dv) {  
		if (dv instanceof VAR){
			VAR v = (VAR) dv;
			for (DASHABLE d : defPrimedVars){
				if ( (d instanceof VAR)&&( ( (VAR) d ).getName().getName().equals(v.getName().getName())) ) {
					return true;
				}
			}//for
		}

		if (dv instanceof PAR){
			PAR v = (PAR) dv;			
			for (DASHABLE d : defPrimedVars){
				if ( (d instanceof PAR)&&( ( (PAR) d ).getName().getName().equals(v.getName().getName())) ) {
					return true;
				}
			}//for
		}		

		if (dv instanceof FA){
			FA v = (FA) dv;
			for (DASHABLE d : defPrimedVars){
				if ( (d instanceof FA)&&( ( (FA) d ).getName().getName().equals(v.getName().getName())) ) {
					return ( strPrefix((FA)d).equals(strPrefix(v)) );
				}
			}//for
		}

		return false;
	}

	private String strPrefix(FA v) {
		String s =  expToString(v);
		if (s.charAt(s.length()-1) == '\'')  s = s.substring(0, s.length()-1);
		return s;
	}

	//removes the undefined expressions from exprsUndefined when it is possible to calculate all of its primed variables
	private void clearExprsUndefined() throws RefinementException{  
		int oldSize = defPrimedVars.size();
		int i=0;
		while (i<= exprsUndefined.size()-1){
			RELOP be = exprsUndefined.get(i);
			if (addRelOp(be)) exprsUndefined.remove(be); else i++;
		}//for
		if ( (oldSize < defPrimedVars.size()) && (exprsUndefined.size()>0) ) clearExprsUndefined();
	}

	public void resetVars(){
		varsDecl.clear();
		exprsDefined.clear();
		varsEnd.clear();
		exprsUndefined.clear();
		defPrimedVars.clear();
		success = true;
	}

	// The public method called from outside, the input is the postcondition in the form of disjunctions of predicates, 
	//and will be translatede to sequential compositions of designs 
	public DESIGN postToSEQ(LOGOP  post) throws RefinementException, GetTypeException, WellformedException{// the main function should be called from outside
		if (cl == null) System.out.println("EXPTOComposition: Classifier (cl) is null, may cause error");
		DESIGN s;
		if (composableConjuctions(post)){
			s = createSeq(createSkip(),createSkip());
			for (VARDECL vd : varsDecl){
				s = createSeq(s, vd);
			}
			for (BINEXP rop : exprsDefined){
				PP pp = createPP(createTrue(), rop);
				s = createSeq(s, pp);
			}
			for (ENDDECL ed : varsEnd){
				s = createSeq(s, ed);
			}
		} else {
			s = createPP(createTrue(), post);
		}
		try {
			s = simplifyDesign(simplifyDesign(s));
		} catch (VisitorException e) {
			throw new RefinementException(e);
		}

		return s;
	}

    //First traverses the AST by calling visit, then moves all expressions from exprsUndefined to exprsDefined 
	//and finally renames variables as needed
	private boolean composableConjuctions(LOGOP lop) throws RefinementException, GetTypeException, WellformedException{//Traverse the AST by calling "visit" then move all expressions from the exprsUndefined array to the exprDefined array by calling "clearExprsUndefined", and finally rename the variables as needed   
		visit(lop);
		clearExprsUndefined();
		if (exprsUndefined.size()>0) { 
			success = false; 
		}
		if (!success) System.out.println("EXPTOComposition: No change, the original design is returned");
		else {
			//Do renaming of variables
			List<RELOP>  exprsAssign= new ArrayList<RELOP>();						
			defPrimedVars.clear();
			List<DASHABLE> tmpDefPrimedVars = new ArrayList<DASHABLE>();
			List<String> strVarDelc = new ArrayList<String>();  
			for (RELOP be : exprsDefined){
				Collection<DASHABLE> allVars = getVariables(be);
				for (DASHABLE v : allVars){
					if (v.isDash()){
						if (!isVarDefined(v))  tmpDefPrimedVars.add(v);
						else v.setDash(false);
					}else{
						if (isVarDefined(v)){
							if (v instanceof VAR){
								String s = ((VAR) v).getName().getName();
								((VAR) v).getName().setName( s+"_b"  ); // XXX Check if name is actually free
								if (strVarDelc.indexOf(s+"_b")==-1) { 
									VAR x_b = createVar(createVarname(s+"_b"),((VAR) copyExpression(v)).getType(), true); 	
									VAR x = createVar(createVarname(s),((VAR) copyExpression(v)).getType(), false);
									exprsAssign.add(createRelOp(RelOps.EQ, x_b, x));//add x_b' = x							
									strVarDelc.add( s+"_b");
									varsDecl.add(createVarDecl(((VAR) copyExpression(v)).getType(), createVarname(new String( x_b.getName().getName()))));
									varsEnd.add(createEndDecl(createVarname(new String( x_b.getName().getName()))));
								}
							} else
								if (v instanceof PAR){
									String s = ((PAR) v).getName().getName();													 	
									VAR x_b2 = createVar(createVarname(s+"_b"),((PAR) copyExpression(v)).getType(), false);
									replaceDASHABLE(be,(PAR)v,x_b2);
									if (strVarDelc.indexOf(s+"_b")==-1) {
										VAR x_b = createVar(createVarname(s+"_b"),((PAR) copyExpression(v)).getType(), true);
										PAR x = createPar(createVarname(s),((PAR) copyExpression(v)).getType(), false);
										exprsAssign.add(createRelOp(RelOps.EQ, x_b, x));//add x_b' = x
										strVarDelc.add( s+"_b");
										varsDecl.add(createVarDecl(((PAR) copyExpression(v)).getType(), createVarname(new String( x_b.getName().getName()))));
										varsEnd.add(createEndDecl(createVarname(new String( x_b.getName().getName()))));
									}
								} else
									if (v instanceof FA){
										FA fa = (FA) v;
										String s = expToString(fa.getOwner())+"_"+fa.getName().getName();
										s = s.replace('.', '_');
										Type tt = th.getType(fa, cl);
										VAR x_b2 = createVar(createVarname(s+"_b"), th.createTYPE(tt), false);
										replaceDASHABLE(be,fa,x_b2);
										if (strVarDelc.indexOf(s+"_b")==-1) {
											FA new_fa = createFA((PREFIX)copyExpression(fa.getOwner()),
													createVarname( new String(fa.getName().getName())),
													false,null);
											VAR x_b = createVar(createVarname(s+"_b"), th.createTYPE(tt), true);
											exprsAssign.add(createRelOp(RelOps.EQ, x_b, new_fa));
											varsDecl.add(createVarDecl( th.createTYPE(tt), createVarname(x_b.getName().getName())));
											varsEnd.add(createEndDecl(createVarname(x_b.getName().getName())));
										}
									}
						}
					}
				}
				defPrimedVars.addAll(tmpDefPrimedVars);
				tmpDefPrimedVars.clear();
			}
			exprsDefined.addAll(0, exprsAssign);
		}
		//End do renaming of variables
		return success;
	}

	//replace a field and a parameter
	private void replaceDASHABLE(EXP be, DASHABLE da, VAR x_b2) {
		if (be instanceof BINEXP){
			replaceDASHABLE_BINEXP((BINEXP) be,da,x_b2);
		}else if (be instanceof NEG){
			replaceDASHABLE_NEG((NEG) be,da,x_b2);
		}else if (be instanceof INSTANCEOF){
			replaceDASHABLE_INSOF((INSTANCEOF) be,da,x_b2);
		}		
	}

	private void replaceDASHABLE_BINEXP(BINEXP be, DASHABLE da, VAR x_b2) {
		EXP right = be.getRight();
		EXP left = be.getLeft();
		if (right instanceof DASHABLE){
			if (right == da){
				be.setRight(x_b2);
				return;
			}
		}else replaceDASHABLE(right, da, x_b2);

		if (left instanceof DASHABLE){
			if (da == left){
				be.setLeft(x_b2);
			}
		}else replaceDASHABLE(left, da, x_b2);
	}

	private void replaceDASHABLE_INSOF(INSTANCEOF be, DASHABLE da, VAR x_b2) {
		EXP ref = be.getRef();
		if (ref instanceof DASHABLE){
			if (ref == da){
				be.setRef(x_b2);
				return;
			}
		}else replaceDASHABLE(ref, da, x_b2);
	}

	private void replaceDASHABLE_NEG(NEG be, DASHABLE da, VAR x_b2) {
		EXP e = be.getE();
		if (e instanceof DASHABLE){
			if (e == da){
				be.setE(x_b2);
			}
		}else replaceDASHABLE(e, da, x_b2);
	}
	//replace a field and a parameter
}