package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.WrappedException;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;

import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert.EPUtility;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert.ExpertPatternException;

/**
 * @author Charles Morisset
 *  Replace each x.y with the corresponding getter get_x_y(), assuming that the getter has been created 
 *  using ExpCreateGetter.  
 *  Throws an exception when the getter doesn't exist (this could be refactored by not doing anything instead) 
 */
public class ExpReplaceGetter extends ExpRefinement<ExpertPatternException> {

	final protected EPUtility epu;
	final private boolean voidOk;

	ExpReplaceGetter(Classifier c, boolean voidOk){
		if (c == null)
			throw new IllegalArgumentException();
		epu = new EPUtility(c);
		this.voidOk = voidOk;
	}

	protected ExpReplaceGetter(ExpReplaceGetter erg, boolean voidOk) {
		this(erg.epu.cl,voidOk);
	}

	protected void visit(MCALL m) throws ExpertPatternException {
		visit((PATH) m);
	}

	protected void visit(FA m) throws ExpertPatternException{
		visit((PATH) m);
	}

	protected void visit(final PATH f) throws ExpertPatternException {

		if ((f.getOwner() instanceof REF)) {
			// If this is a method call, we have to make sure to call the expert pattern 
			// also for the arguments of the call

			if (f instanceof MCALL){
				MCALL m = (MCALL) f;
				EList<EXP> mylexp = new BasicEList<EXP>();


				for (EXP e : m.getIn()){
					try {
						mylexp.add(new ExpReplaceGetter(this,false).run(e));
					} catch (ExpertPatternException e1) {
						throw new WrappedException(e1);
					}
				}
				for (int i = 0; i < mylexp.size(); i++){
					// destructive update of AST.
					m.getIn().set(i, mylexp.get(i));
				}
			}
			PATH f_copy = (PATH) DesignUtility.copyExpression(f);
			REF r = (REF) f_copy.getOwner();
			if (r instanceof PATH && ((PATH) r).getOwner() instanceof REF){
				PATH o = (PATH) r;
				REF o2 = (REF) o.getOwner();
				if (! (o2 instanceof THIS)){
					o.setOwner(MyDesign.createThis());
					final MCALL m = getMCallEP(o2, f_copy);
					setModified(m);
				} else {
					set(f_copy);
				}

			} else {
				set(f_copy);
			}
		} else {
			set(DesignUtility.copyExpression(f));
		}
	}

	/**
	 * @author Charles Morisset
	 * @param owner an object different from THIS
	 * @param path a dotted expression x.y, where x and y can be field or method calls 
	 * @return owner.get_x_y_*(in) where "in" are the arguments present in path 
	 * @throws ExpertPatternException if the getter doesn't exist
	 * 
	 */
	private MCALL getMCallEP(REF owner, final REF path) throws ExpertPatternException {
		try {
			// We first get the type of the owner
			final Class c = (Class) epu.getType(owner);
			assert(c != null);

			// We need to collect all the input parameters of the path, to propagate them.
			// exp_in contains the concrete arguments, which will be passed to the method call
			List<EXP> exp_in = new ArrayList<EXP>();

			// paras_list contains the formal parameters created for the method definition, using the 
			// parameter generator pg (see below).
			final EList<String> paras_list = new BasicEList<String>();
			//type_list[i] corresponds to the type of paras_list[i]
			final EList<Type> type_list = new BasicEList<Type>();

			final FreshNameGenerator pg = new FreshNameGenerator(epu,c);
			epu.getPathParameters(path, pg, exp_in, paras_list, type_list);

			// [vs] Call the typechecker INSIDE the comment to get up-to-date result
			// We get the type of the path					
			Type t_path = epu.getType(path, c, voidOk);

			// If the path is not void, then we need to add a return parameter
			if (! PrimitiveTypes.isVoid(t_path)) {
				paras_list.add("ret");
				type_list.add(t_path);
			}

			boolean conflict = true;
			// As long as we have a conflict, ie a method with the same signature but a different method body
			// we try we another name
			int num = 0;
			String name = "";
			do {
				// The name of the method 
				name = epu.createGetter(path, num);

				// Then we check the existing methods of c
				Operation op1 = c.getOwnedOperation(name, paras_list, type_list);
				if (op1 == null) { 
					// There is no such a method, so we throw an exception
					StringBuffer args = new StringBuffer();
					for (Type t : type_list) {
						args.append(t.getName());
					}
					throw new ExpertPatternException("The getter " + name + " doesn't exist in " + c.getLabel());

				} else if (epu.isGeneratedGetter(op1, path, voidOk)) {
					// We do not need to create a new method, so we just say that there is no conflict
					conflict = false;
				} else {
					// NADA.
					// There is already a method with this name, but it's not one we generated, so we 
					// redo the same process with a different name.
					num++;
				}
			} while (conflict);


			// Finally, we create the method call which will substitute the original path
			MCALL m = MyDesign.createMcall(owner, name, exp_in, null);
			return m;
		} catch (GetTypeException e1) {
			throw new ExpertPatternException(e1);
		}
	}
}
