package edu.unu.iist.rcos.spec.tc;

import java.util.*;

import org.antlr.runtime.RecognitionException;
import org.eclipse.core.runtime.Assert;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.util.*;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DASHABLE;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGNEXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PAR;
import edu.unu.iist.rcos.spec.antlr.visitor.check.DesignCheckType;
import edu.unu.iist.rcos.spec.antlr.visitor.check.ExpGetType;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

public abstract class RuleTC<M extends Namespace, Cmd, F extends RCOSRuleFeature<Cmd>, T extends DESIGNEXP> extends RuleIF<M,Cmd,F> {

	public RuleTC(F rf) {
		super(rf);
	}

	public String toString(M me, T t) throws WellformedException {
		if (!(me instanceof Classifier || me instanceof Operation))
			throw new IllegalArgumentException();
		if (t instanceof EXP)
			return expToString(getContext(me),(EXP) t);
		else if (t instanceof DESIGN)
			return designToString(getContext(me),(DESIGN) t);
		else
			throw new UnsupportedOperationException();
	}

	public abstract void check(final M me, final String newDoc) throws TypingException, RecognitionException, WellformedException;
	
	/**
	 * Extract rule of kind F from 'me' and type check it.
	 * @param me
	 * @param missingOK	Whether to accept a missing rule, throws WellFormedException otherwise.
	 * @throws WellformedException
	 * @throws RecognitionException 
	 */
	public void check(final M me) throws WellformedException, RecognitionException {
		String spec = rf.get(me);
		if (spec != null)
			check(me,spec);
		else
			if (!rf.getMissingOK())
				throw new WellformedException("Rule "+ getRuleName()+" missing on "+me.getQualifiedName());
	}
	
	public abstract void check(final M me, final T t) throws TypingException, PrimedVariableForbiddenException, WellformedException;
	
	/**
	 * Type check boolean expression 'newDoc' in context 'me'. Also checks for dashed variables. 
	 * @param me
	 * @param newDoc
	 * @throws TypingException
	 * @see {@link #checkBool(Namespace, EXP)}
	 */
	protected void checkBool(final M me, final String newDoc) throws RecognitionException, TypingException, WellformedException {
		EXP e = Parse.getExpFromString(newDoc);
		Assert.isNotNull(e, newDoc);
		checkBool(me,e);
	}

	protected void checkDesign(final Operation op, final String newDoc) throws RecognitionException, WellformedException {		
		checkDesign(op,Parse.getDesign(newDoc,op));
	}

	protected void checkDesign(final Operation op, final DESIGN d) throws TypingException {
		final Classifier c = (Classifier) op.getOwner();
		if (c == null)
			throw new IllegalArgumentException("Can't type check operation '"+op.getName()+"' without an owner");
		try {
			DesignCheckType dv;
			(dv = new DesignCheckType(op, mkVardecls(op))).run(d);

			/* Check that all output parameters have been assigned to */
			new CheckOutputParameters(op,dv);

			/* If the operation is marked as a query check that this is indeed the case. */
			if (op.isQuery()) {
				final Set<String> filteredList;
				if ((filteredList = isPure(d, (Classifier)op.getOwner())).size() >0)
					throw new GetTypeException("Pure method calls non-pure method(s):\n"+filteredList, d);
				if (!dv.query)
					throw new GetTypeException("Method is not pure.",d);
			}
		} catch (GetTypeException e) {
			throw new TypingException("In design for operation "+op.getQualifiedName()+": "+e.getMessage(),e);
		}
	}

	public static class TypingWarningException extends TypingException {
		/* These exceptions will be rendered as warnings only during validation.
		 * Or should we introduce error levels in the TypingException?
		 */
		private static final long serialVersionUID = 1L;
		public TypingWarningException(String msg) {
			super(msg);
		}
	}

	static class CheckOutputParameters {
		
		public CheckOutputParameters(Operation op, DesignCheckType dv) throws TypingWarningException {
			Collection<String> outs = dv.getOutParams();
			Collection<String> outArgs = new ArrayList<String>();
			for (Parameter p : op.getOwnedParameters()) {
				if (p.getDirection() != ParameterDirectionKind.IN_LITERAL)
					outArgs.add(p.getName());
			}
			for (String s : outs) {
				outArgs.remove(s);
			}
			if (!outArgs.isEmpty())
				throw new TypingWarningException("Output parameters in "+op.getQualifiedName()+" not assigned to: "+printOuts(outArgs));
		}

		private String printOuts(Collection<String> outs) {
			StringBuffer buf = new StringBuffer();
			for (String p : outs) {
				buf.append(p+" ");
			}
			return buf.toString();
		}
	}

	/**
	 * This method can be used to inject a particular context for TCing
	 * by overloading it.
	 */
	protected VardeclsFIS mkVardecls(Operation op) {
		return new VardeclsFIS(op);
	}

	/**
	 * Type checks an expression. Also checks for dashed variables and side effects.
	 * @param me
	 * @param e
	 * @throws WellformedException 
	 * @throws NotPureException 
	 */
	protected void checkBool(final M me, final EXP e) throws WellformedException {
		final Classifier ctx = getContext(me);
		final String exp = expToString(ctx,e);
		try {
			VardeclsFIS vs = mkVardecls(me);
			if (!checkBoolean(getComponent(ctx), vs, e, ctx))
				throw new TypingException(exp+" is not a boolean expression");
			for (DASHABLE var : DesignUtility.getVariables(e)) {
				if (var.isDash())
					throw new PrimedVariableForbiddenException("The expression '" + exp + "' contains dashed variables");
				if (var instanceof PAR)
					checkPAR(me, vs, (PAR) var);
			}

			{
				// check for purity
				isPure(e, ctx);
			}
		} catch (GetTypeException e1) {
			throw new TypingException(me.getQualifiedName()+" is not a boolean expression:\n"+e1.getMessage(), e1);
		}
	}

	/**
	 * Call back, so that for a guard we can check that we don't access output parameters
	 * @param vs 
	 */
	protected abstract void checkPAR(M me, Vardecls vs, PAR par) throws WellformedException;

	protected abstract VardeclsFIS mkVardecls(M me) throws WellformedException;

	/**
	 * @deprecated Not efficient :-/
	 */
	private Component getComponent(Classifier cl) throws TypingException {
		return Utility.getScope(cl);
	}

	static boolean checkBoolean(Component co, VardeclsFIS vs, EXP e, Classifier cl) throws GetTypeException {
		ExpGetType egt = new ExpGetType(co, cl, vs.th, vs);
		egt.visit(e);
		return (egt.isBooleanType()); 
	}

}
